@@ -1,7 +1,7 @@
 //
 // Core client web interface functions for the Printer Application Framework
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -144,7 +144,7 @@ papplClientGetForm(
   size_t	body_alloc,		// Allocated message body size
 		body_size = 0;		// Size of message body
   ssize_t	bytes;			// Bytes read
-  int		num_form = 0;		// Number of form variables
+  cups_len_t	num_form = 0;		// Number of form variables
   http_state_t	initial_state;		// Initial HTTP state
 
 
@@ -191,7 +191,7 @@ papplClientGetForm(
       return (0);
     }
 
-    for (bodyptr = body, bodyend = body + body_alloc; (bytes = httpRead2(client->http, bodyptr, (size_t)(bodyend - bodyptr))) > 0; bodyptr += bytes)
+    for (bodyptr = body, bodyend = body + body_alloc; (bytes = httpRead(client->http, bodyptr, (size_t)(bodyend - bodyptr))) > 0; bodyptr += bytes)
     {
       body_size += (size_t)bytes;
 
@@ -396,7 +396,7 @@ papplClientGetForm(
       {
 	if ((ptr = strstr(line + 20, " name=\"")) != NULL)
 	{
-	  strlcpy(name, ptr + 7, sizeof(name));
+	  papplCopyString(name, ptr + 7, sizeof(name));
 
 	  if ((ptr = strchr(name, '\"')) != NULL)
 	    *ptr = '\0';
@@ -404,7 +404,7 @@ papplClientGetForm(
 
 	if ((ptr = strstr(line + 20, " filename=\"")) != NULL)
 	{
-	  strlcpy(filename, ptr + 11, sizeof(filename));
+	  papplCopyString(filename, ptr + 11, sizeof(filename));
 
 	  if ((ptr = strchr(filename, '\"')) != NULL)
 	    *ptr = '\0';
@@ -416,7 +416,7 @@ papplClientGetForm(
   free(body);
 
   // Return whatever we got...
-  return (num_form);
+  return ((int)num_form);
 }
 
 
@@ -426,8 +426,9 @@ papplClientGetForm(
 // The web interface supports both authentication against user accounts and
 // authentication using a single administrative access password.  This function
 // handles the details of authentication for the web interface based on the
-// system authentication service configuration (the "auth_service" argument to
-// @link papplSystemCreate@).
+// system authentication service configuration = the "auth_service" argument to
+// @link papplSystemCreate@ and any callback set using
+// @link papplSystemSetAuthCallback@.
 //
 // > Note: IPP operation callbacks needing to perform authorization should use
 // > the @link papplClientIsAuthorized@ function instead.
@@ -446,11 +447,11 @@ papplClientHTMLAuthorize(
 
 
   // Don't authorize if we have no auth service or we don't have a password set.
-  if (!client || (!client->system->auth_service && !client->system->password_hash[0]))
+  if (!client || (!client->system->auth_service && !client->system->auth_cb && !client->system->password_hash[0]))
     return (true);
 
   // When using an auth service, use HTTP Basic authentication...
-  if (client->system->auth_service)
+  if (client->system->auth_service || client->system->auth_cb)
   {
     http_status_t code = papplClientIsAuthorized(client);
 
@@ -473,7 +474,7 @@ papplClientHTMLAuthorize(
     if (!strcmp(auth_cookie, auth_text))
     {
       // Hashes match so we are authorized.  Use "web-admin" as the username.
-      strlcpy(client->username, "web-admin", sizeof(client->username));
+      papplCopyString(client->username, "web-admin", sizeof(client->username));
 
       return (true);
     }
@@ -483,15 +484,15 @@ papplClientHTMLAuthorize(
   if (client->operation == HTTP_STATE_POST)
   {
     // Yes, grab the login information and try to authorize...
-    int			num_form = 0;	// Number of form variable
+    cups_len_t		num_form = 0;	// Number of form variable
     cups_option_t	*form = NULL;	// Form variables
     const char		*password;	// Password from user
 
-    if ((num_form = papplClientGetForm(client, &form)) == 0)
+    if ((num_form = (cups_len_t)papplClientGetForm(client, &form)) == 0)
     {
       status = "Invalid form data.";
     }
-    else if (!papplClientIsValidForm(client, num_form, form))
+    else if (!papplClientIsValidForm(client, (int)num_form, form))
     {
       status = "Invalid form submission.";
     }
@@ -530,7 +531,7 @@ papplClientHTMLAuthorize(
     if (!status)
     {
       // Hashes match so we are authorized.  Use "web-admin" as the username.
-      strlcpy(client->username, "web-admin", sizeof(client->username));
+      papplCopyString(client->username, "web-admin", sizeof(client->username));
 
       return (true);
     }
@@ -586,14 +587,14 @@ papplClientHTMLEscape(
     if (*s == '&' || *s == '<' || *s == '\"')
     {
       if (s > start)
-        httpWrite2(client->http, start, (size_t)(s - start));
+        httpWrite(client->http, start, (size_t)(s - start));
 
       if (*s == '&')
-        httpWrite2(client->http, "&amp;", 5);
+        httpWrite(client->http, "&amp;", 5);
       else if (*s == '<')
-        httpWrite2(client->http, "&lt;", 4);
+        httpWrite(client->http, "&lt;", 4);
       else
-        httpWrite2(client->http, "&quot;", 6);
+        httpWrite(client->http, "&quot;", 6);
 
       start = s + 1;
     }
@@ -602,7 +603,7 @@ papplClientHTMLEscape(
   }
 
   if (s > start)
-    httpWrite2(client->http, start, (size_t)(s - start));
+    httpWrite(client->http, start, (size_t)(s - start));
 }
 
 
@@ -619,7 +620,7 @@ void
 papplClientHTMLFooter(
     pappl_client_t *client)		// I - Client
 {
-  const char *footer = papplSystemGetFooterHTML(papplClientGetSystem(client));
+  const char *footer = papplClientGetLocString(client, papplSystemGetFooterHTML(papplClientGetSystem(client)));
 					// Footer HTML
 
   if (footer)
@@ -638,7 +639,7 @@ papplClientHTMLFooter(
   papplClientHTMLPuts(client,
 		      "  </body>\n"
 		      "</html>\n");
-  httpWrite2(client->http, "", 0);
+  httpWrite(client->http, "", 0);
 }
 
 
@@ -669,7 +670,7 @@ papplClientHTMLHeader(
 
 
   pthread_rwlock_rdlock(&system->rwlock);
-  printer = (pappl_printer_t *)cupsArrayFirst(system->printers);
+  printer = (pappl_printer_t *)cupsArrayGetFirst(system->printers);
   pthread_rwlock_unlock(&system->rwlock);
 
   if ((system->options & PAPPL_SOPTIONS_MULTI_QUEUE) || !printer)
@@ -684,7 +685,7 @@ papplClientHTMLHeader(
 			"    <title>%s%s%s</title>\n"
 			"    <link rel=\"shortcut icon\" href=\"/favicon.png\" type=\"image/png\">\n"
 			"    <link rel=\"stylesheet\" href=\"/style.css\">\n"
-			"    <meta http-equiv=\"X-UA-Compatible\" content=\"IE=9\">\n", title ? title : "", title ? " - " : "", name);
+			"    <meta http-equiv=\"X-UA-Compatible\" content=\"IE=9\">\n", title ? papplClientGetLocString(client, title) : "", title ? " - " : "", name);
   if (refresh > 0)
     papplClientHTMLPrintf(client, "<meta http-equiv=\"refresh\" content=\"%d\">\n", refresh);
   papplClientHTMLPuts(client,
@@ -704,7 +705,7 @@ papplClientHTMLHeader(
 
   if (!(system->options & PAPPL_SOPTIONS_MULTI_QUEUE) && printer)
   {
-    if (cupsArrayCount(system->links) > 0)
+    if (cupsArrayGetCount(system->links) > 0)
       papplClientHTMLPuts(client, "          <span class=\"spacer\"></span>\n");
 
     pthread_rwlock_rdlock(&printer->rwlock);
@@ -746,30 +747,41 @@ _papplClientHTMLInfo(
     papplClientHTMLStartForm(client, client->uri, false);
 
   // DNS-SD name...
-  papplClientHTMLPuts(client,
-		      "          <table class=\"form\">\n"
-		      "            <tbody>\n"
-		      "              <tr><th>Name:</th><td>");
+  papplClientHTMLPrintf(client,
+		        "          <table class=\"form\">\n"
+		        "            <tbody>\n"
+		        "              <tr><th>%s:</th><td>", papplClientGetLocString(client, _PAPPL_LOC("Name")));
   if (is_form)
-    papplClientHTMLPrintf(client, "<input type=\"text\" name=\"dns_sd_name\" value=\"%s\" placeholder=\"DNS-SD Service Name\">", dns_sd_name ? dns_sd_name : "");
+    papplClientHTMLPrintf(client, "<input type=\"text\" name=\"dns_sd_name\" value=\"%s\" placeholder=\"%s\">", dns_sd_name ? dns_sd_name : "", papplClientGetLocString(client, _PAPPL_LOC("DNS-SD Service Name")));
   else
-    papplClientHTMLEscape(client, dns_sd_name ? dns_sd_name : "Not set", 0);
+    papplClientHTMLEscape(client, dns_sd_name ? dns_sd_name : papplClientGetLocString(client, _PAPPL_LOC("Not set")), 0);
 
   // Location and geo-location...
-  papplClientHTMLPuts(client,
-                      "</td></tr>\n"
-		      "              <tr><th>Location:</th><td>");
+  papplClientHTMLPrintf(client,
+                        "</td></tr>\n"
+		        "              <tr><th>%s:</th><td>", papplClientGetLocString(client, _PAPPL_LOC("Location")));
   if (is_form)
   {
     papplClientHTMLPrintf(client,
-                          "<input type=\"text\" name=\"location\" value=\"%s\" placeholder=\"Human-Readable Location\"><br>\n"
-                          "<input type=\"number\" name=\"geo_location_lat\" min=\"-90\" max=\"90\" step=\"0.0001\" value=\"%.4f\" onChange=\"updateMap();\">&nbsp;&deg;&nbsp;latitude x <input type=\"number\" name=\"geo_location_lon\" min=\"-180\" max=\"180\" step=\"0.0001\" value=\"%.4f\" onChange=\"updateMap();\">&nbsp;&deg;&nbsp;longitude <button id=\"geo_location_lookup\" onClick=\"event.preventDefault(); navigator.geolocation.getCurrentPosition(setGeoLocation);\">Use My Position</button>", location ? location : "", lat, lon);
+                          "<input type=\"text\" name=\"location\" placeholder=\"%s\" value=\"%s\"><br>\n"
+                          "<input type=\"number\" name=\"geo_location_lat\" min=\"-90\" max=\"90\" step=\"0.0001\" value=\"%.4f\" onChange=\"updateMap();\">&nbsp;&deg;&nbsp;latitude x <input type=\"number\" name=\"geo_location_lon\" min=\"-180\" max=\"180\" step=\"0.0001\" value=\"%.4f\" onChange=\"updateMap();\">&nbsp;&deg;&nbsp;longitude", papplClientGetLocString(client, _PAPPL_LOC("Human-Readable Location")), location ? location : "", lat, lon);
+
+    if (httpIsEncrypted(client->http))
+    {
+      // If the connection is encrypted, show a button to lookup the position...
+      papplClientHTMLPrintf(client, " <button id=\"geo_location_lookup\" onClick=\"event.preventDefault(); navigator.geolocation.getCurrentPosition(setGeoLocation);\">%s</button>", papplClientGetLocString(client, _PAPPL_LOC("Use My Position")));
+    }
+    else if (!(client->system->options & PAPPL_SOPTIONS_NO_TLS))
+    {
+      // If the connection is not encrypted, redirect to a secure page...
+      papplClientHTMLPrintf(client, " <a class=\"btn\" href=\"https://%s:%d%s?get-location\">%s</a>", client->host_field, client->host_port, client->uri, papplClientGetLocString(client, _PAPPL_LOC("Use My Position")));
+    }
   }
   else
   {
-    papplClientHTMLPrintf(client, "%s", location ? location : "Not set");
+    papplClientHTMLPrintf(client, "%s", location ? location : papplClientGetLocString(client, _PAPPL_LOC("Not set")));
     if (geo_location)
-      papplClientHTMLPrintf(client, "<br>\n%g&deg;&nbsp;%c&nbsp;latitude x %g&deg;&nbsp;%c&nbsp;longitude", fabs(lat), lat < 0.0 ? 'S' : 'N', fabs(lon), lon < 0.0 ? 'W' : 'E');
+      papplClientHTMLPrintf(client, "<br>\n%g&deg;&nbsp;%c %g&deg;&nbsp;%c", fabs(lat), lat < 0.0 ? 'S' : 'N', fabs(lon), lon < 0.0 ? 'W' : 'E');
   }
 
   // Show an embedded map of the location...
@@ -779,29 +791,29 @@ _papplClientHTMLInfo(
 			  "<iframe id=\"map\" frameborder=\"0\" scrolling=\"no\" marginheight=\"0\" marginwidth=\"0\" src=\"https://www.openstreetmap.org/export/embed.html?bbox=%g,%g,%g,%g&amp;layer=mapnik&amp;marker=%g,%g\"></iframe>", lon - 0.00025, lat - 0.00025, lon + 0.00025, lat + 0.00025, lat, lon);
 
   // Organization
-  papplClientHTMLPuts(client,
-                      "</td></tr>\n"
-		      "              <tr><th>Organization:</th><td>");
+  papplClientHTMLPrintf(client,
+                        "</td></tr>\n"
+		        "              <tr><th>%s:</th><td>", papplClientGetLocString(client, _PAPPL_LOC("Organization")));
 
   if (is_form)
     papplClientHTMLPrintf(client,
-                          "<input type=\"text\" name=\"organization\" placeholder=\"Organization Name\" value=\"%s\"><br>\n"
-                          "<input type=\"text\" name=\"organizational_unit\" placeholder=\"Organizational Unit\" value=\"%s\">", organization ? organization : "", org_unit ? org_unit : "");
+                          "<input type=\"text\" name=\"organization\" placeholder=\"%s\" value=\"%s\"><br>\n"
+                          "<input type=\"text\" name=\"organizational_unit\" placeholder=\"%s\" value=\"%s\">", papplClientGetLocString(client, _PAPPL_LOC("Organization Name")), organization ? organization : "", papplClientGetLocString(client, _PAPPL_LOC("Organizational Unit")), org_unit ? org_unit : "");
   else
-    papplClientHTMLPrintf(client, "%s%s%s", organization ? organization : "Not set", org_unit ? ", " : "", org_unit ? org_unit : "");
+    papplClientHTMLPrintf(client, "%s%s%s", organization ? organization : papplClientGetLocString(client, _PAPPL_LOC("Not set")), org_unit ? ", " : "", org_unit ? org_unit : "");
 
   // Contact
-  papplClientHTMLPuts(client,
-                      "</td></tr>\n"
-                      "              <tr><th>Contact:</th><td>");
+  papplClientHTMLPrintf(client,
+                        "</td></tr>\n"
+                        "              <tr><th>%s:</th><td>", papplClientGetLocString(client, _PAPPL_LOC("Contact")));
 
   if (is_form)
   {
     papplClientHTMLPrintf(client,
-                          "<input type=\"text\" name=\"contact_name\" placeholder=\"Name\" value=\"%s\"><br>\n"
+                          "<input type=\"text\" name=\"contact_name\" placeholder=\"%s\" value=\"%s\"><br>\n"
                           "<input type=\"email\" name=\"contact_email\" placeholder=\"name@domain\" value=\"%s\"><br>\n"
                           "<input type=\"tel\" name=\"contact_telephone\" placeholder=\"867-5309\" value=\"%s\"></td></tr>\n"
-		      "              <tr><th></th><td><input type=\"submit\" value=\"Save Changes\">", contact->name, contact->email, contact->telephone);
+		      "              <tr><th></th><td><input type=\"submit\" value=\"Save Changes\">", papplClientGetLocString(client, _PAPPL_LOC("Name")), contact->name, contact->email, contact->telephone);
   }
   else if (contact->email[0])
   {
@@ -822,7 +834,7 @@ _papplClientHTMLInfo(
     papplClientHTMLPrintf(client, "<a href=\"tel:%s\">%s</a>", contact->telephone, contact->telephone);
   }
   else
-    papplClientHTMLPuts(client, "Not set");
+    papplClientHTMLPuts(client, papplClientGetLocString(client, _PAPPL_LOC("Not set")));
 
   papplClientHTMLPuts(client,
 		      "</td></tr>\n"
@@ -856,8 +868,10 @@ _papplClientHTMLInfo(
                         "  document.forms['form']['geo_location_lat'].value = lat;\n"
                         "  document.forms['form']['geo_location_lon'].value = lon;\n"
                         "  updateMap();\n"
-                        "}\n"
-                        "</script>\n");
+                        "}\n");
+    if (client->options && !strcmp(client->options, "get-location"))
+      papplClientHTMLPuts(client, "navigator.geolocation.getCurrentPosition(setGeoLocation);\n");
+    papplClientHTMLPuts(client, "</script>\n");
   }
 }
 
@@ -920,7 +934,8 @@ papplClientHTMLPrinterHeader(
     {
       char	full_title[1024];	// Full title
 
-      snprintf(full_title, sizeof(full_title), "%s - %s", title, printer->name);
+      // Need to grab the localized title here since the title includes the printer name...
+      snprintf(full_title, sizeof(full_title), "%s - %s", papplClientGetLocString(client, title), printer->name);
       papplClientHTMLHeader(client, full_title, refresh);
     }
     else
@@ -930,7 +945,7 @@ papplClientHTMLPrinterHeader(
   }
   else
   {
-    // Single queue mode - the function will automatically add the printer name...
+    // Single queue mode - the function will automatically add the printer name and localize the title...
     papplClientHTMLHeader(client, title, refresh);
   }
 
@@ -965,9 +980,9 @@ papplClientHTMLPrinterHeader(
     papplClientHTMLPrintf(client,
 			  "      <div class=\"row\">\n"
 			  "        <div class=\"col-12\">\n"
-			  "          <h1 class=\"title\">%s", title);
+			  "          <h1 class=\"title\">%s", papplClientGetLocString(client, title));
     if (label && path_or_url)
-      papplClientHTMLPrintf(client, " <a class=\"btn\" href=\"%s\">%s</a>", path_or_url, label);
+      papplClientHTMLPrintf(client, " <a class=\"btn\" href=\"%s\">%s</a>", path_or_url, papplClientGetLocString(client, label));
     papplClientHTMLPuts(client, "</h1>\n");
   }
 }
@@ -1002,6 +1017,7 @@ papplClientHTMLPrintf(
 
 
   // Loop through the format string, formatting as needed...
+  // TODO: Support positional parameters, e.g. "%2$s" to access the second string argument
   va_start(ap, format);
   start = format;
 
@@ -1010,14 +1026,14 @@ papplClientHTMLPrintf(
     if (*format == '%')
     {
       if (format > start)
-        httpWrite2(client->http, start, (size_t)(format - start));
+        httpWrite(client->http, start, (size_t)(format - start));
 
       tptr    = tformat;
       *tptr++ = *format++;
 
       if (*format == '%')
       {
-        httpWrite2(client->http, "%", 1);
+        httpWrite(client->http, "%", 1);
         format ++;
 	start = format;
 	continue;
@@ -1124,7 +1140,7 @@ papplClientHTMLPrintf(
 
 	    snprintf(temp, sizeof(temp), tformat, va_arg(ap, double));
 
-            httpWrite2(client->http, temp, strlen(temp));
+            httpWrite(client->http, temp, strlen(temp));
 	    break;
 
         case 'B' : // Integer formats
@@ -1148,7 +1164,7 @@ papplClientHTMLPrintf(
 	    else
 	      snprintf(temp, sizeof(temp), tformat, va_arg(ap, int));
 
-            httpWrite2(client->http, temp, strlen(temp));
+            httpWrite(client->http, temp, strlen(temp));
 	    break;
 
 	case 'p' : // Pointer value
@@ -1157,7 +1173,7 @@ papplClientHTMLPrintf(
 
 	    snprintf(temp, sizeof(temp), tformat, va_arg(ap, void *));
 
-            httpWrite2(client->http, temp, strlen(temp));
+            httpWrite(client->http, temp, strlen(temp));
 	    break;
 
         case 'c' : // Character or character array
@@ -1184,7 +1200,7 @@ papplClientHTMLPrintf(
   }
 
   if (format > start)
-    httpWrite2(client->http, start, (size_t)(format - start));
+    httpWrite(client->http, start, (size_t)(format - start));
 
   va_end(ap);
 }
@@ -1200,32 +1216,33 @@ _papplClientHTMLPutLinks(
     cups_array_t     *links,		// I - Array of links
     pappl_loptions_t which)		// I - Which links to show
 {
-  int			i,		// Looping var
-			count;		// Number of links
-  _pappl_link_t		*l;		// Current link
+  size_t	i,			// Looping var
+		count;			// Number of links
+  _pappl_link_t	*l;			// Current link
+  const char	*webscheme = _papplClientGetAuthWebScheme(client);
+					// URL scheme for authenticated links
 
 
   // Loop through the links.
   //
-  // Note: We use a loop and not cupsArrayFirst/Last because other threads may
+  // Note: We use a loop and not cupsArrayGetFirst/Last because other threads may
   // be enumerating the same array of links.
-
-  for (i = 0, count = cupsArrayCount(links); i < count; i ++)
+  for (i = 0, count = cupsArrayGetCount(links); i < count; i ++)
   {
-    l = (_pappl_link_t *)cupsArrayIndex(links, i);
+    l = (_pappl_link_t *)cupsArrayGetElement(links, i);
 
     if (!l || !(l->options & which))
       continue;
 
     if (strcmp(client->uri, l->path_or_url))
     {
-      if (l->path_or_url[0] != '/' || !(l->options & PAPPL_LOPTIONS_HTTPS_REQUIRED) || (!client->system->auth_service && !client->system->password_hash[0]))
-	papplClientHTMLPrintf(client, "          <a class=\"btn\" href=\"%s\">%s</a>\n", l->path_or_url, l->label);
+      if (l->path_or_url[0] != '/' || !(l->options & PAPPL_LOPTIONS_HTTPS_REQUIRED))
+	papplClientHTMLPrintf(client, "          <a class=\"btn\" href=\"%s\">%s</a>\n", l->path_or_url, papplClientGetLocString(client, l->label));
       else
-	papplClientHTMLPrintf(client, "          <a class=\"btn\" href=\"https://%s:%d%s\">%s</a>\n", client->host_field, client->host_port, l->path_or_url, l->label);
+	papplClientHTMLPrintf(client, "          <a class=\"btn\" href=\"%s://%s:%d%s\">%s</a>\n", webscheme, client->host_field, client->host_port, l->path_or_url, papplClientGetLocString(client, l->label));
     }
     else
-      papplClientHTMLPrintf(client, "          <span class=\"active\">%s</span>\n", l->label);
+      papplClientHTMLPrintf(client, "          <span class=\"active\">%s</span>\n", papplClientGetLocString(client, l->label));
   }
 }
 
@@ -1243,7 +1260,7 @@ papplClientHTMLPuts(
     const char     *s)			// I - String
 {
   if (client && s && *s)
-    httpWrite2(client->http, s, strlen(s));
+    httpWrite(client->http, s, strlen(s));
 }
 
 
@@ -1303,7 +1320,7 @@ papplClientIsValidForm(
   const char	*session;		// Form variable
 
 
-  if ((session = cupsGetOption("session", num_form, form)) == NULL)
+  if ((session = cupsGetOption("session", (cups_len_t)num_form, form)) == NULL)
     return (false);
 
   return (!strcmp(session, papplClientGetCSRFToken(client, token, sizeof(token))));
@@ -1342,7 +1359,7 @@ papplClientSetCookie(
     return;
 
   if (expires > 0)
-    snprintf(cookie, sizeof(cookie), "%s=%s; path=/; expires=%s; httponly; secure;", name, value, httpGetDateString2(time(NULL) + expires, expireTime, sizeof(expireTime)));
+    snprintf(cookie, sizeof(cookie), "%s=%s; path=/; expires=%s; httponly; secure;", name, value, httpGetDateString(time(NULL) + expires, expireTime, sizeof(expireTime)));
   else
     snprintf(cookie, sizeof(cookie), "%s=%s; path=/; httponly; secure;", name, value);
 
