@@ -1,7 +1,7 @@
 //
 // Client processing code for the Printer Application Framework
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -128,9 +128,15 @@ _papplClientCreateTempFile(
 
   close(fd);
 
-  client->files[client->num_files ++] = strdup(tempfile);
+  if ((client->files[client->num_files] = strdup(tempfile)) != NULL)
+    client->num_files ++;
+  else
+    unlink(tempfile);
 
-  return (client->files[client->num_files - 1]);
+  if (client->num_files > 0)
+    return (client->files[client->num_files - 1]);
+  else
+    return (NULL);
 }
 
 
@@ -145,6 +151,10 @@ void
 _papplClientDelete(
     pappl_client_t *client)		// I - Client
 {
+  pappl_system_t *system = client->system;
+					// System
+
+
   papplLogClient(client, PAPPL_LOGLEVEL_INFO, "Closing connection from '%s'.", client->hostname);
 
   // Flush pending writes before closing...
@@ -159,6 +169,11 @@ _papplClientDelete(
   ippDelete(client->response);
 
   free(client);
+
+  // Update the number of active clients...
+  pthread_rwlock_wrlock(&system->rwlock);
+  system->num_clients --;
+  pthread_rwlock_unlock(&system->rwlock);
 }
 
 
@@ -182,31 +197,13 @@ _papplClientProcessHTTP(
   int			port;		// Port number
   char			*ptr;		// Pointer into string
   _pappl_resource_t	*resource;	// Current resource
-  static const char * const http_states[] =
-  {					// Strings for logging HTTP method
-    "WAITING",
-    "OPTIONS",
-    "GET",
-    "GET_SEND",
-    "HEAD",
-    "POST",
-    "POST_RECV",
-    "POST_SEND",
-    "PUT",
-    "PUT_RECV",
-    "DELETE",
-    "TRACE",
-    "CONNECT",
-    "STATUS",
-    "UNKNOWN_METHOD",
-    "UNKNOWN_VERSION"
-  };
 
 
   // Clear state variables...
   ippDelete(client->request);
   ippDelete(client->response);
 
+  client->loc       = NULL;
   client->request   = NULL;
   client->response  = NULL;
   client->operation = HTTP_STATE_WAITING;
@@ -257,13 +254,16 @@ _papplClientProcessHTTP(
 
   if (http_status != HTTP_STATUS_OK)
   {
+    papplLogClient(client, PAPPL_LOGLEVEL_DEBUG, "http_status=%d", http_status);
     papplClientRespond(client, HTTP_STATUS_BAD_REQUEST, NULL, NULL, 0, 0);
     return (false);
   }
 
   http_version = httpGetVersion(client->http);
 
-  papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s %s://%s%s HTTP/%d.%d", http_states[http_state], httpIsEncrypted(client->http) ? "https" : "http", httpGetField(client->http, HTTP_FIELD_HOST), uri, http_version / 100, http_version % 100);
+  papplCopyString(client->language, httpGetField(client->http, HTTP_FIELD_ACCEPT_LANGUAGE), sizeof(client->language));
+
+  papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s %s://%s%s HTTP/%d.%d (%s)", httpStateString(http_state), httpIsEncrypted(client->http) ? "https" : "http", httpGetField(client->http, HTTP_FIELD_HOST), uri, http_version / 100, http_version % 100, client->language);
 
   // Validate the host header...
   if (!httpGetField(client->http, HTTP_FIELD_HOST)[0] &&
@@ -274,7 +274,7 @@ _papplClientProcessHTTP(
     return (false);
   }
 
-  strlcpy(client->host_field, httpGetField(client->http, HTTP_FIELD_HOST), sizeof(client->host_field));
+  papplCopyString(client->host_field, httpGetField(client->http, HTTP_FIELD_HOST), sizeof(client->host_field));
   if ((ptr = strrchr(client->host_field, ':')) != NULL)
   {
     char *end;				// End of port number
@@ -304,7 +304,7 @@ _papplClientProcessHTTP(
   // Handle HTTP Upgrade...
   if (!strcasecmp(httpGetField(client->http, HTTP_FIELD_CONNECTION), "Upgrade"))
   {
-    if (strstr(httpGetField(client->http, HTTP_FIELD_UPGRADE), "TLS/") != NULL && !httpIsEncrypted(client->http))
+    if (strstr(httpGetField(client->http, HTTP_FIELD_UPGRADE), "TLS/") != NULL && !httpIsEncrypted(client->http) && !(client->system->options & PAPPL_SOPTIONS_NO_TLS))
     {
       if (!papplClientRespond(client, HTTP_STATUS_SWITCHING_PROTOCOLS, NULL, NULL, 0, 0))
         return (false);
@@ -349,7 +349,7 @@ _papplClientProcessHTTP(
 
     case HTTP_STATE_HEAD :
         // See if we have a matching resource to serve...
-        if ((resource = _papplSystemFindResource(client->system, client->uri)) != NULL)
+        if ((resource = _papplSystemFindResourceForPath(client->system, client->uri)) != NULL)
         {
           if (eval_if_modified(client, resource))
 	    return (papplClientRespond(client, HTTP_STATUS_OK, NULL, resource->format, resource->last_modified, 0));
@@ -362,7 +362,7 @@ _papplClientProcessHTTP(
 
     case HTTP_STATE_GET :
         // See if we have a matching resource to serve...
-        if ((resource = _papplSystemFindResource(client->system, client->uri)) != NULL)
+        if ((resource = _papplSystemFindResourceForPath(client->system, client->uri)) != NULL)
         {
           if (!eval_if_modified(client, resource))
           {
@@ -389,9 +389,9 @@ _papplClientProcessHTTP(
 	      }
 
               while ((bytes = read(fd, buffer, sizeof(buffer))) > 0)
-                httpWrite2(client->http, buffer, (size_t)bytes);
+                httpWrite(client->http, buffer, (size_t)bytes);
 
-	      httpWrite2(client->http, "", 0);
+	      httpWrite(client->http, "", 0);
 
 	      close(fd);
 
@@ -404,7 +404,7 @@ _papplClientProcessHTTP(
 	    if (!papplClientRespond(client, HTTP_STATUS_OK, NULL, resource->format, resource->last_modified, resource->length))
 	      return (false);
 
-	    httpWrite2(client->http, (const char *)resource->data, resource->length);
+	    httpWrite(client->http, (const char *)resource->data, resource->length);
 	    httpFlushWrite(client->http);
 	    return (true);
 	  }
@@ -432,7 +432,7 @@ _papplClientProcessHTTP(
 	  // Now that we have the IPP request, process the request...
 	  return (_papplClientProcessIPP(client));
 	}
-	else if ((resource = _papplSystemFindResource(client->system, client->uri)) != NULL)
+	else if ((resource = _papplSystemFindResourceForPath(client->system, client->uri)) != NULL)
         {
 	  // Serve a matching resource...
           if (resource->cb)
@@ -489,13 +489,14 @@ papplClientRespond(
     time_t         last_modified,	// I - Last-Modified date/time or `0` for none
     size_t         length)		// I - Length of response or `0` for variable-length
 {
-  char	message[1024];			// Text message
+  char	message[1024],			// Text message
+	last_str[256];			// Date string
 
 
   if (type)
-    papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s %s %d", httpStatus(code), type, (int)length);
+    papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s %s %d", httpStatusString(code), type, (int)length);
   else
-    papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s", httpStatus(code));
+    papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s", httpStatusString(code));
 
   if (code == HTTP_STATUS_CONTINUE)
   {
@@ -506,7 +507,7 @@ papplClientRespond(
   // Format an error message...
   if (!type && !length && code != HTTP_STATUS_OK && code != HTTP_STATUS_SWITCHING_PROTOCOLS)
   {
-    snprintf(message, sizeof(message), "%d - %s\n", code, httpStatus(code));
+    snprintf(message, sizeof(message), "%d - %s\n", code, httpStatusString(code));
 
     type   = "text/plain";
     length = strlen(message);
@@ -518,7 +519,7 @@ papplClientRespond(
   httpClearFields(client->http);
   httpSetField(client->http, HTTP_FIELD_SERVER, papplSystemGetServerHeader(client->system));
   if (last_modified)
-    httpSetField(client->http, HTTP_FIELD_LAST_MODIFIED, httpGetDateString(last_modified));
+    httpSetField(client->http, HTTP_FIELD_LAST_MODIFIED, httpGetDateString(last_modified, last_str, sizeof(last_str)));
 
   if (code == HTTP_STATUS_METHOD_NOT_ALLOWED || client->operation == HTTP_STATE_OPTIONS)
     httpSetField(client->http, HTTP_FIELD_ALLOW, "GET, HEAD, OPTIONS, POST");
@@ -527,7 +528,7 @@ papplClientRespond(
   {
     char	value[HTTP_MAX_VALUE];	// WWW-Authenticate value
 
-    snprintf(value, sizeof(value), "Basic realm=\"%s\"", client->system->name);
+    snprintf(value, sizeof(value), "%s realm=\"%s\"", client->system->auth_scheme ? client->system->auth_scheme : "Basic", client->system->name);
     httpSetField(client->http, HTTP_FIELD_WWW_AUTHENTICATE, value);
   }
 
@@ -564,7 +565,7 @@ papplClientRespond(
     if (httpPrintf(client->http, "%s", message) < 0)
       return (false);
 
-    if (httpWrite2(client->http, "", 0) < 0)
+    if (httpWrite(client->http, "", 0) < 0)
       return (false);
   }
   else if (client->response)
@@ -595,7 +596,7 @@ papplClientRespondRedirect(
     http_status_t  code,		// I - `HTTP_STATUS_MOVED_PERMANENTLY` or `HTTP_STATUS_FOUND`
     const char     *path)		// I - Redirection path/URL
 {
-  papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s %s", httpStatus(code), path);
+  papplLogClient(client, PAPPL_LOGLEVEL_INFO, "%s %s", httpStatusString(code), path);
 
   // Send the HTTP response header...
   httpClearFields(client->http);
@@ -623,7 +624,7 @@ papplClientRespondRedirect(
   if (httpWriteResponse(client->http, code) < 0)
     return (false);
 
-  return (httpWrite2(client->http, "", 0) >= 0);
+  return (httpWrite(client->http, "", 0) >= 0);
 }
 
 
@@ -641,7 +642,7 @@ _papplClientRun(
   // Loop until we are out of requests or timeout (30 seconds)...
   while (httpWait(client->http, 30000))
   {
-    if (first_time)
+    if (first_time && !(client->system->options & PAPPL_SOPTIONS_NO_TLS))
     {
       // See if we need to negotiate a TLS connection...
       char buf[1];			// First byte from client
@@ -709,7 +710,7 @@ eval_if_modified(
     {
       char	*next;			// Next character
 
-      size = strtoll(ptr + 7, &next, 10);
+      size = (off_t)strtoll(ptr + 7, &next, 10);
       if (!next)
       {
         papplLogClient(client, PAPPL_LOGLEVEL_DEBUG, "If-Modified-Since: Bad length.");
