@@ -367,7 +367,7 @@ pappl_usb_find(
 			copy_did[1024],	// Copy of device ID
 			temp[256];	// Temporary string for serial #
 
-	      strlcpy(copy_did, device_id, sizeof(copy_did));
+	      papplCopyString(copy_did, device_id, sizeof(copy_did));
 
               if ((make = strstr(copy_did, "MANUFACTURER:")) != NULL)
                 make += 13;
@@ -589,7 +589,6 @@ pappl_usb_open_cb(
 
   return (match);
 }
-#endif // HAVE_LIBUSB
 
 
 //
@@ -679,3 +678,4 @@ pappl_usb_write(pappl_device_t *device,	// I - Device
   else
     return ((ssize_t)icount);
 }
+#endif // HAVE_LIBUSB
