@@ -1,7 +1,7 @@
 //
 // Common device support code for the Printer Application Framework
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2007-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -32,6 +32,7 @@ typedef struct _pappl_devscheme_s	// Device scheme data
   pappl_devwrite_cb_t	write_cb;		// Write callback
   pappl_devid_cb_t	id_cb;			// IEEE-1284 device ID callback, if any
   pappl_devstatus_cb_t	status_cb;		// Status callback, if any
+  pappl_devsupplies_cb_t supplies_cb;		// Supplies callback, if any
 } _pappl_devscheme_t;
 
 
@@ -96,7 +97,7 @@ static ssize_t		pappl_write(pappl_device_t *device, const void *buffer, size_t b
 void
 papplDeviceAddScheme(
     const char           *scheme,	// I - URI scheme
-    pappl_devtype_t        dtype,		// I - Device type (`PAPPL_DEVTYPE_CUSTOM_LOCAL` or `PAPPL_DEVTYPE_CUSTOM_NETWORK`)
+    pappl_devtype_t      dtype,		// I - Device type (`PAPPL_DEVTYPE_CUSTOM_LOCAL` or `PAPPL_DEVTYPE_CUSTOM_NETWORK`)
     pappl_devlist_cb_t   list_cb,	// I - List devices callback, if any
     pappl_devopen_cb_t   open_cb,	// I - Open callback
     pappl_devclose_cb_t  close_cb,	// I - Close callback
@@ -104,6 +105,63 @@ papplDeviceAddScheme(
     pappl_devwrite_cb_t  write_cb,	// I - Write callback
     pappl_devstatus_cb_t status_cb,	// I - Status callback, if any
     pappl_devid_cb_t     id_cb)		// I - IEEE-1284 device ID callback, if any
+{
+  papplDeviceAddScheme2(scheme, dtype, list_cb, open_cb, close_cb, read_cb, write_cb, status_cb, NULL, id_cb);
+}
+
+
+//
+// 'papplDeviceAddScheme2()' - Add a device URI scheme with supply-level queries.
+//
+// This function registers a device URI scheme with PAPPL, so that devices using
+// the named scheme can receive print data, report status information, and so
+// forth.  PAPPL includes support for the following URI schemes:
+//
+// - `dnssd`: Network printers discovered using DNS-SD.
+// - `file`: Character device files, plain files, and directories.
+// - `snmp`: Network printers discovered using SNMPv1.
+// - `socket`: Network printers using a hostname or numeric IP address.
+// - `usb`: Class 1 (unidirectional) or 2 (bidirectional) USB printers.
+//
+// The "scheme" parameter specifies the URI scheme and must consist of lowercase
+// letters, digits, "-", "_", and/or ".", for example "x-foo" or
+// "com.example.bar".
+//
+// The "dtype" parameter specifies the device type and should be
+// `PAPPL_DTYPE_CUSTOM_LOCAL` for locally connected printers and
+// `PAPPL_DTYPE_CUSTOM_NETWORK` for network printers.
+//
+// Each of the callbacks corresponds to one of the `papplDevice` functions:
+//
+// - "list_cb": Implements discovery of devices (optional)
+// - "open_cb": Opens communication with a device and allocates any device-
+//   specific data as needed
+// - "close_cb": Closes communication with a device and frees any device-
+//   specific data as needed
+// - "read_cb": Reads data from a device
+// - "write_cb": Write data to a device
+// - "status_cb": Gets basic printer state information from a device (optional)
+// - "supplies_cb": Gets supply level information from a device (optional)
+// - "id_cb": Gets the current IEEE-1284 device ID from a device (optional)
+//
+// The "open_cb" callback typically calls @link papplDeviceSetData@ to store a
+// pointer to contextual information for the connection while the "close_cb",
+// "id_cb", "read_cb", "write_cb", "status_cb", and "supplies_cb" callbacks
+// typically call @link papplDeviceGetData@ to retrieve it.
+//
+
+void
+papplDeviceAddScheme2(
+    const char             *scheme,	// I - URI scheme
+    pappl_devtype_t        dtype,	// I - Device type (`PAPPL_DEVTYPE_CUSTOM_LOCAL` or `PAPPL_DEVTYPE_CUSTOM_NETWORK`)
+    pappl_devlist_cb_t     list_cb,	// I - List devices callback, if any
+    pappl_devopen_cb_t     open_cb,	// I - Open callback
+    pappl_devclose_cb_t    close_cb,	// I - Close callback
+    pappl_devread_cb_t     read_cb,	// I - Read callback
+    pappl_devwrite_cb_t    write_cb,	// I - Write callback
+    pappl_devstatus_cb_t   status_cb,	// I - Status callback, if any
+    pappl_devsupplies_cb_t supplies_cb,	// I - Supply level callback, if any
+    pappl_devid_cb_t       id_cb)	// I - IEEE-1284 device ID callback, if any
 {
   _pappl_devscheme_t	*ds,		// Device URI scheme data
 			dkey;		// Search key
@@ -114,7 +172,7 @@ papplDeviceAddScheme(
   // Create the schemes array as needed...
   if (!device_schemes)
   {
-    if ((device_schemes = cupsArrayNew((cups_array_func_t)pappl_compare_schemes, NULL)) == NULL)
+    if ((device_schemes = cupsArrayNew((cups_array_cb_t)pappl_compare_schemes, NULL, NULL, 0, NULL, NULL)) == NULL)
     {
       pthread_rwlock_unlock(&device_rwlock);
       return;
@@ -134,14 +192,15 @@ papplDeviceAddScheme(
   {
     if ((ds->scheme = strdup(scheme)) != NULL)
     {
-      ds->dtype     = dtype;
-      ds->list_cb   = list_cb;
-      ds->open_cb   = open_cb;
-      ds->close_cb  = close_cb;
-      ds->read_cb   = read_cb;
-      ds->write_cb  = write_cb;
-      ds->status_cb = status_cb;
-      ds->id_cb     = id_cb;
+      ds->dtype       = dtype;
+      ds->list_cb     = list_cb;
+      ds->open_cb     = open_cb;
+      ds->close_cb    = close_cb;
+      ds->read_cb     = read_cb;
+      ds->write_cb    = write_cb;
+      ds->status_cb   = status_cb;
+      ds->supplies_cb = supplies_cb;
+      ds->id_cb       = id_cb;
 
       cupsArrayAdd(device_schemes, ds);
     }
@@ -163,16 +222,16 @@ void
 _papplDeviceAddSupportedSchemes(
     ipp_t *attrs)			// I - Attributes
 {
-  int			i;		// Looping var
+  cups_len_t		i;		// Looping var
   ipp_attribute_t	*attr;		// IPP attribute
   _pappl_devscheme_t	*devscheme;	// Current device scheme
 
 
   pthread_rwlock_rdlock(&device_rwlock);
 
-  attr = ippAddStrings(attrs, IPP_TAG_SYSTEM, IPP_TAG_URISCHEME, "smi2699-device-uri-schemes-supported", cupsArrayCount(device_schemes), NULL, NULL);
+  attr = ippAddStrings(attrs, IPP_TAG_SYSTEM, IPP_TAG_URISCHEME, "smi2699-device-uri-schemes-supported", IPP_NUM_CAST cupsArrayGetCount(device_schemes), NULL, NULL);
 
-  for (i = 0, devscheme = (_pappl_devscheme_t *)cupsArrayFirst(device_schemes); devscheme; i ++, devscheme = (_pappl_devscheme_t *)cupsArrayNext(device_schemes))
+  for (i = 0, devscheme = (_pappl_devscheme_t *)cupsArrayGetFirst(device_schemes); devscheme; i ++, devscheme = (_pappl_devscheme_t *)cupsArrayGetNext(device_schemes))
     ippSetString(attrs, &attr, i, devscheme->scheme);
 
   pthread_rwlock_unlock(&device_rwlock);
@@ -410,6 +469,36 @@ papplDeviceGetStatus(
 }
 
 
+//
+// 'papplDeviceGetSupplies()' - Get the current printer supplies.
+//
+// This function returns the number, type, and level of current printer supply
+// levels, as applicable to the current device.
+//
+// The supply levels for network devices come from the prtSupplyTable and
+// prtMarkerColorantTable properties that are defined in the SNMP Printer MIB
+// v2 (RFC 3805).
+//
+// The supply levels for other devices are not standardized and must be queried
+// using other methods.
+//
+// > Note: This function can block for several seconds while getting the supply
+// > information.
+//
+
+int					// O - Number of supplies
+papplDeviceGetSupplies(
+    pappl_device_t *device,		// I - Device
+    int            max_supplies,	// I - Maximum supplies
+    pappl_supply_t *supplies)		// I - Supplies
+{
+  if (device && device->supplies_cb)
+    return ((device->supplies_cb)(device, max_supplies, supplies));
+  else
+    return (0);
+}
+
+
 //
 // 'papplDeviceIsSupported()' - Determine whether a given URI is supported.
 //
@@ -510,7 +599,7 @@ papplDeviceList(
   if (!err_cb)
     err_cb = pappl_default_error_cb;
 
-  for (ds = (_pappl_devscheme_t *)cupsArrayFirst(device_schemes); ds && !ret; ds = (_pappl_devscheme_t *)cupsArrayNext(device_schemes))
+  for (ds = (_pappl_devscheme_t *)cupsArrayGetFirst(device_schemes); ds && !ret; ds = (_pappl_devscheme_t *)cupsArrayGetNext(device_schemes))
   {
     if ((types & ds->dtype) && ds->list_cb)
       ret = (ds->list_cb)(cb, data, err_cb, err_data);
@@ -593,13 +682,14 @@ papplDeviceOpen(
     return (NULL);
   }
 
-  device->close_cb   = ds->close_cb;
-  device->error_cb   = err_cb ? err_cb : pappl_default_error_cb;
-  device->error_data = err_data;
-  device->id_cb      = ds->id_cb;
-  device->read_cb    = ds->read_cb;
-  device->status_cb  = ds->status_cb;
-  device->write_cb   = ds->write_cb;
+  device->close_cb     = ds->close_cb;
+  device->error_cb     = err_cb ? err_cb : pappl_default_error_cb;
+  device->error_data   = err_data;
+  device->id_cb        = ds->id_cb;
+  device->read_cb      = ds->read_cb;
+  device->status_cb    = ds->status_cb;
+  device->supplies_cb  = ds->supplies_cb;
+  device->write_cb     = ds->write_cb;
 
   if (!(ds->open_cb)(device, device_uri, name))
   {
@@ -624,10 +714,10 @@ papplDeviceParseID(
     const char    *device_id,		// I - IEEE-1284 device ID string
     cups_option_t **pairs)		// O - Key/value pairs
 {
-  int	num_pairs = 0;			// Number of key/value pairs
-  char	name[256],			// Key name
-	value[256],			// Value
-	*ptr;				// Pointer into key/value
+  cups_len_t	num_pairs = 0;		// Number of key/value pairs
+  char		name[256],		// Key name
+		value[256],		// Value
+		*ptr;			// Pointer into key/value
 
 
   // Range check input...
@@ -679,7 +769,7 @@ papplDeviceParseID(
     num_pairs = cupsAddOption(name, value, num_pairs, pairs);
   }
 
-  return (num_pairs);
+  return ((int)num_pairs);
 }
 
 
