@@ -1,7 +1,7 @@
 //
 // DNS-SD support for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -45,7 +45,7 @@ static AvahiThreadedPoll *pappl_dns_sd_poll = NULL;
 //
 
 static void		dns_sd_geo_to_loc(const char *geo, unsigned char loc[16]);
-#ifdef HAVE_DNSSD
+#ifdef HAVE_MDNSRESPONDER
 static void DNSSD_API	dns_sd_printer_callback(DNSServiceRef sdRef, DNSServiceFlags flags, DNSServiceErrorType errorCode, const char *name, const char *regtype, const char *domain, pappl_printer_t *printer);
 static void		*dns_sd_run(void *data);
 static void DNSSD_API	dns_sd_system_callback(DNSServiceRef sdRef, DNSServiceFlags flags, DNSServiceErrorType errorCode, const char *name, const char *regtype, const char *domain, pappl_system_t *system);
@@ -53,7 +53,7 @@ static void DNSSD_API	dns_sd_system_callback(DNSServiceRef sdRef, DNSServiceFlag
 static void		dns_sd_client_cb(AvahiClient *c, AvahiClientState state, void *data);
 static void		dns_sd_printer_callback(AvahiEntryGroup *p, AvahiEntryGroupState state, pappl_printer_t *printer);
 static void		dns_sd_system_callback(AvahiEntryGroup *p, AvahiEntryGroupState state, pappl_system_t *system);
-#endif // HAVE_DNSSD
+#endif // HAVE_MDNSRESPONDER
 
 
 //
@@ -75,7 +75,7 @@ _pappl_dns_sd_t				// O - DNS-SD master reference
 _papplDNSSDInit(
     pappl_system_t *system)		// I - System
 {
-#ifdef HAVE_DNSSD
+#ifdef HAVE_MDNSRESPONDER
   int		error;			// Error code, if any
   pthread_t	tid;			// Thread ID
 
@@ -139,7 +139,7 @@ _papplDNSSDInit(
   }
 
   pthread_mutex_unlock(&pappl_dns_sd_mutex);
-#endif // HAVE_DNSSD
+#endif // HAVE_MDNSRESPONDER
 
   return (pappl_dns_sd_master);
 }
@@ -166,7 +166,7 @@ _papplDNSSDLock(void)
 const char *				// O - Error message
 _papplDNSSDStrError(int error)		// I - Error code
 {
-#ifdef HAVE_DNSSD
+#ifdef HAVE_MDNSRESPONDER
   switch (error)
   {
     case kDNSServiceErr_NoError :
@@ -265,9 +265,10 @@ _papplDNSSDStrError(int error)		// I - Error code
 
     case kDNSServiceErr_Timeout :
         return ("Timeout");
-
+#if !_WIN32
     case kDNSServiceErr_DefunctConnection :
         return ("Connection lost");
+#endif // !_WIN32
   }
 
 #elif defined(HAVE_AVAHI)
@@ -275,7 +276,7 @@ _papplDNSSDStrError(int error)		// I - Error code
 
 #else
   return ("");
-#endif // HAVE_DNSSD
+#endif // HAVE_MDNSRESPONDER
 }
 
 
@@ -301,12 +302,12 @@ bool					// O - `true` on success, `false` on failure
 _papplPrinterRegisterDNSSDNoLock(
     pappl_printer_t *printer)		// I - Printer
 {
-#if defined(HAVE_DNSSD) || defined(HAVE_AVAHI)
   bool			ret = true;	// Return value
+#ifdef HAVE_DNSSD
   pappl_system_t	*system = printer->system;
 					// System
   _pappl_txt_t		txt;		// DNS-SD TXT record
-  int			i,		// Looping var
+  cups_len_t		i,		// Looping var
 			count;		// Number of values
   ipp_attribute_t	*color_supported,
 			*document_format_supported,
@@ -323,12 +324,12 @@ _papplPrinterRegisterDNSSDNoLock(
   char			product[248];	// Make and model (legacy)
   int			max_width;	// Maximum media width (legacy)
   const char		*papermax;	// PaperMax string value (legacy)
-#  ifdef HAVE_DNSSD
+#  ifdef HAVE_MDNSRESPONDER
   DNSServiceErrorType	error;		// Error from mDNSResponder
 #  else
   int			error;		// Error from Avahi
   char			fullname[256];	// Full service name
-#  endif // HAVE_DNSSD
+#  endif // HAVE_MDNSRESPONDER
   _pappl_dns_sd_t	master;		// DNS-SD master reference
 
 
@@ -354,7 +355,7 @@ _papplPrinterRegisterDNSSDNoLock(
     if (ptr > formats && ptr < (formats + sizeof(formats) - 1))
       *ptr++ = ',';
 
-    strlcpy(ptr, value, sizeof(formats) - (size_t)(ptr - formats));
+    papplCopyString(ptr, value, sizeof(formats) - (size_t)(ptr - formats));
     ptr += strlen(ptr);
 
     if (ptr >= (formats + sizeof(formats) - 1))
@@ -369,7 +370,7 @@ _papplPrinterRegisterDNSSDNoLock(
     if (ptr > kind && ptr < (kind + sizeof(kind) - 1))
       *ptr++ = ',';
 
-    strlcpy(ptr, value, sizeof(kind) - (size_t)(ptr - kind));
+    papplCopyString(ptr, value, sizeof(kind) - (size_t)(ptr - kind));
     ptr += strlen(ptr);
 
     if (ptr >= (kind + sizeof(kind) - 1))
@@ -378,7 +379,7 @@ _papplPrinterRegisterDNSSDNoLock(
 
   snprintf(product, sizeof(product), "(%s)", printer->driver_data.make_and_model);
 
-  for (i = 0, max_width = 0; i < printer->driver_data.num_media; i ++)
+  for (i = 0, max_width = 0; i < (cups_len_t)printer->driver_data.num_media; i ++)
   {
     pwg_media_t *media = pwgMediaForPWG(printer->driver_data.media[i]);
 					// Current media size
@@ -406,7 +407,7 @@ _papplPrinterRegisterDNSSDNoLock(
     if (ptr > urf && ptr < (urf + sizeof(urf) - 1))
       *ptr++ = ',';
 
-    strlcpy(ptr, value, sizeof(urf) - (size_t)(ptr - urf));
+    papplCopyString(ptr, value, sizeof(urf) - (size_t)(ptr - urf));
     ptr += strlen(ptr);
 
     if (ptr >= (urf + sizeof(urf) - 1))
@@ -442,7 +443,7 @@ _papplPrinterRegisterDNSSDNoLock(
     {
       char	base_dns_sd_name[256];	// Base DNS-SD name
 
-      strlcpy(base_dns_sd_name, printer->dns_sd_name, sizeof(base_dns_sd_name));
+      papplCopyString(base_dns_sd_name, printer->dns_sd_name, sizeof(base_dns_sd_name));
       if ((ptr = strrchr(base_dns_sd_name, '(')) != NULL)
         *ptr = '\0';
 
@@ -465,9 +466,9 @@ _papplPrinterRegisterDNSSDNoLock(
 
   if ((master = _papplDNSSDInit(printer->system)) == NULL)
     return (false);
-#endif // HAVE_DNSSD || HAVE_AVAHI
+#endif // HAVE_DNSSD
 
-#ifdef HAVE_DNSSD
+#ifdef HAVE_MDNSRESPONDER
   // Build the TXT record for IPP...
   TXTRecordCreate(&txt, 1024, NULL);
   TXTRecordSetValue(&txt, "rp", (uint8_t)strlen(printer->resource) - 1, printer->resource + 1);
@@ -522,7 +523,7 @@ _papplPrinterRegisterDNSSDNoLock(
   if (system->subtypes && *system->subtypes)
     snprintf(regtype, sizeof(regtype), "_ipp._tcp,%s", system->subtypes);
   else
-    strlcpy(regtype, "_ipp._tcp", sizeof(regtype));
+    papplCopyString(regtype, "_ipp._tcp", sizeof(regtype));
 
   if ((error = DNSServiceRegister(&printer->dns_sd_ipp_ref, kDNSServiceFlagsShareConnection | kDNSServiceFlagsNoAutoRename, 0 /* interfaceIndex */, printer->dns_sd_name, regtype, NULL /* domain */, system->hostname, htons(system->port), TXTRecordGetLength(&txt), TXTRecordGetBytesPtr(&txt), (DNSServiceRegisterReply)dns_sd_printer_callback, printer)) != kDNSServiceErr_NoError)
   {
@@ -542,27 +543,32 @@ _papplPrinterRegisterDNSSDNoLock(
   if (printer->dns_sd_ipps_ref)
     DNSServiceRefDeallocate(printer->dns_sd_ipps_ref);
 
-  printer->dns_sd_ipps_ref = master;
-
-  if (system->subtypes && *system->subtypes)
-    snprintf(regtype, sizeof(regtype), "_ipps._tcp,%s", system->subtypes);
-  else
-    strlcpy(regtype, "_ipps._tcp", sizeof(regtype));
-
-  if ((error = DNSServiceRegister(&printer->dns_sd_ipps_ref, kDNSServiceFlagsShareConnection | kDNSServiceFlagsNoAutoRename, 0 /* interfaceIndex */, printer->dns_sd_name, regtype, NULL /* domain */, system->hostname, htons(system->port), TXTRecordGetLength(&txt), TXTRecordGetBytesPtr(&txt), (DNSServiceRegisterReply)dns_sd_printer_callback, printer)) != kDNSServiceErr_NoError)
+  if (!(printer->system->options & PAPPL_SOPTIONS_NO_TLS))
   {
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s.%s': %s", printer->dns_sd_name, regtype, _papplDNSSDStrError(error));
-    ret = false;
-  }
+    printer->dns_sd_ipps_ref = master;
 
-  if (printer->geo_location && ret)
-  {
-    if ((error = DNSServiceAddRecord(printer->dns_sd_ipps_ref, &printer->dns_sd_ipps_loc_ref, 0, kDNSServiceType_LOC, sizeof(printer->dns_sd_loc), printer->dns_sd_loc, 0)) != kDNSServiceErr_NoError)
+    if (system->subtypes && *system->subtypes)
+      snprintf(regtype, sizeof(regtype), "_ipps._tcp,%s", system->subtypes);
+    else
+      papplCopyString(regtype, "_ipps._tcp", sizeof(regtype));
+
+    if ((error = DNSServiceRegister(&printer->dns_sd_ipps_ref, kDNSServiceFlagsShareConnection | kDNSServiceFlagsNoAutoRename, 0 /* interfaceIndex */, printer->dns_sd_name, regtype, NULL /* domain */, system->hostname, htons(system->port), TXTRecordGetLength(&txt), TXTRecordGetBytesPtr(&txt), (DNSServiceRegisterReply)dns_sd_printer_callback, printer)) != kDNSServiceErr_NoError)
     {
-      papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register LOC record for '%s.%s': %s", printer->dns_sd_name, regtype, _papplDNSSDStrError(error));
+      papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s.%s': %s", printer->dns_sd_name, regtype, _papplDNSSDStrError(error));
       ret = false;
     }
+
+    if (printer->geo_location && ret)
+    {
+      if ((error = DNSServiceAddRecord(printer->dns_sd_ipps_ref, &printer->dns_sd_ipps_loc_ref, 0, kDNSServiceType_LOC, sizeof(printer->dns_sd_loc), printer->dns_sd_loc, 0)) != kDNSServiceErr_NoError)
+      {
+	papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register LOC record for '%s.%s': %s", printer->dns_sd_name, regtype, _papplDNSSDStrError(error));
+	ret = false;
+      }
+    }
   }
+  else
+    printer->dns_sd_ipps_ref = NULL;
 
   TXTRecordDeallocate(&txt);
 
@@ -705,33 +711,36 @@ _papplPrinterRegisterDNSSDNoLock(
     free(temptypes);
   }
 
-  if ((error = avahi_entry_group_add_service_strlst(printer->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, printer->dns_sd_name, "_ipps._tcp", NULL, system->hostname, system->port, txt)) < 0)
+  if (!(printer->system->options & PAPPL_SOPTIONS_NO_TLS))
   {
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s._ipps._tcp': %s", printer->dns_sd_name, _papplDNSSDStrError(error));
-    ret = false;
-  }
-
-  if (system->subtypes && *system->subtypes)
-  {
-    char *temptypes = strdup(system->subtypes), *start, *end;
-					// Pointers into sub-types...
-
-    for (start = temptypes; start && *start; start = end)
+    if ((error = avahi_entry_group_add_service_strlst(printer->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, printer->dns_sd_name, "_ipps._tcp", NULL, system->hostname, system->port, txt)) < 0)
     {
-      if ((end = strchr(start, ',')) != NULL)
-	*end++ = '\0';
-      else
-	end = start + strlen(start);
-
-      snprintf(regtype, sizeof(regtype), "%s._sub._ipps._tcp", start);
-      if ((error = avahi_entry_group_add_service_subtype(printer->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, printer->dns_sd_name, "_ipps._tcp", NULL, regtype)) < 0)
-      {
-        papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s.%s': %s", printer->dns_sd_name, regtype, _papplDNSSDStrError(error));
-        ret = false;
-      }
+      papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s._ipps._tcp': %s", printer->dns_sd_name, _papplDNSSDStrError(error));
+      ret = false;
     }
 
-    free(temptypes);
+    if (system->subtypes && *system->subtypes)
+    {
+      char *temptypes = strdup(system->subtypes), *start, *end;
+					  // Pointers into sub-types...
+
+      for (start = temptypes; start && *start; start = end)
+      {
+	if ((end = strchr(start, ',')) != NULL)
+	  *end++ = '\0';
+	else
+	  end = start + strlen(start);
+
+	snprintf(regtype, sizeof(regtype), "%s._sub._ipps._tcp", start);
+	if ((error = avahi_entry_group_add_service_subtype(printer->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, printer->dns_sd_name, "_ipps._tcp", NULL, regtype)) < 0)
+	{
+	  papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s.%s': %s", printer->dns_sd_name, regtype, _papplDNSSDStrError(error));
+	  ret = false;
+	}
+      }
+
+      free(temptypes);
+    }
   }
 
   avahi_string_list_free(txt);
@@ -800,7 +809,7 @@ _papplPrinterRegisterDNSSDNoLock(
   // Commit it...
   avahi_entry_group_commit(printer->dns_sd_ref);
   _papplDNSSDUnlock();
-#endif // HAVE_DNSSD
+#endif // HAVE_MDNSRESPONDER
 
   return (ret);
 }
@@ -814,7 +823,7 @@ void
 _papplPrinterUnregisterDNSSDNoLock(
     pappl_printer_t *printer)		// I - Printer
 {
-#if HAVE_DNSSD
+#if HAVE_MDNSRESPONDER
   if (printer->dns_sd_printer_ref)
   {
     DNSServiceRefDeallocate(printer->dns_sd_printer_ref);
@@ -851,7 +860,7 @@ _papplPrinterUnregisterDNSSDNoLock(
 
 #else
   (void)printer;
-#endif /* HAVE_DNSSD */
+#endif // HAVE_MDNSRESPONDER
 }
 
 
@@ -863,16 +872,16 @@ bool					// O - `true` on success, `false` on failure
 _papplSystemRegisterDNSSDNoLock(
     pappl_system_t *system)		// I - System
 {
-#if defined(HAVE_DNSSD) || defined(HAVE_AVAHI)
   bool			ret = true;	// Return value
+#ifdef HAVE_DNSSD
   _pappl_dns_sd_t	master;		// DNS-SD master reference
   _pappl_txt_t		txt;		// DNS-SD TXT record
-#  ifdef HAVE_DNSSD
+#  ifdef HAVE_MDNSRESPONDER
   DNSServiceErrorType	error;		// Error from mDNSResponder
 #  else
   int			error;		// Error from Avahi
   char			fullname[256];	// Full name of services
-#  endif // HAVE_DNSSD
+#  endif // HAVE_MDNSRESPONDER
 
 
   // Make sure we have all of the necessary information to register the system...
@@ -903,7 +912,7 @@ _papplSystemRegisterDNSSDNoLock(
       char	base_dns_sd_name[256],	// Base DNS-SD name
 		*ptr;			// Pointer into name
 
-      strlcpy(base_dns_sd_name, system->dns_sd_name, sizeof(base_dns_sd_name));
+      papplCopyString(base_dns_sd_name, system->dns_sd_name, sizeof(base_dns_sd_name));
       if ((ptr = strrchr(base_dns_sd_name, '(')) != NULL)
         *ptr = '\0';
 
@@ -927,9 +936,9 @@ _papplSystemRegisterDNSSDNoLock(
 
   if ((master = _papplDNSSDInit(system)) == NULL)
     return (false);
-#endif // HAVE_DNSSD || HAVE_AVAHI
+#endif // HAVE_DNSSD
 
-#ifdef HAVE_DNSSD
+#ifdef HAVE_MDNSRESPONDER
   // Build the TXT record...
   TXTRecordCreate(&txt, 1024, NULL);
   if (system->location != NULL)
@@ -940,24 +949,29 @@ _papplSystemRegisterDNSSDNoLock(
   if (system->dns_sd_ipps_ref)
     DNSServiceRefDeallocate(system->dns_sd_ipps_ref);
 
-  system->dns_sd_ipps_ref = master;
-
-  if ((error = DNSServiceRegister(&system->dns_sd_ipps_ref, kDNSServiceFlagsShareConnection | kDNSServiceFlagsNoAutoRename, 0 /* interfaceIndex */, system->dns_sd_name, "_ipps-system._tcp", NULL /* domain */, system->hostname, htons(system->port), TXTRecordGetLength(&txt), TXTRecordGetBytesPtr(&txt), (DNSServiceRegisterReply)dns_sd_system_callback, system)) != kDNSServiceErr_NoError)
+  if (!(system->options & PAPPL_SOPTIONS_NO_TLS))
   {
-    papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s._ipps-system._tcp': %s", system->dns_sd_name, _papplDNSSDStrError(error));
-    ret = false;
-  }
+    system->dns_sd_ipps_ref = master;
 
-  if (system->geo_location && ret)
-  {
-    papplLog(system, PAPPL_LOGLEVEL_DEBUG, "Registering LOC record for '%s._ipps-system._tcp' with data %02X %02X %02X %02X %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X", system->dns_sd_name, system->dns_sd_loc[0], system->dns_sd_loc[1], system->dns_sd_loc[2], system->dns_sd_loc[3], system->dns_sd_loc[4], system->dns_sd_loc[5], system->dns_sd_loc[6], system->dns_sd_loc[7], system->dns_sd_loc[8], system->dns_sd_loc[9], system->dns_sd_loc[10], system->dns_sd_loc[11], system->dns_sd_loc[12], system->dns_sd_loc[13], system->dns_sd_loc[14], system->dns_sd_loc[15]);
-
-    if ((error = DNSServiceAddRecord(system->dns_sd_ipps_ref, &system->dns_sd_loc_ref, 0, kDNSServiceType_LOC, sizeof(system->dns_sd_loc), system->dns_sd_loc, 0)) != kDNSServiceErr_NoError)
+    if ((error = DNSServiceRegister(&system->dns_sd_ipps_ref, kDNSServiceFlagsShareConnection | kDNSServiceFlagsNoAutoRename, 0 /* interfaceIndex */, system->dns_sd_name, "_ipps-system._tcp", NULL /* domain */, system->hostname, htons(system->port), TXTRecordGetLength(&txt), TXTRecordGetBytesPtr(&txt), (DNSServiceRegisterReply)dns_sd_system_callback, system)) != kDNSServiceErr_NoError)
     {
-      papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register LOC record for '%s._ipps-system._tcp': %s", system->dns_sd_name, _papplDNSSDStrError(error));
+      papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s._ipps-system._tcp': %s", system->dns_sd_name, _papplDNSSDStrError(error));
       ret = false;
     }
+
+    if (system->geo_location && ret)
+    {
+      papplLog(system, PAPPL_LOGLEVEL_DEBUG, "Registering LOC record for '%s._ipps-system._tcp' with data %02X %02X %02X %02X %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X", system->dns_sd_name, system->dns_sd_loc[0], system->dns_sd_loc[1], system->dns_sd_loc[2], system->dns_sd_loc[3], system->dns_sd_loc[4], system->dns_sd_loc[5], system->dns_sd_loc[6], system->dns_sd_loc[7], system->dns_sd_loc[8], system->dns_sd_loc[9], system->dns_sd_loc[10], system->dns_sd_loc[11], system->dns_sd_loc[12], system->dns_sd_loc[13], system->dns_sd_loc[14], system->dns_sd_loc[15]);
+
+      if ((error = DNSServiceAddRecord(system->dns_sd_ipps_ref, &system->dns_sd_loc_ref, 0, kDNSServiceType_LOC, sizeof(system->dns_sd_loc), system->dns_sd_loc, 0)) != kDNSServiceErr_NoError)
+      {
+	papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register LOC record for '%s._ipps-system._tcp': %s", system->dns_sd_name, _papplDNSSDStrError(error));
+	ret = false;
+      }
+    }
   }
+  else
+    system->dns_sd_ipps_ref = NULL;
 
   TXTRecordDeallocate(&txt);
 
@@ -998,24 +1012,27 @@ _papplSystemRegisterDNSSDNoLock(
     return (false);
   }
 
-  if ((error = avahi_entry_group_add_service_strlst(system->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, system->dns_sd_name, "_ipps-system._tcp", NULL, system->hostname, system->port, txt)) < 0)
+  if (!(system->options & PAPPL_SOPTIONS_NO_TLS))
   {
-    papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s._ipps-system._tcp': %s", system->dns_sd_name, _papplDNSSDStrError(error));
-    ret = false;
-  }
-
-  // Register the geolocation of the service...
-  if (system->geo_location && ret)
-  {
-    snprintf(fullname, sizeof(fullname), "%s._ipps-system._tcp.local.", system->dns_sd_name);
-
-    papplLog(system, PAPPL_LOGLEVEL_DEBUG, "Registering LOC record for '%s' with data %02X %02X %02X %02X %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X", fullname, system->dns_sd_loc[0], system->dns_sd_loc[1], system->dns_sd_loc[2], system->dns_sd_loc[3], system->dns_sd_loc[4], system->dns_sd_loc[5], system->dns_sd_loc[6], system->dns_sd_loc[7], system->dns_sd_loc[8], system->dns_sd_loc[9], system->dns_sd_loc[10], system->dns_sd_loc[11], system->dns_sd_loc[12], system->dns_sd_loc[13], system->dns_sd_loc[14], system->dns_sd_loc[15]);
-
-    if ((error = avahi_entry_group_add_record(system->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, fullname, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_LOC, 75 * 60, system->dns_sd_loc, sizeof(system->dns_sd_loc))) < 0)
+    if ((error = avahi_entry_group_add_service_strlst(system->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, system->dns_sd_name, "_ipps-system._tcp", NULL, system->hostname, system->port, txt)) < 0)
     {
-      papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register LOC record for '%s': %s", fullname, _papplDNSSDStrError(error));
+      papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register '%s._ipps-system._tcp': %s", system->dns_sd_name, _papplDNSSDStrError(error));
       ret = false;
     }
+
+    // Register the geolocation of the service...
+    if (system->geo_location && ret)
+    {
+      snprintf(fullname, sizeof(fullname), "%s._ipps-system._tcp.local.", system->dns_sd_name);
+
+      papplLog(system, PAPPL_LOGLEVEL_DEBUG, "Registering LOC record for '%s' with data %02X %02X %02X %02X %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X", fullname, system->dns_sd_loc[0], system->dns_sd_loc[1], system->dns_sd_loc[2], system->dns_sd_loc[3], system->dns_sd_loc[4], system->dns_sd_loc[5], system->dns_sd_loc[6], system->dns_sd_loc[7], system->dns_sd_loc[8], system->dns_sd_loc[9], system->dns_sd_loc[10], system->dns_sd_loc[11], system->dns_sd_loc[12], system->dns_sd_loc[13], system->dns_sd_loc[14], system->dns_sd_loc[15]);
+
+      if ((error = avahi_entry_group_add_record(system->dns_sd_ref, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, fullname, AVAHI_DNS_CLASS_IN, AVAHI_DNS_TYPE_LOC, 75 * 60, system->dns_sd_loc, sizeof(system->dns_sd_loc))) < 0)
+      {
+	papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to register LOC record for '%s': %s", fullname, _papplDNSSDStrError(error));
+	ret = false;
+      }
+    }
   }
 
   // Finally _http.tcp (HTTP) for the web interface...
@@ -1030,7 +1047,7 @@ _papplSystemRegisterDNSSDNoLock(
   _papplDNSSDUnlock();
 
   avahi_string_list_free(txt);
-#endif // HAVE_DNSSD
+#endif // HAVE_MDNSRESPONDER
 
   return (ret);
 }
@@ -1044,7 +1061,7 @@ void
 _papplSystemUnregisterDNSSDNoLock(
     pappl_system_t *system)		// I - System
 {
-#if HAVE_DNSSD
+#if HAVE_MDNSRESPONDER
   if (system->dns_sd_ipps_ref)
   {
     DNSServiceRefDeallocate(system->dns_sd_ipps_ref);
@@ -1069,8 +1086,8 @@ _papplSystemUnregisterDNSSDNoLock(
   _papplDNSSDUnlock();
 
 #else
-  (void)printer;
-#endif /* HAVE_DNSSD */
+  (void)system;
+#endif // HAVE_MDNSRESPONDER
 }
 
 
@@ -1117,7 +1134,7 @@ dns_sd_geo_to_loc(const char    *geo,	// I - "geo:" URI
 }
 
 
-#ifdef HAVE_DNSSD
+#ifdef HAVE_MDNSRESPONDER
 //
 // 'dns_sd_printer_callback()' - Handle DNS-SD printer registration events.
 //
@@ -1160,14 +1177,34 @@ dns_sd_run(void *data)			// I - System object
   int		err;			// Status
   pappl_system_t *system = (pappl_system_t *)data;
 					// System object
+  struct pollfd	pfd;			// Poll data
+
+
+  pfd.events = POLLIN | POLLERR;
+  pfd.fd     = DNSServiceRefSockFD(pappl_dns_sd_master);
 
   for (;;)
   {
-    if ((err = DNSServiceProcessResult(pappl_dns_sd_master)) != kDNSServiceErr_NoError)
+#if _WIN32
+    if (poll(&pfd, 1, 1000) < 0 && WSAGetLastError() == WSAEINTR)
+#else
+    if (poll(&pfd, 1, 1000) < 0 && errno != EINTR && errno != EAGAIN)
+#endif // _WIN32
     {
-      papplLog(system, PAPPL_LOGLEVEL_ERROR, "DNSServiceProcessResult returned %d (%s).", err, _papplDNSSDStrError(err));
+      papplLog(system, PAPPL_LOGLEVEL_ERROR, "DNS-SD poll failed: %s", strerror(errno));
       break;
     }
+
+    if (pfd.revents & POLLIN)
+    {
+      if ((err = DNSServiceProcessResult(pappl_dns_sd_master)) != kDNSServiceErr_NoError)
+      {
+	papplLog(system, PAPPL_LOGLEVEL_ERROR, "DNSServiceProcessResult returned %d (%s).", err, _papplDNSSDStrError(err));
+	break;
+      }
+    }
+    else if (pfd.revents)
+      break;
   }
 
   return (NULL);
@@ -1275,4 +1312,4 @@ dns_sd_system_callback(
     system->dns_sd_any_collision = true;
   }
 }
-#endif // HAVE_DNSSD
+#endif // HAVE_MDNSRESPONDER
