@@ -86,8 +86,12 @@ papplJobFilterImage(
 			byte,		// Byte in line
 			bit;		// Current bit
   const unsigned char	*pixbase,	// Pointer to first pixel
-			*pixptr;	// Pointer into image
-  int			img_width,	// Rotated image width
+			*pixline,	// Pointer to start of current line
+			*pixptr,	// Pointer into image
+			*pixend;	// End of image
+  int			pixel0,		// Temporary pixel value
+			pixel1,		// ...
+			img_width,	// Rotated image width
 			img_height,	// Rotated image height
 			x,		// X position
 			xsize,		// Scaled width
@@ -101,12 +105,12 @@ papplJobFilterImage(
 			xerr,		// X error accumulator
 			xmod,		// X modulus
 			xstep,		// X step
+			yerr,		// Y error accumulator
+			ymod,		// Y modulus
+			ystep,		// Y step
 			ydir;		// Y direction
 
 
-  // TODO: Implement interpolation (Issue #64)
-  (void)smoothing;
-
   // Images contain a single page/impression...
   papplJobSetImpressions(job, 1);
 
@@ -313,6 +317,9 @@ papplJobFilterImage(
   xmod   = (int)(img_width % xsize);
   xstep  = (int)(img_width / xsize) * xdir;
 
+  ymod   = (int)(img_height % ysize);
+  ystep  = (int)(img_height / ysize) * ydir;
+
   if (xend > (int)options->header.cupsWidth)
     xend = (int)options->header.cupsWidth;
 
@@ -320,7 +327,7 @@ papplJobFilterImage(
     yend = (int)options->header.cupsHeight;
 
   papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "xsize=%d, xstart=%d, xend=%d, xdir=%d, xmod=%d, xstep=%d", xsize, xstart, xend, xdir, xmod, xstep);
-  papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "ysize=%d, ystart=%d, yend=%d, ydir=%d", ysize, ystart, yend, ydir);
+  papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "ysize=%d, ystart=%d, yend=%d, ydir=%d, ymod=%d, ystep=%d", ysize, ystart, yend, ydir, ymod, ystep);
 
   papplPrinterGetDriverData(papplJobGetPrinter(job), &driver_data);
 
@@ -344,6 +351,8 @@ papplJobFilterImage(
   else
     white = 0xff;
 
+  pixend = pixels + width * height * depth;
+
   // Print every copy...
   for (i = 0; i < options->copies; i ++)
   {
@@ -364,10 +373,21 @@ papplJobFilterImage(
       }
     }
 
+    if (ystart < 0)
+    {
+      pixline = pixbase - (ystart * ymod / ysize) * ydir;
+      yerr    = -ymod / 2 - (ystart * ymod) % ysize;
+    }
+    else
+    {
+      pixline = pixbase;
+      yerr    = -ymod / 2;
+    }
+
     // Now RIP the image...
     for (; y < yend && !job->is_canceled; y ++)
     {
-      pixptr = pixbase + ydir * (int)((y - ystart) * (img_height - 1) / (ysize - 1));
+      pixptr = pixline;
 
       if (xstart < 0)
       {
@@ -423,7 +443,28 @@ papplJobFilterImage(
 	for (lineptr = line + x; x < xend; x ++)
 	{
 	  // Copy an inverted grayscale pixel...
-	  *lineptr++ = ~*pixptr;
+	  if (smoothing && yerr >= 0 && xerr >= 0)
+	  {
+	    const unsigned char	*rt = pixptr + xdir,
+				*dn = pixptr + ydir,
+				*dnrt = pixptr + xdir + ydir;
+					// Pointers to adjacent pixels
+
+	    if (rt < pixels || rt >= pixend)
+	      rt = pixptr;
+	    if (dn < pixels || dn >= pixend)
+	      dn = pixptr;
+	    if (dnrt < pixels || dnrt >= pixend)
+	      dnrt = pixptr;
+
+	    pixel0     = ((xsize - xerr) * *pixptr + xerr * *rt) / xsize;
+            pixel1     = ((xsize - xerr) * *dn + xerr * *dnrt) / xsize;
+	    *lineptr++ = (unsigned char)(255 - ((ysize - yerr) * pixel0 + yerr * pixel1) / ysize);
+	  }
+	  else
+	  {
+	    *lineptr++ = ~*pixptr;
+	  }
 
 	  // Advance to the next pixel...
 	  pixptr += xstep;
@@ -444,8 +485,33 @@ papplJobFilterImage(
 	for (lineptr = line + x * bpp; x < xend; x ++)
 	{
 	  // Copy a grayscale or RGB pixel...
-	  memcpy(lineptr, pixptr, (unsigned)bpp);
-	  lineptr += bpp;
+	  if (smoothing && yerr >= 0 && xerr >= 0)
+	  {
+	    int			j;	// Looping var
+	    const unsigned char	*rt = pixptr + xdir,
+				*dn = pixptr + ydir,
+				*dnrt = pixptr + xdir + ydir;
+					// Pointers to adjacent pixels
+
+	    if (rt < pixels || rt >= pixend)
+	      rt = pixptr;
+	    if (dn < pixels || dn >= pixend)
+	      dn = pixptr;
+	    if (dnrt < pixels || dnrt >= pixend)
+	      dnrt = pixptr;
+
+            for (j = 0; j < bpp; j ++)
+            {
+	      pixel0     = ((xsize - xerr) * pixptr[j] + xerr * rt[j]) / xsize;
+              pixel1     = ((xsize - xerr) * dn[j] + xerr * dnrt[j]) / xsize;
+	      *lineptr++ = (unsigned char)(((ysize - yerr) * pixel0 + yerr * pixel1) / ysize);
+	    }
+	  }
+	  else
+	  {
+	    memcpy(lineptr, pixptr, (unsigned)bpp);
+	    lineptr += bpp;
+	  }
 
 	  // Advance to the next pixel...
 	  pixptr += xstep;
@@ -464,6 +530,14 @@ papplJobFilterImage(
 	papplLogJob(job, PAPPL_LOGLEVEL_ERROR, "Unable to write raster line %u.", y);
 	goto abort_job;
       }
+
+      pixline += ystep;
+      yerr += ymod;
+      if (yerr >= ysize)
+      {
+        pixline += ydir;
+        yerr -= ysize;
+      }
     }
 
     // Trailing blank space...
@@ -622,11 +696,12 @@ _papplJobFilterJPEG(
 
   ret = papplJobFilterImage(job, device, options, pixels, (int)dinfo.output_width, (int)dinfo.output_height, dinfo.output_components, ppi, true);
 
+  jpeg_finish_decompress(&dinfo);
+
   finish_jpeg:
 
   papplJobDeletePrintOptions(options);
   free(pixels);
-  jpeg_finish_decompress(&dinfo);
   jpeg_destroy_decompress(&dinfo);
   fclose(fp);
 
