@@ -1,7 +1,7 @@
 //
 // Job IPP processing for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -43,8 +43,8 @@ static void		ipp_send_document(pappl_client_t *client);
 
 void
 _papplJobCopyAttributes(
-    pappl_client_t *client,		// I - Client
     pappl_job_t    *job,		// I - Job
+    pappl_client_t *client,		// I - Client
     cups_array_t   *ra)			// I - requested-attributes
 {
   _papplCopyAttributes(client->response, job->attrs, ra, IPP_TAG_JOB, 0);
@@ -77,114 +77,7 @@ _papplJobCopyAttributes(
   if (!ra || cupsArrayFind(ra, "job-printer-up-time"))
     ippAddInteger(client->response, IPP_TAG_JOB, IPP_TAG_INTEGER, "job-printer-up-time", (int)(time(NULL) - client->printer->start_time));
 
-  if (!ra || cupsArrayFind(ra, "job-state"))
-    ippAddInteger(client->response, IPP_TAG_JOB, IPP_TAG_ENUM, "job-state", (int)job->state);
-
-  if (!ra || cupsArrayFind(ra, "job-state-message"))
-  {
-    if (job->message)
-    {
-      ippAddString(client->response, IPP_TAG_JOB, IPP_TAG_TEXT, "job-state-message", NULL, job->message);
-    }
-    else
-    {
-      switch (job->state)
-      {
-	case IPP_JSTATE_PENDING :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job pending.");
-	    break;
-
-	case IPP_JSTATE_HELD :
-	    if (job->fd >= 0)
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job incoming.");
-	    else if (ippFindAttribute(job->attrs, "job-hold-until", IPP_TAG_ZERO))
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job held.");
-	    else
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job created.");
-	    break;
-
-	case IPP_JSTATE_PROCESSING :
-	    if (job->is_canceled)
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job canceling.");
-	    else
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job printing.");
-	    break;
-
-	case IPP_JSTATE_STOPPED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job stopped.");
-	    break;
-
-	case IPP_JSTATE_CANCELED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job canceled.");
-	    break;
-
-	case IPP_JSTATE_ABORTED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job aborted.");
-	    break;
-
-	case IPP_JSTATE_COMPLETED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job completed.");
-	    break;
-      }
-    }
-  }
-
-  if (!ra || cupsArrayFind(ra, "job-state-reasons"))
-  {
-    if (job->state_reasons)
-    {
-      int		num_values = 0;	// Number of string values
-      const char	*svalues[32];	// String values
-      pappl_jreason_t	bit;		// Current reason bit
-
-      for (bit = PAPPL_JREASON_ABORTED_BY_SYSTEM; bit <= PAPPL_JREASON_WARNINGS_DETECTED; bit *= 2)
-      {
-        if (bit & job->state_reasons)
-          svalues[num_values ++] = _papplJobReasonString(bit);
-      }
-
-      ippAddStrings(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", num_values, NULL, svalues);
-    }
-    else
-    {
-      switch (job->state)
-      {
-	case IPP_JSTATE_PENDING :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "none");
-	    break;
-
-	case IPP_JSTATE_HELD :
-	    if (job->fd >= 0)
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-incoming");
-	    else
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-data-insufficient");
-	    break;
-
-	case IPP_JSTATE_PROCESSING :
-	    if (job->is_canceled)
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "processing-to-stop-point");
-	    else
-	      ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-printing");
-	    break;
-
-	case IPP_JSTATE_STOPPED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-stopped");
-	    break;
-
-	case IPP_JSTATE_CANCELED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-canceled-by-user");
-	    break;
-
-	case IPP_JSTATE_ABORTED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "aborted-by-system");
-	    break;
-
-	case IPP_JSTATE_COMPLETED :
-	    ippAddString(client->response, IPP_TAG_JOB, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-completed-successfully");
-	    break;
-      }
-    }
-  }
+  _papplJobCopyState(job, IPP_TAG_JOB, client->response, ra);
 
   if (!ra || cupsArrayFind(ra, "time-at-creation"))
     ippAddInteger(client->response, IPP_TAG_JOB, IPP_TAG_INTEGER, "time-at-creation", (int)(job->created - client->printer->start_time));
@@ -209,7 +102,8 @@ _papplJobCopyDocumentData(
 {
   char			filename[1024],	// Filename buffer
 			buffer[4096];	// Copy buffer
-  ssize_t		bytes;		// Bytes read
+  ssize_t		bytes,		// Bytes read
+			total = 0;	// Total bytes copied
   cups_array_t		*ra;		// Attributes to send in response
 
 
@@ -241,8 +135,10 @@ _papplJobCopyDocumentData(
 
   papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "Created job file \"%s\", format \"%s\".", filename, job->format);
 
-  while ((bytes = httpRead2(client->http, buffer, sizeof(buffer))) > 0)
+  while ((bytes = httpRead(client->http, buffer, sizeof(buffer))) > 0)
   {
+    papplLogClient(client, PAPPL_LOGLEVEL_DEBUG, "Read %d bytes...", (int)bytes);
+
     if (write(job->fd, buffer, (size_t)bytes) < bytes)
     {
       int error = errno;		// Write error
@@ -256,6 +152,8 @@ _papplJobCopyDocumentData(
 
       goto abort_job;
     }
+
+    total += (size_t)bytes;
   }
 
   if (bytes < 0)
@@ -284,6 +182,8 @@ _papplJobCopyDocumentData(
     goto abort_job;
   }
 
+  papplLogJob(job, PAPPL_LOGLEVEL_INFO, "Received %lu bytes of document data.", (unsigned long)total);
+
   job->fd = -1;
 
   // Submit the job for processing...
@@ -294,14 +194,14 @@ _papplJobCopyDocumentData(
   // Return the job info...
   papplClientRespondIPP(client, IPP_STATUS_OK, NULL);
 
-  ra = cupsArrayNew((cups_array_func_t)strcmp, NULL);
+  ra = cupsArrayNew((cups_array_cb_t)strcmp, NULL, NULL, 0, NULL, NULL);
   cupsArrayAdd(ra, "job-id");
   cupsArrayAdd(ra, "job-state");
   cupsArrayAdd(ra, "job-state-message");
   cupsArrayAdd(ra, "job-state-reasons");
   cupsArrayAdd(ra, "job-uri");
 
-  _papplJobCopyAttributes(client, job, ra);
+  _papplJobCopyAttributes(job, client, ra);
   cupsArrayDelete(ra);
   return;
 
@@ -323,17 +223,139 @@ _papplJobCopyDocumentData(
 
   pthread_rwlock_unlock(&client->printer->rwlock);
 
-  ra = cupsArrayNew((cups_array_func_t)strcmp, NULL);
+  ra = cupsArrayNew((cups_array_cb_t)strcmp, NULL, NULL, 0, NULL, NULL);
   cupsArrayAdd(ra, "job-id");
   cupsArrayAdd(ra, "job-state");
   cupsArrayAdd(ra, "job-state-reasons");
   cupsArrayAdd(ra, "job-uri");
 
-  _papplJobCopyAttributes(client, job, ra);
+  _papplJobCopyAttributes(job, client, ra);
   cupsArrayDelete(ra);
 }
 
 
+//
+// '_papplJobCopyState()' - Copy the job-state-xxx sttributes.
+//
+
+void
+_papplJobCopyState(
+    pappl_job_t    *job,	// I - Job
+    ipp_tag_t      group_tag,	// I - Group tag
+    ipp_t          *ipp,	// I - IPP message
+    cups_array_t   *ra)		// I - Requested attributes
+{
+  if (!ra || cupsArrayFind(ra, "job-state"))
+    ippAddInteger(ipp, group_tag, IPP_TAG_ENUM, "job-state", (int)job->state);
+
+  if (!ra || cupsArrayFind(ra, "job-state-message"))
+  {
+    if (job->message)
+    {
+      ippAddString(ipp, group_tag, IPP_TAG_TEXT, "job-state-message", NULL, job->message);
+    }
+    else
+    {
+      switch (job->state)
+      {
+	case IPP_JSTATE_PENDING :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job pending.");
+	    break;
+
+	case IPP_JSTATE_HELD :
+	    if (job->fd >= 0)
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job incoming.");
+	    else if (ippFindAttribute(job->attrs, "job-hold-until", IPP_TAG_ZERO))
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job held.");
+	    else
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job created.");
+	    break;
+
+	case IPP_JSTATE_PROCESSING :
+	    if (job->is_canceled)
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job canceling.");
+	    else
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job printing.");
+	    break;
+
+	case IPP_JSTATE_STOPPED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job stopped.");
+	    break;
+
+	case IPP_JSTATE_CANCELED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job canceled.");
+	    break;
+
+	case IPP_JSTATE_ABORTED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job aborted.");
+	    break;
+
+	case IPP_JSTATE_COMPLETED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_TEXT), "job-state-message", NULL, "Job completed.");
+	    break;
+      }
+    }
+  }
+
+  if (!ra || cupsArrayFind(ra, "job-state-reasons"))
+  {
+    if (job->state_reasons)
+    {
+      size_t		num_values = 0;	// Number of string values
+      const char	*svalues[32];	// String values
+      pappl_jreason_t	bit;		// Current reason bit
+
+      for (bit = PAPPL_JREASON_ABORTED_BY_SYSTEM; bit <= PAPPL_JREASON_WARNINGS_DETECTED; bit *= 2)
+      {
+        if (bit & job->state_reasons)
+          svalues[num_values ++] = _papplJobReasonString(bit);
+      }
+
+      ippAddStrings(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", IPP_NUM_CAST num_values, NULL, svalues);
+    }
+    else
+    {
+      switch (job->state)
+      {
+	case IPP_JSTATE_PENDING :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "none");
+	    break;
+
+	case IPP_JSTATE_HELD :
+	    if (job->fd >= 0)
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-incoming");
+	    else
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-data-insufficient");
+	    break;
+
+	case IPP_JSTATE_PROCESSING :
+	    if (job->is_canceled)
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "processing-to-stop-point");
+	    else
+	      ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-printing");
+	    break;
+
+	case IPP_JSTATE_STOPPED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-stopped");
+	    break;
+
+	case IPP_JSTATE_CANCELED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-canceled-by-user");
+	    break;
+
+	case IPP_JSTATE_ABORTED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "aborted-by-system");
+	    break;
+
+	case IPP_JSTATE_COMPLETED :
+	    ippAddString(ipp, group_tag, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-state-reasons", NULL, "job-completed-successfully");
+	    break;
+      }
+    }
+  }
+}
+
+
 //
 // '_papplJobProcessIPP()' - Process an IPP Job request.
 //
@@ -508,6 +530,10 @@ ipp_cancel_job(pappl_client_t *client)	// I - Client
   pappl_job_t	*job;			// Job information
 
 
+  // Authorize access...
+  if (!_papplPrinterIsAuthorized(client))
+    return;
+
   // Get the job...
   if (ippGetOperation(client->request) == IPP_OP_CANCEL_CURRENT_JOB)
     job = client->printer->processing_job;
@@ -556,6 +582,10 @@ ipp_close_job(pappl_client_t *client)	// I - Client
   pappl_job_t	*job = client->job;	// Job information
 
 
+  // Authorize access...
+  if (!_papplPrinterIsAuthorized(client))
+    return;
+
   // Get the job...
   if (!job)
   {
@@ -603,6 +633,10 @@ ipp_get_job_attributes(
   cups_array_t	*ra;			// requested-attributes
 
 
+  // Authorize access...
+  if (!_papplPrinterIsAuthorized(client))
+    return;
+
   if (!job)
   {
     papplClientRespondIPP(client, IPP_STATUS_ERROR_NOT_FOUND, "Job not found.");
@@ -612,7 +646,7 @@ ipp_get_job_attributes(
   papplClientRespondIPP(client, IPP_STATUS_OK, NULL);
 
   ra = ippCreateRequestedArray(client->request);
-  _papplJobCopyAttributes(client, job, ra);
+  _papplJobCopyAttributes(job, client, ra);
   cupsArrayDelete(ra);
 }
 
@@ -631,6 +665,10 @@ ipp_send_document(
   bool		have_data;		// Do we have document data?
 
 
+  // Authorize access...
+  if (!_papplPrinterIsAuthorized(client))
+    return;
+
   // Get the job...
   if (!job)
   {
