@@ -1,7 +1,7 @@
 //
 // Job processing (printing) functions for the Printer Application Framework
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -21,7 +21,7 @@
 static const char *cups_cspace_string(cups_cspace_t cspace);
 static bool	filter_raw(pappl_job_t *job, pappl_device_t *device);
 static void	finish_job(pappl_job_t *job);
-static void	start_job(pappl_job_t *job);
+static bool	start_job(pappl_job_t *job);
 
 
 //
@@ -43,12 +43,73 @@ papplJobCreatePrintOptions(
     bool             color)		// I - Is the document in color?
 {
   pappl_pr_options_t	*options;	// New options data
-  int			i,		// Looping var
+  cups_len_t		i,		// Looping var
 			count;		// Number of values
   ipp_attribute_t	*attr;		// Attribute
   pappl_printer_t	*printer = job->printer;
 					// Printer
   const char		*raster_type;	// Raster type for output
+  static const char * const media_positions[] =
+  {					// "media-source" to MediaPosition mapping
+    "auto",
+    "main",
+    "alternate",
+    "large-capacity",
+    "manual",
+    "envelope",
+    "disc",
+    "photo",
+    "hagaki",
+    "main-roll",
+    "alternate-roll",
+    "top",
+    "middle",
+    "bottom",
+    "side",
+    "left",
+    "right",
+    "center",
+    "rear",
+    "by-pass-tray",
+    "tray-1",
+    "tray-2",
+    "tray-3",
+    "tray-4",
+    "tray-5",
+    "tray-6",
+    "tray-7",
+    "tray-8",
+    "tray-9",
+    "tray-10",
+    "tray-11",
+    "tray-12",
+    "tray-13",
+    "tray-14",
+    "tray-15",
+    "tray-16",
+    "tray-17",
+    "tray-18",
+    "tray-19",
+    "tray-20",
+    "roll-1",
+    "roll-2",
+    "roll-3",
+    "roll-4",
+    "roll-5",
+    "roll-6",
+    "roll-7",
+    "roll-8",
+    "roll-9",
+    "roll-10"
+  };
+  static const cups_orient_t orientations[] =
+  {					// "orientation-requested" to Orientation  mapping
+    CUPS_ORIENT_0,
+    CUPS_ORIENT_90,
+    CUPS_ORIENT_270,
+    CUPS_ORIENT_180,
+    CUPS_ORIENT_0
+  };
   static const char * const sheet_back[] =
   {					// "pwg-raster-document-sheet-back values
     "normal",
@@ -120,7 +181,7 @@ papplJobCreatePrintOptions(
 
     if (pwg_name && pwg_media)
     {
-      strlcpy(options->media.size_name, pwg_name, sizeof(options->media.size_name));
+      papplCopyString(options->media.size_name, pwg_name, sizeof(options->media.size_name));
       options->media.size_width  = pwg_media->width;
       options->media.size_length = pwg_media->length;
       options->media.source[0]   = '\0';
@@ -129,17 +190,17 @@ papplJobCreatePrintOptions(
 
   if (!options->media.source[0])
   {
-    for (i = 0; i < printer->driver_data.num_source; i ++)
+    for (i = 0; i < (cups_len_t)printer->driver_data.num_source; i ++)
     {
       if (!strcmp(options->media.size_name, printer->driver_data.media_ready[i].size_name))
       {
-        strlcpy(options->media.source, printer->driver_data.source[i], sizeof(options->media.source));
+        papplCopyString(options->media.source, printer->driver_data.source[i], sizeof(options->media.source));
         break;
       }
     }
 
     if (!options->media.source[0])
-      strlcpy(options->media.source, printer->driver_data.media_default.source, sizeof(options->media.source));
+      papplCopyString(options->media.source, printer->driver_data.media_default.source, sizeof(options->media.source));
   }
 
   // orientation-requested
@@ -156,9 +217,9 @@ papplJobCreatePrintOptions(
     const char		*value;		// Attribute string value
 
     if ((value = ippGetString(ippFindAttribute(job->attrs, "output-bin", IPP_TAG_ZERO), 0, NULL)) != NULL)
-      strlcpy(options->output_bin, value, sizeof(options->output_bin));
+      papplCopyString(options->output_bin, value, sizeof(options->output_bin));
     else
-      strlcpy(options->output_bin, printer->driver_data.bin[printer->driver_data.bin_default], sizeof(options->output_bin));
+      papplCopyString(options->output_bin, printer->driver_data.bin[printer->driver_data.bin_default], sizeof(options->output_bin));
   }
 
   // page-ranges
@@ -271,14 +332,14 @@ papplJobCreatePrintOptions(
   else if (options->print_quality == IPP_QUALITY_NORMAL)
   {
     // print-quality=normal
-    i = printer->driver_data.num_resolution / 2;
+    i = (cups_len_t)printer->driver_data.num_resolution / 2;
     options->printer_resolution[0] = printer->driver_data.x_resolution[i];
     options->printer_resolution[1] = printer->driver_data.y_resolution[i];
   }
   else
   {
     // print-quality=high
-    i = printer->driver_data.num_resolution - 1;
+    i = (cups_len_t)printer->driver_data.num_resolution - 1;
     options->printer_resolution[0] = printer->driver_data.x_resolution[i];
     options->printer_resolution[1] = printer->driver_data.y_resolution[i];
   }
@@ -292,7 +353,7 @@ papplJobCreatePrintOptions(
     options->sides = PAPPL_SIDES_ONE_SIDED;
 
   // Vendor options...
-  for (i = 0; i < printer->driver_data.num_vendor; i ++)
+  for (i = 0; i < (cups_len_t)printer->driver_data.num_vendor; i ++)
   {
     const char *name = printer->driver_data.vendor[i];
 					// Vendor attribute name
@@ -310,7 +371,7 @@ papplJobCreatePrintOptions(
       char	value[1024];		// Value of attribute
 
       ippAttributeString(attr, value, sizeof(value));
-      options->num_vendor = cupsAddOption(name, value, options->num_vendor, &options->vendor);
+      options->num_vendor = (int)cupsAddOption(name, value, (cups_len_t)options->num_vendor, &options->vendor);
     }
   }
 
@@ -346,9 +407,32 @@ papplJobCreatePrintOptions(
       raster_type = "black_8";
   }
 
-  cupsRasterInitPWGHeader(&options->header, pwgMediaForPWG(options->media.size_name), raster_type, options->printer_resolution[0], options->printer_resolution[1], _papplSidesString(options->sides), sheet_back[printer->driver_data.duplex]);
+  if (options->print_quality == IPP_QUALITY_HIGH)
+    memcpy(options->dither, printer->driver_data.pdither, sizeof(options->dither));
+  else
+    memcpy(options->dither, printer->driver_data.gdither, sizeof(options->dither));
 
+  // Generate the raster header...
+  cupsRasterInitPWGHeader(&options->header, pwgMediaForPWG(options->media.size_name), raster_type, options->printer_resolution[0], options->printer_resolution[1], _papplSidesString(options->sides), sheet_back[printer->driver_data.duplex]);
+  for (i = 0; i < (int)(sizeof(media_positions) / sizeof(media_positions[0])); i ++)
+  {
+    if (!strcmp(media_positions[i], options->media.source))
+    {
+      options->header.MediaPosition = (unsigned)i;
+      break;
+    }
+  }
+  papplCopyString(options->header.MediaType, options->media.type, sizeof(options->header.MediaType));
+  papplCopyString(options->header.OutputType, _papplContentString(options->print_content_optimize), sizeof(options->header.OutputType));
+  if (options->finishings & PAPPL_FINISHINGS_TRIM)
+    options->header.CutMedia = CUPS_CUT_PAGE;
+  options->header.Orientation = orientations[options->orientation_requested - IPP_ORIENT_PORTRAIT];
   options->header.cupsInteger[CUPS_RASTER_PWG_TotalPageCount] = (unsigned)options->copies * options->num_pages;
+  options->header.cupsInteger[CUPS_RASTER_PWG_ImageBoxLeft]   = (unsigned)options->media.left_margin * options->header.HWResolution[0] / 2540;
+  options->header.cupsInteger[CUPS_RASTER_PWG_ImageBoxTop]    = options->header.cupsHeight - (unsigned)options->media.top_margin * options->header.HWResolution[1] / 2540;
+  options->header.cupsInteger[CUPS_RASTER_PWG_ImageBoxRight]  = options->header.cupsWidth - (unsigned)options->media.right_margin * options->header.HWResolution[0] / 2540;
+  options->header.cupsInteger[CUPS_RASTER_PWG_ImageBoxBottom] = (unsigned)options->media.bottom_margin * options->header.HWResolution[1] / 2540;
+  options->header.cupsInteger[CUPS_RASTER_PWG_PrintQuality]   = options->print_quality;
 
   // Log options...
   papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "header.cupsWidth=%u", options->header.cupsWidth);
@@ -390,7 +474,7 @@ papplJobCreatePrintOptions(
   papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "print-speed=%d", options->print_speed);
   papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "printer-resolution=%dx%ddpi", options->printer_resolution[0], options->printer_resolution[1]);
 
-  for (i = 0; i < options->num_vendor; i ++)
+  for (i = 0; i < (cups_len_t)options->num_vendor; i ++)
     papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "%s=%s", options->vendor[i].name, options->vendor[i].value);
 
   pthread_rwlock_unlock(&printer->rwlock);
@@ -411,7 +495,7 @@ papplJobDeletePrintOptions(
 {
   if (options)
   {
-    cupsFreeOptions(options->num_vendor, options->vendor);
+    cupsFreeOptions((cups_len_t)options->num_vendor, options->vendor);
     free(options);
   }
 }
@@ -428,27 +512,28 @@ _papplJobProcess(pappl_job_t *job)	// I - Job
 
 
   // Start processing the job...
-  start_job(job);
+  if (start_job(job))
+  {
+    // Do file-specific conversions...
+    if ((filter = _papplSystemFindMIMEFilter(job->system, job->format, job->printer->driver_data.format)) == NULL)
+      filter =_papplSystemFindMIMEFilter(job->system, job->format, "image/pwg-raster");
 
-  // Do file-specific conversions...
-  if ((filter = _papplSystemFindMIMEFilter(job->system, job->format, job->printer->driver_data.format)) == NULL)
-    filter =_papplSystemFindMIMEFilter(job->system, job->format, "image/pwg-raster");
-
-  if (filter)
-  {
-    if (!(filter->cb)(job, job->printer->device, filter->cbdata))
+    if (filter)
+    {
+      if (!(filter->cb)(job, job->printer->device, filter->cbdata))
+	job->state = IPP_JSTATE_ABORTED;
+    }
+    else if (!strcmp(job->format, job->printer->driver_data.format))
+    {
+      if (!filter_raw(job, job->printer->device))
+	job->state = IPP_JSTATE_ABORTED;
+    }
+    else
+    {
+      // Abort a job we can't process...
+      papplLogJob(job, PAPPL_LOGLEVEL_ERROR, "Unable to process job with format '%s'.", job->format);
       job->state = IPP_JSTATE_ABORTED;
-  }
-  else if (!strcmp(job->format, job->printer->driver_data.format))
-  {
-    if (!filter_raw(job, job->printer->device))
-      job->state = IPP_JSTATE_ABORTED;
-  }
-  else
-  {
-    // Abort a job we can't process...
-    papplLogJob(job, PAPPL_LOGLEVEL_ERROR, "Unable to process job with format '%s'.", job->format);
-    job->state = IPP_JSTATE_ABORTED;
+    }
   }
 
   // Move the job to a completed state...
@@ -471,7 +556,7 @@ _papplJobProcessRaster(
 					// Printer for job
   pappl_pr_options_t	*options = NULL;// Job options
   cups_raster_t		*ras = NULL;	// Raster stream
-  cups_page_header2_t	header;		// Page header
+  cups_page_header_t	header;		// Page header
   unsigned		header_pages;	// Number of pages from page header
   const unsigned char	*dither;	// Dither line
   unsigned char		*pixels,	// Incoming pixel line
@@ -488,10 +573,11 @@ _papplJobProcessRaster(
   // Start processing the job...
   job->streaming = true;
 
-  start_job(job);
+  if (!start_job(job))
+    goto complete_job;
 
   // Open the raster stream...
-  if ((ras = cupsRasterOpenIO((cups_raster_iocb_t)httpRead2, client->http, CUPS_RASTER_READ)) == NULL)
+  if ((ras = cupsRasterOpenIO((cups_raster_cb_t)httpRead, client->http, CUPS_RASTER_READ)) == NULL)
   {
     papplLogJob(job, PAPPL_LOGLEVEL_ERROR, "Unable to open raster stream from client - %s", cupsLastErrorString());
     job->state = IPP_JSTATE_ABORTED;
@@ -499,7 +585,7 @@ _papplJobProcessRaster(
   }
 
   // Prepare options...
-  if (!cupsRasterReadHeader2(ras, &header))
+  if (!cupsRasterReadHeader(ras, &header))
   {
     papplLogJob(job, PAPPL_LOGLEVEL_ERROR, "Unable to read raster stream from client - %s", cupsLastErrorString());
     job->state = IPP_JSTATE_ABORTED;
@@ -528,6 +614,8 @@ _papplJobProcessRaster(
 
     papplLogJob(job, PAPPL_LOGLEVEL_INFO, "Page %u raster data is %ux%ux%u (%s)", page, header.cupsWidth, header.cupsHeight, header.cupsBitsPerPixel, cups_cspace_string(header.cupsColorSpace));
 
+    papplSystemAddEvent(printer->system, printer, job, PAPPL_EVENT_JOB_PROGRESS, NULL);
+
     // Set options for this page...
     papplJobDeletePrintOptions(options);
     options = papplJobCreatePrintOptions(job, (unsigned)job->impressions, header.cupsBitsPerPixel > 8);
@@ -709,7 +797,7 @@ _papplJobProcessRaster(
       break;
     }
   }
-  while (cupsRasterReadHeader2(ras, &header));
+  while (cupsRasterReadHeader(ras, &header));
 
   if (!(printer->driver_data.rendjob_cb)(job, options, job->printer->device))
     job->state = IPP_JSTATE_ABORTED;
@@ -725,7 +813,7 @@ _papplJobProcessRaster(
     // Flush excess data...
     char	buffer[8192];		// Read buffer
 
-    while (httpRead2(client->http, buffer, sizeof(buffer)) > 0)
+    while (httpRead(client->http, buffer, sizeof(buffer)) > 0)
       ;				// Read all document data
   }
 
@@ -831,7 +919,7 @@ filter_raw(pappl_job_t    *job,		// I - Job
 
 
   papplJobSetImpressions(job, 1);
-  options = papplJobCreatePrintOptions(job, 1, false);
+  options = papplJobCreatePrintOptions(job, 0, job->printer->driver_data.ppm_color > 0);
 
   if (!(job->printer->driver_data.printfile_cb)(job, options, device))
   {
@@ -857,8 +945,8 @@ finish_job(pappl_job_t  *job)		// I - Job
 					// Printer
 
 
-  pthread_rwlock_wrlock(&job->rwlock);
   pthread_rwlock_wrlock(&printer->rwlock);
+  pthread_rwlock_wrlock(&job->rwlock);
 
   if (job->is_canceled)
     job->state = IPP_JSTATE_CANCELED;
@@ -867,10 +955,15 @@ finish_job(pappl_job_t  *job)		// I - Job
 
   papplLogJob(job, PAPPL_LOGLEVEL_INFO, "%s, job-impressions-completed=%d.", job->state == IPP_JSTATE_COMPLETED ? "Completed" : job->state == IPP_JSTATE_CANCELED ? "Canceled" : "Aborted", job->impcompleted);
 
-  job->completed          = time(NULL);
+  if (job->state >= IPP_JSTATE_CANCELED)
+    job->completed = time(NULL);
+
   printer->processing_job = NULL;
 
-  _papplJobRemoveFile(job);
+  if (!printer->max_preserved_jobs)
+    _papplJobRemoveFile(job);
+
+  _papplSystemAddEventNoLock(job->system, job->printer, job, PAPPL_EVENT_JOB_COMPLETED, NULL);
 
   pthread_rwlock_unlock(&job->rwlock);
 
@@ -896,6 +989,11 @@ finish_job(pappl_job_t  *job)		// I - Job
   if (!job->system->clean_time)
     job->system->clean_time = time(NULL) + 60;
 
+  _papplSystemAddEventNoLock(printer->system, printer, NULL, PAPPL_EVENT_PRINTER_STATE_CHANGED, NULL);
+
+  if (printer->max_preserved_jobs > 0)
+    _papplPrinterCleanJobsNoLock(printer);
+
   pthread_rwlock_unlock(&printer->rwlock);
 
   _papplSystemConfigChanged(printer->system);
@@ -904,7 +1002,7 @@ finish_job(pappl_job_t  *job)		// I - Job
   {
     papplPrinterDelete(printer);
   }
-  else if (cupsArrayCount(printer->active_jobs) > 0)
+  else if (cupsArrayGetCount(printer->active_jobs) > 0)
   {
     _papplPrinterCheckJobs(printer);
   }
@@ -930,7 +1028,7 @@ finish_job(pappl_job_t  *job)		// I - Job
 // 'start_job()' - Start processing a job...
 //
 
-static void
+static bool				// O - `true` on success, `false` otherwise
 start_job(pappl_job_t *job)		// I - Job
 {
   pappl_printer_t *printer = job->printer;
@@ -939,8 +1037,8 @@ start_job(pappl_job_t *job)		// I - Job
 
 
   // Move the job to the 'processing' state...
-  pthread_rwlock_wrlock(&job->rwlock);
   pthread_rwlock_wrlock(&printer->rwlock);
+  pthread_rwlock_wrlock(&job->rwlock);
 
   papplLogJob(job, PAPPL_LOGLEVEL_INFO, "Starting print job.");
 
@@ -948,14 +1046,16 @@ start_job(pappl_job_t *job)		// I - Job
   job->processing         = time(NULL);
   printer->processing_job = job;
 
+  _papplSystemAddEventNoLock(printer->system, printer, job, PAPPL_EVENT_JOB_STATE_CHANGED, NULL);
+
   pthread_rwlock_unlock(&job->rwlock);
 
   // Open the output device...
-  while (!printer->device)
+  while (!printer->device && !printer->is_deleted && !job->is_canceled && papplSystemIsRunning(printer->system))
   {
     printer->device = papplDeviceOpen(printer->device_uri, job->name, papplLogDevice, job->system);
 
-    if (!printer->device)
+    if (!printer->device && !printer->is_deleted && !job->is_canceled)
     {
       // Log that the printer is unavailable then sleep for 5 seconds to retry.
       if (first_open)
@@ -973,9 +1073,25 @@ start_job(pappl_job_t *job)		// I - Job
     }
   }
 
-  // Move the printer to the 'processing' state...
-  printer->state      = IPP_PSTATE_PROCESSING;
-  printer->state_time = time(NULL);
+  if (!papplSystemIsRunning(printer->system))
+  {
+    job->state = IPP_JSTATE_PENDING;
+
+    pthread_rwlock_rdlock(&job->rwlock);
+    _papplSystemAddEventNoLock(job->system, job->printer, job, PAPPL_EVENT_JOB_STATE_CHANGED, NULL);
+    pthread_rwlock_unlock(&job->rwlock);
+  }
+
+  if (printer->device)
+  {
+    // Move the printer to the 'processing' state...
+    printer->state      = IPP_PSTATE_PROCESSING;
+    printer->state_time = time(NULL);
+  }
+
+  _papplSystemAddEventNoLock(printer->system, printer, NULL, PAPPL_EVENT_PRINTER_STATE_CHANGED, NULL);
 
   pthread_rwlock_unlock(&printer->rwlock);
+
+  return (printer->device != NULL);
 }
