@@ -1,7 +1,7 @@
 //
 // Job object for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -29,8 +29,8 @@ papplJobCancel(pappl_job_t *job)	// I - Job
   if (!job)
     return;
 
-  pthread_rwlock_wrlock(&job->rwlock);
   pthread_rwlock_wrlock(&job->printer->rwlock);
+  pthread_rwlock_wrlock(&job->rwlock);
 
   if (job->state == IPP_JSTATE_PROCESSING || (job->state == IPP_JSTATE_HELD && job->fd >= 0))
   {
@@ -47,12 +47,13 @@ papplJobCancel(pappl_job_t *job)	// I - Job
     cupsArrayAdd(job->printer->completed_jobs, job);
   }
 
-  pthread_rwlock_unlock(&job->printer->rwlock);
-
   if (!job->system->clean_time)
     job->system->clean_time = time(NULL) + 60;
 
   pthread_rwlock_unlock(&job->rwlock);
+  pthread_rwlock_unlock(&job->printer->rwlock);
+
+  papplSystemAddEvent(job->system, job->printer, job, PAPPL_EVENT_JOB_COMPLETED, NULL);
 }
 
 
@@ -80,7 +81,7 @@ _papplJobCreate(
 
   pthread_rwlock_wrlock(&printer->rwlock);
 
-  if (printer->max_active_jobs > 0 && cupsArrayCount(printer->active_jobs) >= printer->max_active_jobs)
+  if (printer->max_active_jobs > 0 && (int)cupsArrayGetCount(printer->active_jobs) >= printer->max_active_jobs)
   {
     pthread_rwlock_unlock(&printer->rwlock);
     return (NULL);
@@ -94,6 +95,8 @@ _papplJobCreate(
     return (NULL);
   }
 
+  pthread_rwlock_init(&job->rwlock, NULL);
+
   job->attrs   = ippNew();
   job->fd      = -1;
   job->format  = format;
@@ -132,7 +135,7 @@ _papplJobCreate(
 
   if ((attr = ippFindAttribute(attrs, "printer-uri", IPP_TAG_URI)) != NULL)
   {
-    strlcpy(job_printer_uri, ippGetString(attr, 0, NULL), sizeof(job_printer_uri));
+    papplCopyString(job_printer_uri, ippGetString(attr, 0, NULL), sizeof(job_printer_uri));
 
     snprintf(job_uri, sizeof(job_uri), "%s/%d", ippGetString(attr, 0, NULL), job->job_id);
   }
@@ -156,12 +159,65 @@ _papplJobCreate(
 
   pthread_rwlock_unlock(&printer->rwlock);
 
+  papplSystemAddEvent(printer->system, printer, job, PAPPL_EVENT_JOB_CREATED, NULL);
+
   _papplSystemConfigChanged(printer->system);
 
   return (job);
 }
 
 
+//
+// 'papplJobCreateWithFile()' - Create a job with a local file.
+//
+// This function creates a new print job with a local file.  The "num_options"
+// and "options" parameters specify additional print options, as needed.  The
+// file specified by "filename" is removed automatically if it resides in the
+// spool directory.
+//
+
+pappl_job_t *				// O - New job object or `NULL` on error
+papplJobCreateWithFile(
+    pappl_printer_t *printer,		// I - Printer
+    const char      *username,		// I - Submitting user name
+    const char      *format,		// I - MIME media type of file
+    const char      *job_name,		// I - Job name
+    int             num_options,	// I - Number of options
+    cups_option_t   *options,		// I - Options or `NULL` if none
+    const char      *filename)		// I - File to print
+{
+  pappl_job_t	*job;			// New job
+  ipp_t		*attrs;			// Attributes for job
+
+
+  // Range check input...
+  if (!printer || !username || !format || !job_name || !filename)
+    return (NULL);
+
+  // Encode options as needed...
+  if (num_options > 0 && options)
+  {
+    attrs = ippNew();
+
+    pthread_rwlock_rdlock(&printer->rwlock);
+    _papplMainloopAddOptions(attrs, (cups_len_t)num_options, options, printer->driver_attrs);
+    pthread_rwlock_unlock(&printer->rwlock);
+  }
+  else
+  {
+    attrs = NULL;
+  }
+
+  // Create the job...
+  if ((job = _papplJobCreate(printer, 0, username, format, job_name, attrs)) != NULL)
+    _papplJobSubmitFile(job, filename);
+
+  ippDelete(attrs);
+
+  return (job);
+}
+
+
 //
 // '_papplJobDelete()' - Remove a job from the system and free its memory.
 //
@@ -171,6 +227,8 @@ _papplJobDelete(pappl_job_t *job)	// I - Job
 {
   papplLogJob(job, PAPPL_LOGLEVEL_INFO, "Removing job from history.");
 
+  pthread_rwlock_destroy(&job->rwlock);
+
   ippDelete(job->attrs);
 
   free(job->message);
@@ -211,6 +269,15 @@ papplJobOpenFile(
   const char		*job_name;	// job-name value
 
 
+  // Range check input...
+  if (!job || !fname || fnamesize < 256 || !mode)
+  {
+    if (fname)
+      *fname = '\0';
+
+    return (-1);
+  }
+
   // Make sure the spool directory exists...
   if (!directory)
     directory = job->system->directory;
@@ -280,9 +347,9 @@ papplJobOpenFile(
   snprintf(fname, fnamesize, "%s/p%05dj%09d-%s.%s", directory, job->printer->printer_id, job->job_id, name, ext);
 
   if (!strcmp(mode, "r"))
-    return (open(fname, O_RDONLY | O_NOFOLLOW | O_CLOEXEC));
+    return (open(fname, O_RDONLY | O_NOFOLLOW | O_CLOEXEC | O_BINARY));
   else if (!strcmp(mode, "w"))
-    return (open(fname, O_WRONLY | O_CREAT | O_TRUNC | O_NOFOLLOW | O_CLOEXEC, 0600));
+    return (open(fname, O_WRONLY | O_CREAT | O_TRUNC | O_NOFOLLOW | O_CLOEXEC | O_BINARY, 0600));
   else if (!strcmp(mode, "x"))
     return (unlink(fname));
   else
@@ -299,11 +366,17 @@ _papplJobRemoveFile(pappl_job_t *job)	// I - Job
 {
   size_t dirlen = strlen(job->system->directory);
 					// Length of spool directory
+  const char *tempdir = papplGetTempDir();
+					// Location of temporary files
+  size_t templen = strlen(tempdir);	// Length of temporary directory
 
 
-  // Only remove the file if it is in spool directory...
-  if (job->filename && !strncmp(job->filename, job->system->directory, dirlen) && job->filename[dirlen] == '/')
-    unlink(job->filename);
+  // Only remove the file if it is in spool or temporary directory...
+  if (job->filename)
+  {
+    if ((!strncmp(job->filename, job->system->directory, dirlen) && job->filename[dirlen] == '/') || (!strncmp(job->filename, tempdir, templen) && job->filename[templen] == '/'))
+      unlink(job->filename);
+  }
 
   free(job->filename);
   job->filename = NULL;
@@ -387,11 +460,16 @@ _papplJobSubmitFile(
   else
   {
     // Abort the job...
+    size_t dirlen = strlen(job->system->directory);
+					// Length of spool directory
+
     job->state     = IPP_JSTATE_ABORTED;
     job->completed = time(NULL);
 
     papplLogJob(job, PAPPL_LOGLEVEL_ERROR, "Unable to allocate filename.");
-    unlink(filename);
+
+    if (!strncmp(filename, job->system->directory, dirlen) && filename[dirlen] == '/')
+      unlink(filename);
 
     pthread_rwlock_wrlock(&job->printer->rwlock);
     cupsArrayRemove(job->printer->active_jobs, job);
@@ -436,9 +514,9 @@ _papplPrinterCheckJobs(
   pthread_rwlock_wrlock(&printer->rwlock);
 
   // Enumerate the jobs.  Since we have a writer (exclusive) lock, we are the
-  // only thread enumerating and can use cupsArrayFirst/Last...
+  // only thread enumerating and can use cupsArrayGetFirst/Last...
 
-  for (job = (pappl_job_t *)cupsArrayFirst(printer->active_jobs); job; job = (pappl_job_t *)cupsArrayNext(printer->active_jobs))
+  for (job = (pappl_job_t *)cupsArrayGetFirst(printer->active_jobs); job; job = (pappl_job_t *)cupsArrayGetNext(printer->active_jobs))
   {
     if (job->state == IPP_JSTATE_PENDING)
     {
@@ -470,6 +548,46 @@ _papplPrinterCheckJobs(
 }
 
 
+//
+// '_papplPrinterCleanJobsNoLock()' - Clean completed jobs for a printer.
+//
+
+void
+_papplPrinterCleanJobsNoLock(
+    pappl_printer_t *printer)		// I - Printer
+{
+  time_t	cleantime;		// Clean time
+  pappl_job_t	*job;			// Current job
+  int		preserved;		// Number of preserved jobs
+
+
+  if (cupsArrayGetCount(printer->completed_jobs) == 0 || (printer->max_preserved_jobs == 0 && printer->max_completed_jobs <= 0))
+    return;
+
+  // Enumerate the jobs.  Since we have a writer (exclusive) lock, we are the
+  // only thread enumerating and can use cupsArrayGetFirst/Last...
+  for (job = (pappl_job_t *)cupsArrayGetFirst(printer->completed_jobs), cleantime = time(NULL) - 60, preserved = 0; job; job = (pappl_job_t *)cupsArrayGetNext(printer->completed_jobs))
+  {
+    if (job->completed && job->completed < cleantime && printer->max_completed_jobs > 0 && (int)cupsArrayGetCount(printer->completed_jobs) > printer->max_completed_jobs)
+    {
+      cupsArrayRemove(printer->completed_jobs, job);
+      cupsArrayRemove(printer->all_jobs, job);
+    }
+    else if (printer->max_preserved_jobs > 0)
+    {
+      if (job->filename)
+      {
+	preserved ++;
+	if (preserved > printer->max_preserved_jobs)
+	  _papplJobRemoveFile(job);
+      }
+    }
+    else
+      break;
+  }
+}
+
+
 //
 // 'papplPrinterFindJob()' - Find a job.
 //
@@ -510,45 +628,24 @@ void
 papplSystemCleanJobs(
     pappl_system_t *system)		// I - System
 {
-  int			i,		// Looping var
+  size_t		i,		// Looping var
 			count;		// Number of printers
   pappl_printer_t	*printer;	// Current printer
-  pappl_job_t		*job;		// Current job
-  time_t		cleantime;	// Clean time
 
 
-  cleantime = time(NULL) - 60;
-
   pthread_rwlock_rdlock(&system->rwlock);
 
   // Loop through the printers.
   //
-  // Note: Cannot use cupsArrayFirst/Last since other threads might be
+  // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
   // enumerating the printers array.
 
-  for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+  for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
   {
-    printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
-
-    if (cupsArrayCount(printer->completed_jobs) == 0 || printer->max_completed_jobs <= 0)
-      continue;
+    printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
     pthread_rwlock_wrlock(&printer->rwlock);
-
-    // Enumerate the jobs.  Since we have a writer (exclusive) lock, we are the
-    // only thread enumerating and can use cupsArrayFirst/Last...
-
-    for (job = (pappl_job_t *)cupsArrayFirst(printer->completed_jobs); job; job = (pappl_job_t *)cupsArrayNext(printer->completed_jobs))
-    {
-      if (job->completed && job->completed < cleantime && cupsArrayCount(printer->completed_jobs) > printer->max_completed_jobs)
-      {
-	cupsArrayRemove(printer->completed_jobs, job);
-	cupsArrayRemove(printer->all_jobs, job);
-      }
-      else
-	break;
-    }
-
+    _papplPrinterCleanJobsNoLock(printer);
     pthread_rwlock_unlock(&printer->rwlock);
   }
 
