@@ -1,7 +1,7 @@
 //
 // Logging functions for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2021 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -17,7 +17,9 @@
 #include "printer-private.h"
 #include "system-private.h"
 #include <stdarg.h>
-#include <syslog.h>
+#if !_WIN32
+#  include <syslog.h>
+#endif // !_WIN32
 
 
 //
@@ -34,6 +36,7 @@ static void	write_log(pappl_system_t *system, pappl_loglevel_t level, const char
 
 static pthread_mutex_t	log_mutex = PTHREAD_MUTEX_INITIALIZER;
 					// Log rotation mutex
+#if !_WIN32
 static const int	syslevels[] =	// Mapping of log levels to syslog
 {
   LOG_DEBUG | LOG_PID | LOG_LPR,
@@ -42,6 +45,7 @@ static const int	syslevels[] =	// Mapping of log levels to syslog
   LOG_ERR | LOG_PID | LOG_LPR,
   LOG_CRIT | LOG_PID | LOG_LPR
 };
+#endif // !_WIN32
 
 
 //
@@ -93,8 +97,10 @@ papplLog(pappl_system_t   *system,	// I - System
 
   if (system->logfd >= 0)
     write_log(system, level, message, ap);
+#if !_WIN32
   else
     vsyslog(syslevels[level], message, ap);
+#endif // !_WIN32
 
   va_end(ap);
 }
@@ -194,8 +200,10 @@ papplLogClient(
 
   if (client->system->logfd >= 0)
     write_log(client->system, level, cmessage, ap);
+#if !_WIN32
   else
     vsyslog(syslevels[level], cmessage, ap);
+#endif // !_WIN32
 
   va_end(ap);
 }
@@ -259,8 +267,10 @@ papplLogJob(
 
   if (job->system->logfd >= 0)
     write_log(job->system, level, jmessage, ap);
+#if !_WIN32
   else
     vsyslog(syslevels[level], jmessage, ap);
+#endif // !_WIN32
 
   va_end(ap);
 }
@@ -304,7 +314,7 @@ _papplLogOpen(
   }
 
   // Log the system status information
-  papplLog(system, PAPPL_LOGLEVEL_INFO, "Starting log, system up %ld second(s), %d printer(s), listening for connections on '%s:%d'.", (long)(time(NULL) - system->start_time), cupsArrayCount(system->printers), system->hostname, system->port);
+  papplLog(system, PAPPL_LOGLEVEL_INFO, "Starting log, system up %ld second(s), %d printer(s), listening for connections on '%s:%d' from up to %d clients.", (long)(time(NULL) - system->start_time), (int)cupsArrayGetCount(system->printers), system->hostname, system->port, system->max_clients);
 }
 
 
@@ -346,7 +356,7 @@ papplLogPrinter(
 
   // Prefix the message with "[Printer foo]", making sure to not insert any
   // printf format specifiers.
-  strlcpy(pmessage, "[Printer ", sizeof(pmessage));
+  papplCopyString(pmessage, "[Printer ", sizeof(pmessage));
   for (pptr = pmessage + 9, nameptr = printer->name; *nameptr && pptr < (pmessage + 200); pptr ++)
   {
     if (*nameptr == '%')
@@ -355,15 +365,17 @@ papplLogPrinter(
   }
   *pptr++ = ']';
   *pptr++ = ' ';
-  strlcpy(pptr, message, sizeof(pmessage) - (size_t)(pptr - pmessage));
+  papplCopyString(pptr, message, sizeof(pmessage) - (size_t)(pptr - pmessage));
 
   // Write the log message...
   va_start(ap, message);
 
   if (printer->system->logfd >= 0)
     write_log(printer->system, level, pmessage, ap);
+#if !_WIN32
   else
     vsyslog(syslevels[level], pmessage, ap);
+#endif // !_WIN32
 
   va_end(ap);
 }
@@ -385,6 +397,12 @@ rotate_log(pappl_system_t *system)	// I - System
     // Rename existing log file to "xxx.O"
     char	backname[1024];		// Backup log filename
 
+#if _WIN32
+    // Windows doesn't allow an open file to be renamed...
+    close(system->logfd);
+    system->logfd = -1;
+#endif // _WIN32
+
     snprintf(backname, sizeof(backname), "%s.O", system->logfile);
     unlink(backname);
     rename(system->logfile, backname);
@@ -430,7 +448,12 @@ write_log(pappl_system_t   *system,	// I - System
 
   // Each log line starts with a standard prefix of log level and date/time...
   gettimeofday(&curtime, NULL);
+#if _WIN32
+  time_t curtemp = (time_t)curtime.tv_sec;
+  gmtime_s(&curdate, &curtemp);
+#else
   gmtime_r(&curtime.tv_sec, &curdate);
+#endif // _WIN32
 
   snprintf(buffer, sizeof(buffer), "%c [%04d-%02d-%02dT%02d:%02d:%02d.%03dZ] ", prefix[level], curdate.tm_year + 1900, curdate.tm_mon + 1, curdate.tm_mday, curdate.tm_hour, curdate.tm_min, curdate.tm_sec, (int)(curtime.tv_usec / 1000));
   bufptr = buffer + 29;			// Skip level/date/time
@@ -628,7 +651,7 @@ write_log(pappl_system_t   *system,	// I - System
             break;
 
         default : // Something else we don't support
-            strlcpy(bufptr, tformat, (size_t)(bufptr - bufend + 1));
+            papplCopyString(bufptr, tformat, (size_t)(bufptr - bufend + 1));
             bufptr += strlen(bufptr);
             break;
       }
