@@ -1,7 +1,7 @@
 //
 // papplMainloop support functions for the Printer Application Framework
 //
-// Copyright © 2020 by Michael R Sweet.
+// Copyright © 2020-2021 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -12,8 +12,11 @@
 //
 
 #  include "pappl-private.h"
-#  include <spawn.h>
-#  include <libgen.h>
+#  if _WIN32
+#  else
+#    include <spawn.h>
+#    include <libgen.h>
+#  endif // _WIN32
 
 
 //
@@ -37,7 +40,7 @@ static int	get_length(const char *value);
 void
 _papplMainloopAddOptions(
     ipp_t         *request,		// I - IPP request
-    int           num_options,		// I - Number of options
+    cups_len_t    num_options,		// I - Number of options
     cups_option_t *options,		// I - Options
     ipp_t         *supported)		// I - Supported attributes
 {
@@ -56,7 +59,7 @@ _papplMainloopAddOptions(
 
 
   // Determine what kind of options we are adding...
-  group_tag  = ippGetOperation(request) == IPP_PRINT_JOB ? IPP_TAG_JOB : IPP_TAG_PRINTER;
+  group_tag  = ippGetOperation(request) == IPP_OP_PRINT_JOB ? IPP_TAG_JOB : IPP_TAG_PRINTER;
   is_default = (group_tag == IPP_TAG_PRINTER);
 
   if (is_default)
@@ -70,7 +73,7 @@ _papplMainloopAddOptions(
 
     if ((value = cupsGetOption("media-ready", num_options, options)) != NULL)
     {
-      int	num_values;		// Number of values
+      cups_len_t num_values;		// Number of values
       char	*values[PAPPL_MAX_SOURCE],
 					// Pointers to size strings
 		*cvalue,		// Copied value
@@ -80,8 +83,18 @@ _papplMainloopAddOptions(
       cvalue     = strdup(value);
       cptr       = cvalue;
 
+#if _WIN32
+      if ((values[num_values] = strtok_s(cvalue, ",", &cptr)) != NULL)
+      {
+        num_values ++;
+
+        while (num_values < PAPPL_MAX_SOURCE && (values[num_values] = strtok_s(NULL, ",", &cptr)) != NULL)
+          num_values ++;
+      }
+#else
       while (num_values < PAPPL_MAX_SOURCE && (values[num_values] = strsep(&cptr, ",")) != NULL)
         num_values ++;
+#endif // _WIN32
 
       if (num_values > 0)
         ippAddStrings(request, IPP_TAG_PRINTER, IPP_TAG_KEYWORD, "media-ready", num_values, NULL, (const char * const *)values);
@@ -107,6 +120,27 @@ _papplMainloopAddOptions(
     if ((value = cupsGetOption("printer-organizational-unit", num_options, options)) != NULL)
       ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_TEXT, "printer-organizational-unit", NULL, value);
   }
+  else
+  {
+    if ((value = cupsGetOption("compression", num_options, options)) != NULL)
+      ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD, "compression", NULL, value);
+
+    if ((value = cupsGetOption("page-ranges", num_options, options)) != NULL)
+    {
+      int	first_page = 1,		// First page
+		last_page = INT_MAX;	// Last page
+      char	*valptr = (char *)value;// Pointer into value
+
+      if (isdigit(*valptr & 255))
+	first_page = (int)strtol(valptr, &valptr, 10);
+      if (*valptr == '-')
+	valptr ++;
+      if (isdigit(*valptr & 255))
+	last_page = (int)strtol(valptr, &valptr, 10);
+
+      ippAddRange(request, IPP_TAG_JOB, "page-ranges", first_page, last_page);
+    }
+  }
 
   if ((value = cupsGetOption("copies", num_options, options)) == NULL)
     value = cupsGetOption("copies-default", num_options, options);
@@ -214,7 +248,7 @@ _papplMainloopAddOptions(
       {
         xres = 300;
 
-        strlcpy(units, "dpi", sizeof(units));
+        papplCopyString(units, "dpi", sizeof(units));
       }
 
       yres = xres;
@@ -226,7 +260,7 @@ _papplMainloopAddOptions(
   // Vendor attributes/options
   if ((job_attrs = ippFindAttribute(supported, "job-creation-attributes-supported", IPP_TAG_KEYWORD)) != NULL)
   {
-    int		i,			// Looping var
+    cups_len_t	i,			// Looping var
 		count;			// Count
     const char	*name;			// Attribute name
     char	defname[128],		// xxx-default name
@@ -293,10 +327,30 @@ _papplMainloopAddPrinterURI(
     char       *resource,		// I - Resource path buffer
     size_t     rsize)			// I - Size of buffer
 {
-  char	uri[1024];			// printer-uri value
+  char	uri[1024],			// printer-uri value
+	*resptr;			// Pointer into resource path
 
 
   snprintf(resource, rsize, "/ipp/print/%s", printer_name);
+  for (resptr = resource + 11; *resptr; resptr ++)
+  {
+    if ((*resptr & 255) <= ' ' || strchr("\177/\\\'\"?#", *resptr))
+      *resptr = '_';
+  }
+
+  // Eliminate duplicate and trailing underscores...
+  resptr = resource + 11;
+  while (*resptr)
+  {
+    if (resptr[0] == '_' && resptr[1] == '_')
+      memmove(resptr, resptr + 1, strlen(resptr));
+					// Duplicate underscores
+    else if (resptr[0] == '_' && !resptr[1])
+      *resptr = '\0';			// Trailing underscore
+    else
+      resptr ++;
+  }
+
   httpAssembleURI(HTTP_URI_CODING_ALL, uri, sizeof(uri), "ipp", NULL, "localhost", 0, resource);
 
   ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI, "printer-uri", NULL, uri);
@@ -317,13 +371,24 @@ _papplMainloopConnect(
 
 
   // See if the server is running...
-  http = httpConnect2(_papplMainloopGetServerPath(base_name, sockname, sizeof(sockname)), 0, NULL, AF_UNSPEC, HTTP_ENCRYPTION_IF_REQUESTED, 1, 30000, NULL);
+  http = httpConnect(_papplMainloopGetServerPath(base_name, getuid(), sockname, sizeof(sockname)), _papplMainloopGetServerPort(base_name), NULL, AF_UNSPEC, HTTP_ENCRYPTION_IF_REQUESTED, 1, 30000, NULL);
+
+#if !_WIN32
+  if (!http && getuid())
+  {
+    // Try root server...
+    http = httpConnect(_papplMainloopGetServerPath(base_name, 0, sockname, sizeof(sockname)), _papplMainloopGetServerPort(base_name), NULL, AF_UNSPEC, HTTP_ENCRYPTION_IF_REQUESTED, 1, 30000, NULL);
+  }
+#endif // !_WIN32
 
   if (!http && auto_start)
   {
     // Nope, start it now...
+    int		tries;			// Number of retries
+#if !_WIN32
     pid_t	server_pid;		// Server process ID
     posix_spawnattr_t server_attrs;	// Server process attributes
+#endif // !_WIN32
     char * const server_argv[] =	// Server command-line
     {
       _papplMainloopPath,
@@ -333,29 +398,45 @@ _papplMainloopConnect(
       NULL
     };
 
+#if _WIN32
+  int status = (int)_spawnvpe(_P_NOWAIT, _papplMainloopPath, server_argv, environ);
+
+  if (status != 0)
+  {
+    char status_str[32];		// Status string
+
+    snprintf(status_str, sizeof(status_str), "%d", status);
+    _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unable to start server: %s"), base_name, status_str);
+    return (NULL);
+  }
+
+#else
     posix_spawnattr_init(&server_attrs);
     posix_spawnattr_setpgroup(&server_attrs, 0);
 
     if (posix_spawn(&server_pid, _papplMainloopPath, NULL, &server_attrs, server_argv, environ))
     {
-      fprintf(stderr, "%s: Unable to start server: %s\n", base_name, strerror(errno));
+      _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unable to start server: %s"), base_name, strerror(errno));
       posix_spawnattr_destroy(&server_attrs);
       return (NULL);
     }
 
     posix_spawnattr_destroy(&server_attrs);
+#endif // _WIN32
 
     // Wait for it to start...
-    do
+    _papplMainloopGetServerPath(base_name, getuid(), sockname, sizeof(sockname));
+
+    for (tries = 0; tries < 40; tries ++)
     {
       usleep(250000);
-    }
-    while (access(sockname, 0));
 
-    http = httpConnect2(sockname, 0, NULL, AF_UNSPEC, HTTP_ENCRYPTION_IF_REQUESTED, 1, 30000, NULL);
+      if ((http = httpConnect(_papplMainloopGetServerPath(base_name, getuid(), sockname, sizeof(sockname)), _papplMainloopGetServerPort(base_name), NULL, AF_UNSPEC, HTTP_ENCRYPTION_IF_REQUESTED, 1, 30000, NULL)) != NULL)
+        break;
+    }
 
     if (!http)
-      fprintf(stderr, "%s: Unable to connect to server: %s\n", base_name, cupsLastErrorString());
+      _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unable to connect to server: %s"), base_name, cupsLastErrorString());
   }
 
   return (http);
@@ -384,26 +465,26 @@ _papplMainloopConnectURI(
   // First extract the components of the URI...
   if (httpSeparateURI(HTTP_URI_CODING_ALL, printer_uri, scheme, sizeof(scheme), userpass, sizeof(userpass), hostname, sizeof(hostname), &port, resource, (int)rsize) < HTTP_URI_STATUS_OK)
   {
-    fprintf(stderr, "%s: Bad printer URI '%s'.\n", base_name, printer_uri);
+    _papplLocPrintf(stderr, _PAPPL_LOC("%s: Bad printer URI '%s'."), base_name, printer_uri);
     return (NULL);
   }
 
   if (strcmp(scheme, "ipp") && strcmp(scheme, "ipps"))
   {
-    fprintf(stderr, "%s: Unsupported URI scheme '%s'.\n", base_name, scheme);
+    _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unsupported URI scheme '%s'."), base_name, scheme);
     return (NULL);
   }
 
   if (userpass[0])
-    fprintf(stderr, "%s: Warning - user credentials are not supported in URIs.\n", base_name);
+    _papplLocPrintf(stderr, _PAPPL_LOC("%s: Warning - user credentials are not supported in URIs."), base_name);
 
   if (!strcmp(scheme, "ipps") || port == 443)
     encryption = HTTP_ENCRYPTION_ALWAYS;
   else
     encryption = HTTP_ENCRYPTION_IF_REQUESTED;
 
-  if ((http = httpConnect2(hostname, port, NULL, AF_UNSPEC, encryption, 1, 30000, NULL)) == NULL)
-    fprintf(stderr, "%s: Unable to connect to printer at '%s:%d': %s\n", base_name, hostname, port, cupsLastErrorString());
+  if ((http = httpConnect(hostname, port, NULL, AF_UNSPEC, encryption, 1, 30000, NULL)) == NULL)
+    _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unable to connect to printer at '%s:%d': %s"), base_name, hostname, port, cupsLastErrorString());
 
   return (http);
 }
@@ -426,13 +507,13 @@ _papplMainloopGetDefaultPrinter(
 
   // Ask the server for its default printer
   request = ippNewRequest(IPP_OP_CUPS_GET_DEFAULT);
-  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME, "requesting-user-name", NULL, cupsUser());
+  ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME, "requesting-user-name", NULL, cupsGetUser());
   ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD, "requested-attributes", NULL, "printer-name");
 
   response = cupsDoRequest(http, request, "/ipp/system");
 
   if ((printer_name = ippGetString(ippFindAttribute(response, "printer-name", IPP_TAG_NAME), 0, NULL)) != NULL)
-    strlcpy(buffer, printer_name, bufsize);
+    papplCopyString(buffer, printer_name, bufsize);
   else
     *buffer = '\0';
 
@@ -449,27 +530,82 @@ _papplMainloopGetDefaultPrinter(
 char *					// O - Socket filename
 _papplMainloopGetServerPath(
     const char *base_name,		// I - Base name
+    uid_t      uid,			// I - UID for server
     char       *buffer,			// I - Buffer for filename
     size_t     bufsize)			// I - Size of buffer
 {
-  const char	*tmpdir = getenv("TMPDIR");
-					// Temporary directory
+#if _WIN32
+  // Server running as local service...
+  (void)base_name;
+  (void)uid;
+
+  papplCopyString(buffer, "localhost", bufsize);
 
-#ifdef __APPLE__
-  if (!tmpdir)
-    tmpdir = "/private/tmp";
 #else
-  if (!tmpdir)
-    tmpdir = "/tmp";
-#endif // __APPLE__
+  const char	*snap_common;		// SNAP_COMMON environment variable
 
-  snprintf(buffer, bufsize, "%s/%s%d.sock", tmpdir, base_name, (int)getuid());
-  _PAPPL_DEBUG("Creating domain socket as '%s'.\n", buffer);
+
+  if (uid)
+  {
+    // Per-user server...
+    snprintf(buffer, bufsize, "%s/%s%d.sock", papplGetTempDir(), base_name, (int)uid);
+  }
+  else if ((snap_common = getenv("SNAP_COMMON")) != NULL)
+  {
+    // System server running as root inside a snap (https://snapcraft.io)...
+    snprintf(buffer, bufsize, "%s/%s.sock", snap_common, base_name);
+  }
+  else
+  {
+    // System server running as root
+    snprintf(buffer, bufsize, PAPPL_SOCKDIR "/%s.sock", base_name);
+  }
+
+  _PAPPL_DEBUG("Using domain socket '%s'.\n", buffer);
+#endif // _WIN32
 
   return (buffer);
 }
 
 
+//
+// '_papplMainloopGetServerPort()' - Get the socket port number for the server.
+//
+
+int					// O - Port number
+_papplMainloopGetServerPort(
+    const char *base_name)		// I - Base name of printer application
+{
+#if _WIN32
+  char		path[1024];		// Registry path
+  HKEY		key;			// Registry key
+  DWORD		dport = 0;		// Port number value for registry
+
+
+  // The server's port number is saved in SOFTWARE\appname\port
+  snprintf(path, sizeof(path), "SOFTWARE\\%s", base_name);
+
+  if (!RegOpenKeyExA(HKEY_LOCAL_MACHINE, path, 0, KEY_READ, &key))
+  {
+    // Was able to open the registry, get the port number...
+    DWORD dsize = sizeof(dport);	// Size of port number value
+
+    RegGetValueA(key, NULL, "port", RRF_RT_REG_DWORD, NULL, &dport, &dsize);
+    RegCloseKey(key);
+  }
+
+  return ((int)dport);
+
+#else
+  // POSIX platforms use domain instead of TCP/IP sockets...
+  (void)base_name;
+
+  return (0);
+#endif // _WIN32
+}
+
+
+
 //
 // 'get_length()' - Get a length in hundreths of millimeters.
 //
