@@ -1,7 +1,7 @@
 //
 // Command line utilities for the Printer Application Framework
 //
-// Copyright © 2020 by Michael R Sweet.
+// Copyright © 2020-2022 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -11,9 +11,10 @@
 // Include necessary headers
 //
 
-#  include "pappl-private.h"
-#  include <spawn.h>
+#include "pappl-private.h"
+#if !_WIN32
 #  include <libgen.h>
+#endif // !_WIN32
 
 
 //
@@ -76,9 +77,9 @@ papplMainloop(
   int		i, j;			// Looping vars
   const char	*opt;			// Option character
   const char	*subcommand = NULL;	// Sub-command
-  int		num_files = 0;		// File count
+  cups_len_t	num_files = 0;		// File count
   char		*files[1000];		// Files array
-  int		num_options = 0;	// Number of options
+  cups_len_t	num_options = 0;	// Number of options
   cups_option_t	*options = NULL;	// Options
   static const char * const subcommands[] =
   {					// List of standard sub-commands
@@ -92,12 +93,17 @@ papplMainloop(
     "jobs",
     "modify",
     "options",
+    "pause",
     "printers",
+    "resume",
     "server",
     "shutdown",
     "status",
     "submit"
   };
+#ifdef __APPLE__
+  const char	*server_argv[7];	// New command arguments
+#endif // __APPLE__
 
 
   // Range check input...
@@ -115,7 +121,35 @@ papplMainloop(
 
   // Save the path to the printer application and get the base name.
   _papplMainloopPath = argv[0];
-  base_name          = basename(argv[0]);
+
+#if _WIN32
+  if ((base_name = strrchr(argv[0], '\\')) == NULL)
+    if ((base_name = strrchr(argv[0], '/')) == NULL)
+      base_name = argv[0];
+
+#else
+  base_name = basename(argv[0]);
+#endif // _WIN32
+
+#ifdef __APPLE__
+  // When you click on the application icon, macOS' Finder will pass "-psn_..."
+  // on the command-line.  We also recognize when running the application from
+  // within the .app bundle.  If either is true, replace argc and argv to run a
+  // server...
+  if ((argc > 1 && !strncmp(argv[1], "-psn", 4)) || strstr(argv[0], ".app/Contents/MacOS/") != NULL)
+  {
+    server_argv[0] = argv[0];
+    server_argv[1] = "server";
+    server_argv[2] = "-o";
+    server_argv[3] = "log-file=syslog";
+    server_argv[4] = "-o";
+    server_argv[5] = "log-level=info";
+    server_argv[6] = NULL;
+
+    argc = 6;
+    argv = (char **)server_argv;
+  }
+#endif // __APPLE__
 
   // Parse the command-line...
   for (i = 1; i < argc; i ++)
@@ -141,12 +175,12 @@ papplMainloop(
 
       if (i >= argc)
       {
-        fprintf(stderr, "%s: Missing filename after '--'.\n", base_name);
+        _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing filename after '--'."), base_name);
         return (1);
       }
       else if (num_files >= (int)(sizeof(files) / sizeof(files[0])))
       {
-        fprintf(stderr, "%s: Too many files.\n", base_name);
+        _papplLocPrintf(stderr, _PAPPL_LOC("%s: Too many files."), base_name);
         return (1);
       }
 
@@ -154,7 +188,7 @@ papplMainloop(
     }
     else if (!strncmp(argv[i], "--", 2))
     {
-      fprintf(stderr, "%s: Unknown option '%s'.\n", base_name, argv[i]);
+      _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unknown option '%s'."), base_name, argv[i]);
       return (1);
     }
     else if (argv[i][0] == '-' && argv[i][1])
@@ -171,7 +205,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing printer name after '-d'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing printer name after '-d'."), base_name);
                 return (1);
               }
 
@@ -182,7 +216,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing hostname after '-h'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing hostname after '-h'."), base_name);
                 return (1);
               }
 
@@ -193,7 +227,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing job ID after '-j'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing job ID after '-j'."), base_name);
                 return (1);
               }
 
@@ -204,7 +238,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing driver name after '-m'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing driver name after '-m'."), base_name);
                 return (1);
               }
 
@@ -215,7 +249,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing copy count after '-n'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing copy count after '-n'."), base_name);
                 return (1);
               }
 
@@ -225,14 +259,14 @@ papplMainloop(
           case 'o': // -o "NAME=VALUE [... NAME=VALUE]"
               if (opt[1] && strchr(opt, '='))
               {
-                fprintf(stderr, "%s: Missing space after '-o'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing space after '-o'."), base_name);
                 return (1);
               }
 
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing option(s) after '-o'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing option(s) after '-o'."), base_name);
                 return (1);
               }
 
@@ -243,7 +277,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing title after '-t'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing title after '-t'."), base_name);
                 return (1);
 	      }
 
@@ -254,7 +288,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing printer URI after '-u'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing printer URI after '-u'."), base_name);
                 return (1);
               }
 
@@ -265,7 +299,7 @@ papplMainloop(
               i ++;
               if (i >= argc)
               {
-                fprintf(stderr, "%s: Missing device URI after '-v'.\n", base_name);
+                _papplLocPrintf(stderr, _PAPPL_LOC("%s: Missing device URI after '-v'."), base_name);
                 return (1);
               }
 
@@ -273,7 +307,7 @@ papplMainloop(
 	      break;
 
           default:
-              fprintf(stderr, "%s: Unknown option '-%c'.\n", base_name, *opt);
+              _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unknown option '-%c'."), base_name, *opt);
               return (1);
         }
       }
@@ -296,7 +330,7 @@ papplMainloop(
         {
           if (subcommand)
           {
-            fprintf(stderr, "%s: Cannot specify more than one sub-command.\n", base_name);
+            _papplLocPrintf(stderr, _PAPPL_LOC("%s: Cannot specify more than one sub-command."), base_name);
             return (1);
           }
 
@@ -313,12 +347,17 @@ papplMainloop(
           // Extension sub-command...
           if (subcommand)
           {
-            fprintf(stderr, "%s: Cannot specify more than one sub-command.\n", base_name);
+            _papplLocPrintf(stderr, _PAPPL_LOC("%s: Cannot specify more than one sub-command."), base_name);
             return (1);
           }
 
           subcommand = argv[i];
         }
+        else if (access(argv[i], R_OK))
+        {
+          _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unable to access '%s': %s"), base_name, argv[i], strerror(errno));
+          return (1);
+        }
         else if (num_files < (int)(sizeof(files) / sizeof(files[0])))
         {
           // Filename...
@@ -327,7 +366,7 @@ papplMainloop(
         else
 	{
 	  // Too many files...
-	  fprintf(stderr, "%s: Too many files.\n", base_name);
+	  _papplLocPrintf(stderr, _PAPPL_LOC("%s: Too many files."), base_name);
 	  return (1);
 	}
       }
@@ -341,11 +380,11 @@ papplMainloop(
   }
   else if (subcmd_name && !strcmp(subcommand, subcmd_name))
   {
-    return ((subcmd_cb)(base_name, num_options, options, num_files, files, data));
+    return ((subcmd_cb)(base_name, (int)num_options, options, (int)num_files, files, data));
   }
   else if (num_files > 0)
   {
-    fprintf(stderr, "%s: Sub-command '%s' does not accept files.\n", base_name, subcommand);
+    _papplLocPrintf(stderr, _PAPPL_LOC("%s: Sub-command '%s' does not accept files."), base_name, subcommand);
     return (1);
   }
   else if (!strcmp(subcommand, "add"))
@@ -360,7 +399,7 @@ papplMainloop(
     }
     else
     {
-      fprintf(stderr, "%s: Sub-command 'autoadd' is not supported.\n", base_name);
+      _papplLocPrintf(stderr, _PAPPL_LOC("%s: Sub-command 'autoadd' is not supported."), base_name);
       return (1);
     }
   }
@@ -382,7 +421,7 @@ papplMainloop(
   }
   else if (!strcmp(subcommand, "drivers"))
   {
-    return (_papplMainloopShowDrivers(base_name, num_options, options, system_cb, data));
+    return (_papplMainloopShowDrivers(base_name, (cups_len_t)num_drivers, drivers, autoadd_cb, driver_cb, num_options, options, system_cb, data));
   }
   else if (!strcmp(subcommand, "jobs"))
   {
@@ -396,13 +435,21 @@ papplMainloop(
   {
     return (_papplMainloopShowOptions(base_name, num_options, options));
   }
+  else if (!strcmp(subcommand, "pause"))
+  {
+    return (_papplMainloopPausePrinter(base_name, num_options, options));
+  }
   else if (!strcmp(subcommand, "printers"))
   {
     return (_papplMainloopShowPrinters(base_name, num_options, options));
   }
+  else if (!strcmp(subcommand, "resume"))
+  {
+    return (_papplMainloopResumePrinter(base_name, num_options, options));
+  }
   else if (!strcmp(subcommand, "server"))
   {
-    return (_papplMainloopRunServer(base_name, version, footer_html, num_drivers, drivers, autoadd_cb, driver_cb, num_options, options, system_cb, data));
+    return (_papplMainloopRunServer(base_name, version, footer_html, (cups_len_t)num_drivers, drivers, autoadd_cb, driver_cb, num_options, options, system_cb, data));
   }
   else if (!strcmp(subcommand, "shutdown"))
   {
@@ -415,7 +462,7 @@ papplMainloop(
   else
   {
     // This should never happen...
-    fprintf(stderr, "%s: Unknown sub-command '%s'.\n", base_name, subcommand);
+    _papplLocPrintf(stderr, _PAPPL_LOC("%s: Unknown sub-command '%s'."), base_name, subcommand);
     return (1);
   }
 }
@@ -429,35 +476,37 @@ static void
 usage(const char *base_name,		// I - Base name of application
       bool       with_autoadd)		// I - `true` if autoadd command is supported
 {
-  printf("Usage: %s SUB-COMMAND [OPTIONS] [FILENAME]\n", base_name);
-  printf("       %s [OPTIONS] [FILENAME]\n", base_name);
-  printf("       %s [OPTIONS] -\n", base_name);
+  _papplLocPrintf(stdout, _PAPPL_LOC("Usage: %s SUB-COMMAND [OPTIONS] [FILENAME]\n"
+				     "       %s [OPTIONS] [FILENAME]\n"
+				     "       %s [OPTIONS] -"), base_name, base_name, base_name);
   puts("");
-  puts("Sub-commands:");
-  puts("  add PRINTER      Add a printer.");
+  _papplLocPrintf(stdout, _PAPPL_LOC("Sub-commands:"));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  add PRINTER      Add a printer."));
   if (with_autoadd)
-    puts("  autoadd          Automatically add supported printers.");
-  puts("  cancel           Cancel one or more jobs.");
-  puts("  default          Set the default printer.");
-  puts("  delete           Delete a printer.");
-  puts("  devices          List devices.");
-  puts("  drivers          List drivers.");
-  puts("  jobs             List jobs.");
-  puts("  modify           Modify a printer.");
-  puts("  options          List printer options.");
-  puts("  printers         List printers.");
-  puts("  server           Run a server.");
-  puts("  shutdown         Shutdown a running server.");
-  puts("  status           Show server/printer/job status.");
-  puts("  submit           Submit a file for printing.");
+    _papplLocPrintf(stdout, _PAPPL_LOC("  autoadd          Automatically add supported printers."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  cancel           Cancel one or more jobs."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  default          Set the default printer."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  delete           Delete a printer."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  devices          List devices."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  drivers          List drivers."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  jobs             List jobs."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  modify           Modify a printer."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  options          List printer options."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  pause            Pause printing for a printer."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  printers         List printers."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  resume           Resume printing for a printer."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  server           Run a server."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  shutdown         Shutdown a running server."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  status           Show server/printer/job status."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  submit           Submit a file for printing."));
   puts("");
-  puts("Options:");
-  puts("  -a               Cancel all jobs (cancel).");
-  puts("  -d PRINTER       Specify printer.");
-  puts("  -j JOB-ID        Specify job ID (cancel).");
-  puts("  -m DRIVER-NAME   Specify driver (add/modify).");
-  puts("  -n COPIES        Specify number of copies (submit).");
-  puts("  -o NAME=VALUE    Specify option (add,modify,server,submit).");
-  puts("  -u URI           Specify ipp: or ipps: printer/server.");
-  puts("  -v DEVICE-URI    Specify socket: or usb: device (add/modify).");
+  _papplLocPrintf(stdout, _PAPPL_LOC("Options:"));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -a               Cancel all jobs (cancel)."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -d PRINTER       Specify printer."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -j JOB-ID        Specify job ID (cancel)."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -m DRIVER-NAME   Specify driver (add/modify)."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -n COPIES        Specify number of copies (submit)."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -o NAME=VALUE    Specify option (add,modify,server,submit)."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -u URI           Specify ipp: or ipps: printer/server."));
+  _papplLocPrintf(stdout, _PAPPL_LOC("  -v DEVICE-URI    Specify socket: or usb: device (add/modify)."));
 }
