@@ -1,7 +1,7 @@
 //
 // Printer accessor functions for the Printer Application Framework
 //
-// Copyright © 2020-2021 by Michael R Sweet.
+// Copyright © 2020-2022 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -31,19 +31,51 @@ papplPrinterCloseDevice(
 
   pthread_rwlock_wrlock(&printer->rwlock);
 
-  papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Closing device.");
-
   papplDeviceClose(printer->device);
 
   printer->device        = NULL;
   printer->device_in_use = false;
 
-  papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Device closed.");
-
   pthread_rwlock_unlock(&printer->rwlock);
 }
 
 
+//
+// 'papplPrinterDisable()' - Stop accepting jobs on a printer.
+//
+// This function stops accepting jobs on a printer.
+//
+
+void
+papplPrinterDisable(
+    pappl_printer_t *printer)		// I - Printer
+{
+  if (printer)
+  {
+    printer->is_accepting = false;
+    papplSystemAddEvent(printer->system, printer, NULL, PAPPL_EVENT_PRINTER_STATE_CHANGED, NULL);
+  }
+}
+
+
+//
+// 'papplPrinterEnable()' - Start accepting jobs on a printer.
+//
+// This function starts accepting jobs on a printer.
+//
+
+void
+papplPrinterEnable(
+    pappl_printer_t *printer)		// I - Printer
+{
+  if (printer)
+  {
+    printer->is_accepting = true;
+    papplSystemAddEvent(printer->system, printer, NULL, PAPPL_EVENT_PRINTER_STATE_CHANGED, NULL);
+  }
+}
+
+
 //
 // 'papplPrinterGetContact()' - Get the "printer-contact" value.
 //
@@ -125,7 +157,7 @@ papplPrinterGetDNSSDName(
   }
 
   pthread_rwlock_rdlock(&printer->rwlock);
-  strlcpy(buffer, printer->dns_sd_name, bufsize);
+  papplCopyString(buffer, printer->dns_sd_name, bufsize);
   pthread_rwlock_unlock(&printer->rwlock);
 
   return (buffer);
@@ -155,7 +187,7 @@ papplPrinterGetGeoLocation(
   }
 
   pthread_rwlock_rdlock(&printer->rwlock);
-  strlcpy(buffer, printer->geo_location, bufsize);
+  papplCopyString(buffer, printer->geo_location, bufsize);
   pthread_rwlock_unlock(&printer->rwlock);
 
   return (buffer);
@@ -214,7 +246,7 @@ papplPrinterGetLocation(
   }
 
   pthread_rwlock_rdlock(&printer->rwlock);
-  strlcpy(buffer, printer->location, bufsize);
+  papplCopyString(buffer, printer->location, bufsize);
   pthread_rwlock_unlock(&printer->rwlock);
 
   return (buffer);
@@ -229,7 +261,7 @@ papplPrinterGetLocation(
 // supports, as configured by the @link papplPrinterSetMaxActiveJobs@ function.
 //
 
-int					// O - Maximum number of active jobs
+int					// O - Maximum number of active jobs, `0` for unlimited
 papplPrinterGetMaxActiveJobs(
     pappl_printer_t *printer)		// I - Printer
 {
@@ -246,7 +278,7 @@ papplPrinterGetMaxActiveJobs(
 // function.
 //
 
-int					// O - Maximum number of completed jobs
+int					// O - Maximum number of completed jobs, `0` for unlimited
 papplPrinterGetMaxCompletedJobs(
     pappl_printer_t *printer)		// I - Printer
 {
@@ -254,6 +286,23 @@ papplPrinterGetMaxCompletedJobs(
 }
 
 
+//
+// 'papplPrinterGetMaxPreservedJobs()' - Get the maximum number of jobs
+//                                       preserved by the printer.
+//
+// This function returns the maximum number of jobs that are retained (including
+// document data) in the job history as configured by the
+// @link papplPrinterSetMaxPreservedJobs@ function.
+//
+
+int					// O - Maximum number of preserved jobs, `0` for none
+papplPrinterGetMaxPreservedJobs(
+    pappl_printer_t *printer)		// I - Printer
+{
+  return (printer ? printer->max_preserved_jobs : 0);
+}
+
+
 //
 // 'papplPrinterGetName()' - Get the printer name.
 //
@@ -295,7 +344,7 @@ int					// O - Number of active print jobs
 papplPrinterGetNumberOfActiveJobs(
     pappl_printer_t *printer)		// I - Printer
 {
-  return (printer ? cupsArrayCount(printer->active_jobs) : 0);
+  return (printer ? (int)cupsArrayGetCount(printer->active_jobs) : 0);
 }
 
 
@@ -311,7 +360,7 @@ int					// O - Number of completed print jobs
 papplPrinterGetNumberOfCompletedJobs(
     pappl_printer_t *printer)		// I - Printer
 {
-  return (printer ? cupsArrayCount(printer->completed_jobs) : 0);
+  return (printer ? (int)cupsArrayGetCount(printer->completed_jobs) : 0);
 }
 
 
@@ -326,7 +375,7 @@ int					// O - Total number of print jobs
 papplPrinterGetNumberOfJobs(
     pappl_printer_t *printer)		// I - Printer
 {
-  return (printer ? cupsArrayCount(printer->all_jobs) : 0);
+  return (printer ? (int)cupsArrayGetCount(printer->all_jobs) : 0);
 }
 
 
@@ -352,7 +401,7 @@ papplPrinterGetOrganization(
   }
 
   pthread_rwlock_rdlock(&printer->rwlock);
-  strlcpy(buffer, printer->organization, bufsize);
+  papplCopyString(buffer, printer->organization, bufsize);
   pthread_rwlock_unlock(&printer->rwlock);
 
   return (buffer);
@@ -381,7 +430,7 @@ papplPrinterGetOrganizationalUnit(
   }
 
   pthread_rwlock_rdlock(&printer->rwlock);
-  strlcpy(buffer, printer->org_unit, bufsize);
+  papplCopyString(buffer, printer->org_unit, bufsize);
   pthread_rwlock_unlock(&printer->rwlock);
 
   return (buffer);
@@ -414,7 +463,7 @@ papplPrinterGetPath(
   if (subpath)
     snprintf(buffer, bufsize, "%s/%s", printer->uriname, subpath);
   else
-    strlcpy(buffer, printer->uriname, bufsize);
+    papplCopyString(buffer, printer->uriname, bufsize);
 
   return (buffer);
 }
@@ -442,7 +491,7 @@ papplPrinterGetPrintGroup(
   }
 
   pthread_rwlock_rdlock(&printer->rwlock);
-  strlcpy(buffer, printer->print_group, bufsize);
+  papplCopyString(buffer, printer->print_group, bufsize);
   pthread_rwlock_unlock(&printer->rwlock);
 
   return (buffer);
@@ -550,6 +599,21 @@ papplPrinterGetSystem(
 }
 
 
+//
+// '()' - Return whether the printer is accepting jobs.
+//
+// This function returns a boolean value indicating whether a printer is
+// accepting jobs.
+//
+
+bool					// O - `true` if the printer is accepting jobs, `false` otherwise
+papplPrinterIsAcceptingJobs(
+    pappl_printer_t *printer)		// I - Printer
+{
+  return (printer ? printer->is_accepting : false);
+}
+
+
 //
 // 'papplPrinterIterateActiveJobs()' - Iterate over the active jobs.
 //
@@ -571,9 +635,9 @@ papplPrinterIterateActiveJobs(
     int             limit)		// I - Maximum jobs to iterate or `0` for no limit
 {
   pappl_job_t	*job;			// Current job
-  int		j,			// Looping var
-		jcount,			// Number of jobs
-		count;			// Number of jobs iterated
+  size_t	j,			// Looping var
+		jcount;			// Number of jobs
+  int		count;			// Number of jobs iterated
 
 
   if (!printer || !cb)
@@ -581,15 +645,15 @@ papplPrinterIterateActiveJobs(
 
   pthread_rwlock_rdlock(&printer->rwlock);
 
-  // Note: Cannot use cupsArrayFirst/Last since other threads might be
+  // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
   // enumerating the active_jobs array.
 
   if (limit <= 0)
     limit = INT_MAX;
 
-  for (count = 0, j = job_index - 1, jcount = cupsArrayCount(printer->active_jobs); j < jcount && count < limit; j ++, count ++)
+  for (count = 0, j = (size_t)job_index - 1, jcount = cupsArrayGetCount(printer->active_jobs); j < jcount && count < limit; j ++, count ++)
   {
-    job = (pappl_job_t *)cupsArrayIndex(printer->active_jobs, j);
+    job = (pappl_job_t *)cupsArrayGetElement(printer->active_jobs, j);
 
     (cb)(job, data);
   }
@@ -618,9 +682,9 @@ papplPrinterIterateAllJobs(
     int             limit)		// I - Maximum jobs to iterate, `0` for no limit
 {
   pappl_job_t	*job;			// Current job
-  int		j,			// Looping var
-		jcount,			// Number of jobs
-		count;			// Number of jobs iterated
+  size_t	j,			// Looping var
+		jcount;			// Number of jobs
+  int		count;			// Number of jobs iterated
 
 
   if (!printer || !cb)
@@ -628,15 +692,15 @@ papplPrinterIterateAllJobs(
 
   pthread_rwlock_rdlock(&printer->rwlock);
 
-  // Note: Cannot use cupsArrayFirst/Last since other threads might be
+  // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
   // enumerating the all_jobs array.
 
   if (limit <= 0)
     limit = INT_MAX;
 
-  for (count = 0, j = job_index - 1, jcount = cupsArrayCount(printer->all_jobs); j < jcount && count < limit; j ++, count ++)
+  for (count = 0, j = (size_t)job_index - 1, jcount = cupsArrayGetCount(printer->all_jobs); j < jcount && count < limit; j ++, count ++)
   {
-    job = (pappl_job_t *)cupsArrayIndex(printer->all_jobs, j);
+    job = (pappl_job_t *)cupsArrayGetElement(printer->all_jobs, j);
 
     (cb)(job, data);
   }
@@ -666,9 +730,9 @@ papplPrinterIterateCompletedJobs(
     int             limit)		// I - Maximum jobs to iterate, `0` for no limit
 {
   pappl_job_t	*job;			// Current job
-  int		j,			// Looping var
-		jcount,			// Number of jobs
-		count;			// Number of jobs iterated
+  size_t	j,			// Looping var
+		jcount;			// Number of jobs
+  int		count;			// Number of jobs iterated
 
 
   if (!printer || !cb)
@@ -676,15 +740,15 @@ papplPrinterIterateCompletedJobs(
 
   pthread_rwlock_rdlock(&printer->rwlock);
 
-  // Note: Cannot use cupsArrayFirst/Last since other threads might be
+  // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
   // enumerating the completed_jobs array.
 
   if (limit <= 0)
     limit = INT_MAX;
 
-  for (count = 0, j = job_index - 1, jcount = cupsArrayCount(printer->completed_jobs); j < jcount && count < limit; j ++, count ++)
+  for (count = 0, j = (size_t)job_index - 1, jcount = cupsArrayGetCount(printer->completed_jobs); j < jcount && count < limit; j ++, count ++)
   {
-    job = (pappl_job_t *)cupsArrayIndex(printer->completed_jobs, j);
+    job = (pappl_job_t *)cupsArrayGetElement(printer->completed_jobs, j);
 
     (cb)(job, data);
   }
@@ -717,17 +781,10 @@ papplPrinterOpenDevice(
 
   if (!printer->device_in_use && !printer->processing_job)
   {
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Opening device.");
-
     printer->device        = device = papplDeviceOpen(printer->device_uri, "printer", papplLogDevice, printer->system);
     printer->device_in_use = device != NULL;
   }
 
-  if (device)
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Device open.");
-  else
-    papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "Device not open.");
-
   pthread_rwlock_unlock(&printer->rwlock);
 
   return (device);
@@ -755,6 +812,8 @@ papplPrinterPause(
   else
     printer->state = IPP_PSTATE_STOPPED;
 
+  _papplSystemAddEventNoLock(printer->system, printer, NULL, PAPPL_EVENT_PRINTER_STATE_CHANGED | PAPPL_EVENT_PRINTER_STOPPED, NULL);
+
   pthread_rwlock_unlock(&printer->rwlock);
 }
 
@@ -777,6 +836,8 @@ papplPrinterResume(
   printer->is_stopped = false;
   printer->state      = IPP_PSTATE_IDLE;
 
+  _papplSystemAddEventNoLock(printer->system, printer, NULL, PAPPL_EVENT_PRINTER_STATE_CHANGED, NULL);
+
   pthread_rwlock_unlock(&printer->rwlock);
 
   _papplPrinterCheckJobs(printer);
@@ -991,6 +1052,33 @@ papplPrinterSetMaxCompletedJobs(
 }
 
 
+//
+// 'papplPrinterSetMaxPreservedJobs()' - Set the maximum number of preserved
+//                                       jobs for the printer.
+//
+// This function sets the maximum number of aborted, canceled, or completed jobs
+// that are preserved (with document data) in the job history.
+//
+
+void
+papplPrinterSetMaxPreservedJobs(
+    pappl_printer_t *printer,		// I - Printer
+    int             max_preserved_jobs)	// I - Maximum number of preserved jobs, `0` for none
+{
+  if (!printer || max_preserved_jobs < 0)
+    return;
+
+  pthread_rwlock_wrlock(&printer->rwlock);
+
+  printer->max_preserved_jobs = max_preserved_jobs;
+  printer->config_time        = time(NULL);
+
+  pthread_rwlock_unlock(&printer->rwlock);
+
+  _papplSystemConfigChanged(printer->system);
+}
+
+
 //
 // 'papplPrinterSetNextJobID()' - Set the next "job-id" value.
 //
@@ -1098,6 +1186,7 @@ papplPrinterSetPrintGroup(
   printer->print_group = value ? strdup(value) : NULL;
   printer->config_time = time(NULL);
 
+#if !_WIN32
   if (printer->print_group && strcmp(printer->print_group, "none"))
   {
     char		buffer[8192];	// Buffer for strings
@@ -1110,6 +1199,7 @@ papplPrinterSetPrintGroup(
       printer->print_gid = grp->gr_gid;
   }
   else
+#endif // !_WIN32
     printer->print_gid = (gid_t)-1;
 
   pthread_rwlock_unlock(&printer->rwlock);
