@@ -1,7 +1,7 @@
 //
 // Printer driver functions for the Printer Application Framework
 //
-// Copyright © 2020-2021 by Michael R Sweet.
+// Copyright © 2020-2022 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -19,7 +19,7 @@
 // Local functions...
 //
 
-static ipp_t	*make_attrs(pappl_system_t *system, pappl_pr_driver_data_t *data);
+static ipp_t	*make_attrs(pappl_system_t *system, pappl_printer_t *printer, pappl_pr_driver_data_t *data);
 static bool	validate_defaults(pappl_printer_t *printer, pappl_pr_driver_data_t *driver_data, pappl_pr_driver_data_t *data);
 static bool	validate_driver(pappl_printer_t *printer, pappl_pr_driver_data_t *data);
 static bool	validate_ready(pappl_printer_t *printer, pappl_pr_driver_data_t *driver_data, int num_ready, pappl_media_col_t *ready);
@@ -122,11 +122,30 @@ _papplPrinterInitDriverData(
     { 170, 226, 250, 210,  94,  30,  22,  70, 171, 227, 251, 211,  95,  31,  23,  71 },
     { 130, 162, 218, 154, 126,  86,  78, 118, 131, 163, 219, 155, 127,  87,  79, 119 }
   };
+  static const pappl_dither_t blue =	// Blue-noise dither array
+  {
+    { 111,  49, 142, 162, 113, 195,  71, 177, 201,  50, 151,  94,  66,  37,  85, 252 },
+    {  25,  99, 239, 222,  32, 250, 148,  19,  38, 106, 220, 170, 194, 138,  13, 167 },
+    { 125, 178,  79,  15,  65, 173, 123,  87, 213, 131, 247,  23, 116,  54, 229, 212 },
+    {  41, 202, 152, 132, 189, 104,  53, 236, 161,  62,   1, 181,  77, 241, 147,  68 },
+    {   2, 244,  56,  91, 230,   5, 204,  28, 187, 101, 144, 206,  33,  92, 190, 107 },
+    { 223, 164, 114,  36, 214, 156, 139,  70, 245,  84, 226,  48, 126, 158,  17, 135 },
+    {  83, 196,  21, 254,  76,  45, 179, 115,  12,  40, 169, 105, 253, 176, 211,  59 },
+    { 100, 180, 145, 122, 172,  97, 235, 129, 215, 149, 199,   8,  72,  26, 238,  44 },
+    { 232,  31,  69,  11, 205,  58,  18, 193,  88,  60, 112, 221, 140,  86, 120, 153 },
+    { 208, 130, 243, 160, 224, 110,  34, 248, 165,  24, 234, 184,  52, 198, 171,   6 },
+    { 108, 188,  51,  89, 137, 186, 154,  78,  47, 134,  98, 157,  35, 249,  95,  63 },
+    {  16,  75, 219,  39,   0,  67, 228, 121, 197, 240,   3,  74, 127,  20, 227, 143 },
+    { 246, 175, 119, 200, 251, 103, 146,  14, 209, 174, 109, 218, 192,  82, 203, 163 },
+    {  29,  93, 150,  22, 166, 182,  55,  30,  90,  64,  42, 141, 168,  57, 117,  46 },
+    { 216, 233,  61, 128,  81, 237, 217, 118, 159, 255, 185,  27, 242, 102,   4, 133 },
+    {  73, 191,   9, 210,  43,  96,   7, 136, 231,  80,  10, 124, 225, 207, 155, 183 }
+  };
 
 
   memset(d, 0, sizeof(pappl_pr_driver_data_t));
   memcpy(d->gdither, clustered, sizeof(d->gdither));
-  memcpy(d->pdither, clustered, sizeof(d->pdither));
+  memcpy(d->pdither, blue, sizeof(d->pdither));
 
   d->orient_default      = IPP_ORIENT_NONE;
   d->content_default     = PAPPL_CONTENT_AUTO;
@@ -168,10 +187,11 @@ papplPrinterSetDriverData(
 
   // Copy driver data to printer
   memcpy(&printer->driver_data, data, sizeof(printer->driver_data));
+  printer->num_ready = data->num_source;
 
   // Create printer (capability) attributes based on driver data...
   ippDelete(printer->driver_attrs);
-  printer->driver_attrs = make_attrs(printer->system, &printer->driver_data);
+  printer->driver_attrs = make_attrs(printer->system, printer, &printer->driver_data);
 
   if (attrs)
     ippCopyAttributes(printer->driver_attrs, attrs, 0, NULL, NULL);
@@ -236,7 +256,7 @@ papplPrinterSetDriverDefaults(
   // Copy any vendor-specific xxx-default values...
   for (i = 0; i < data->num_vendor; i ++)
   {
-    if ((value = cupsGetOption(data->vendor[i], num_vendor, vendor)) == NULL)
+    if ((value = cupsGetOption(data->vendor[i], (cups_len_t)num_vendor, vendor)) == NULL)
       continue;
 
     snprintf(defname, sizeof(defname), "%s-default", data->vendor[i]);
@@ -297,6 +317,9 @@ papplPrinterSetReadyMedia(
     int               num_ready,	// I - Number of ready media
     pappl_media_col_t *ready)		// I - Array of ready media
 {
+  int	i;				// Looping vars
+
+
   if (!printer || num_ready <= 0 || !ready)
     return (false);
 
@@ -305,11 +328,24 @@ papplPrinterSetReadyMedia(
 
   pthread_rwlock_wrlock(&printer->rwlock);
 
-  if (num_ready > printer->driver_data.num_source)
-    num_ready = printer->driver_data.num_source;
+  // Copy new ready media to printer data...
+  if (num_ready > PAPPL_MAX_SOURCE)
+    num_ready = PAPPL_MAX_SOURCE;
 
   memset(printer->driver_data.media_ready, 0, sizeof(printer->driver_data.media_ready));
   memcpy(printer->driver_data.media_ready, ready, (size_t)num_ready * sizeof(pappl_media_col_t));
+  printer->num_ready = num_ready;
+
+  // Update default media from ready media...
+  for (i = 0; i < num_ready; i ++)
+  {
+    if (!strcmp(ready[i].source, printer->driver_data.media_default.source))
+    {
+      printer->driver_data.media_default = ready[i];
+      break;
+    }
+  }
+
   printer->state_time = time(NULL);
 
   pthread_rwlock_unlock(&printer->rwlock);
@@ -327,16 +363,22 @@ papplPrinterSetReadyMedia(
 static ipp_t *				// O - Driver attributes
 make_attrs(
     pappl_system_t         *system,	// I - System
+    pappl_printer_t        *printer,	// I - Printer
     pappl_pr_driver_data_t *data)	// I - Driver data
 {
   ipp_t			*attrs;		// Driver attributes
+  bool			pdf_supported;	// Is PDF supported?
   unsigned		bit;		// Current bit value
-  int			i, j,		// Looping vars
+  cups_len_t		i, j,		// Looping vars
 			num_values;	// Number of values
-  const char		*svalues[100];	// String values
-  int			ivalues[100];	// Integer values
+  const char		*svalues[PAPPL_MAX_MEDIA];
+					// String values
+  int			ivalues[PAPPL_MAX_MEDIA];
+					// Integer values
   ipp_t			*cvalues[PAPPL_MAX_MEDIA * 2];
 					// Collection values
+  char			vvalues[PAPPL_MAX_VENDOR][128];
+					// Vendor xxx-default values
   char			fn[32],		// FN (finishings) values
 			*ptr;		// Pointer into value
   const char		*preferred;	// "document-format-preferred" value
@@ -386,12 +428,18 @@ make_attrs(
     "media-size-name",
     "media-top-margin"
   };
+  static const char * const pdf_versions_supported[] =
+  {					// "pdf-versions-supported" values
+    "adobe-1.3",
+    "adobe-1.4",
+    "adobe-1.5",
+    "adobe-1.6",
+    "iso-32000-1_2008"			// PDF 1.7
+  };
   static const char * const printer_settable_attributes[] =
   {					// printer-settable-attributes values
     "copies-default",
     "document-format-default",
-    "label-mode-configured",
-    "label-tear-off-configured",
     "media-col-default",
     "media-col-ready",
     "media-default",
@@ -400,10 +448,7 @@ make_attrs(
     "orientation-requested-default",
     "print-color-mode-default",
     "print-content-optimize-default",
-    "print-darkness-default",
     "print-quality-default",
-    "print-speed-default",
-    "printer-darkness-configured",
     "printer-geo-location",
     "printer-location",
     "printer-organization",
@@ -412,6 +457,10 @@ make_attrs(
   };
 
 
+  // Is PDF supported?
+  pdf_supported = (data->format && !strcmp(data->format, "application/pdf")) ||   _papplSystemFindMIMEFilter(system, "application/pdf", "image/pwg-raster") != NULL || _papplSystemFindMIMEFilter(system, "application/pdf", data->format) != NULL;
+  papplLogPrinter(printer, PAPPL_LOGLEVEL_DEBUG, "pdf-supported=%s", pdf_supported ? "true" : "false");
+
   // Create an empty IPP message for the attributes...
   attrs = ippNew();
 
@@ -429,7 +478,7 @@ make_attrs(
   if (data->format && strcmp(data->format, "application/octet-stream"))
     svalues[num_values ++] = data->format;
 
-  for (preferred = "image/urf", filter = (_pappl_mime_filter_t *)cupsArrayFirst(system->filters); filter; filter = (_pappl_mime_filter_t *)cupsArrayNext(system->filters))
+  for (preferred = "image/urf", filter = (_pappl_mime_filter_t *)cupsArrayGetFirst(system->filters); filter; filter = (_pappl_mime_filter_t *)cupsArrayGetNext(system->filters))
   {
     if ((data->format && !strcmp(filter->dst, data->format)) || !strcmp(filter->dst, "image/pwg-raster"))
     {
@@ -461,7 +510,7 @@ make_attrs(
   ivalues[num_values   ] = IPP_FINISHINGS_NONE;
   svalues[num_values ++] = "none";
 
-  strlcpy(fn, "FN3", sizeof(fn));
+  papplCopyString(fn, "FN3", sizeof(fn));
   for (ptr = fn + 3, i = 0, bit = PAPPL_FINISHINGS_PUNCH; bit <= PAPPL_FINISHINGS_TRIM; i ++, bit *= 2)
   {
     if (data->finishings & bit)
@@ -511,10 +560,11 @@ make_attrs(
 
 
   // ipp-features-supported
-  num_values = data->num_features;
+  if ((num_values = (cups_len_t)data->num_features) > PAPPL_MAX_VENDOR)
+    num_values = PAPPL_MAX_VENDOR;
 
-  if (data->num_features > 0)
-    memcpy(svalues, data->features, (size_t)data->num_features * sizeof(char *));
+  if (num_values > 0)
+    memcpy((void *)svalues, data->features, (size_t)num_values * sizeof(char *));
 
   svalues[num_values ++] = "ipp-everywhere";
 
@@ -522,10 +572,10 @@ make_attrs(
 
 
   // job-creation-attributes-supported
-  memcpy(svalues, job_creation_attributes, sizeof(job_creation_attributes));
+  memcpy((void *)svalues, job_creation_attributes, sizeof(job_creation_attributes));
   num_values = (int)(sizeof(job_creation_attributes) / sizeof(job_creation_attributes[0]));
 
-  if (_papplSystemFindMIMEFilter(system, "application/pdf", "image/pwg-raster"))
+  if (pdf_supported)
     svalues[num_values ++] = "page-ranges";
 
   if (data->darkness_supported)
@@ -534,7 +584,7 @@ make_attrs(
   if (data->speed_supported[1])
     svalues[num_values ++] = "print-speed";
 
-  for (i = 0; i < data->num_vendor && i < (int)(sizeof(data->vendor) / sizeof(data->vendor[0])) && i < (int)(sizeof(svalues) / sizeof(svalues[0])); i ++)
+  for (i = 0; i < (cups_len_t)data->num_vendor && i < (cups_len_t)(sizeof(data->vendor) / sizeof(data->vendor[0])); i ++)
     svalues[num_values ++] = data->vendor[i];
 
   ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "job-creation-attributes-supported", num_values, NULL, svalues);
@@ -560,6 +610,10 @@ make_attrs(
   ippAddInteger(attrs, IPP_TAG_PRINTER, IPP_TAG_ENUM, "landscape-orientation-requested-preferred", IPP_ORIENT_LANDSCAPE);
 
 
+  // max-page-ranges-supported
+  if (pdf_supported)
+    ippAddInteger(attrs, IPP_TAG_PRINTER, IPP_TAG_INTEGER, "max-page-ranges-supported", 1);
+
   // media-bottom-margin-supported
   num_values = 0;
   if (data->borderless)
@@ -570,7 +624,7 @@ make_attrs(
 
 
   // media-col-database
-  for (i = 0, num_values = 0; i < data->num_media; i ++)
+  for (i = 0, num_values = 0; i < (cups_len_t)data->num_media; i ++)
   {
     if (!strncmp(data->media[i], "custom_max_", 11) || !strncmp(data->media[i], "roll_max_", 9))
     {
@@ -586,7 +640,7 @@ make_attrs(
       pwg_media_t	*pwg;		// PWG media size info
 
       memset(&col, 0, sizeof(col));
-      strlcpy(col.size_name, data->media[i], sizeof(col.size_name));
+      papplCopyString(col.size_name, data->media[i], sizeof(col.size_name));
       if ((pwg = pwgMediaForPWG(data->media[i])) != NULL)
       {
 	col.size_width  = pwg->width;
@@ -655,7 +709,7 @@ make_attrs(
 
 
   // media-col-supported
-  memcpy(svalues, media_col, sizeof(media_col));
+  memcpy((void *)svalues, media_col, sizeof(media_col));
   num_values = (int)(sizeof(media_col) / sizeof(media_col[0]));
 
   if (data->left_offset_supported[1])
@@ -700,7 +754,7 @@ make_attrs(
 
 
   // media-size-supported
-  for (i = 0, num_values = 0; i < data->num_media; i ++)
+  for (i = 0, num_values = 0; i < (cups_len_t)data->num_media; i ++)
   {
     pwg_media_t	*pwg;			// PWG media size info
 
@@ -750,13 +804,13 @@ make_attrs(
 
 
   // media-source-supported
-  if (data->num_source)
+  if ((num_values = (cups_len_t)data->num_source) > 0)
   {
-    memcpy(svalues, data->source, (size_t)data->num_source * sizeof(char *));
-    num_values = data->num_source;
+    if (num_values > PAPPL_MAX_SOURCE)
+      num_values = PAPPL_MAX_SOURCE;
+
+    memcpy((void *)svalues, data->source, (size_t)num_values * sizeof(char *));
   }
-  else
-    num_values = 0;
 
   svalues[num_values ++] = "auto";
 
@@ -764,7 +818,7 @@ make_attrs(
 
   // media-supported
   if (data->num_media)
-    ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "media-supported", data->num_media, NULL, data->media);
+    ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "media-supported", (cups_len_t)data->num_media, NULL, data->media);
 
 
   // media-top-margin-supported
@@ -796,7 +850,7 @@ make_attrs(
 
   // media-type-supported
   if (data->num_type)
-    ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "media-type-supported", data->num_type, NULL, data->type);
+    ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "media-type-supported", (cups_len_t)data->num_type, NULL, data->type);
 
 
   // mopria-certified (Mopria-specific attribute)
@@ -806,13 +860,18 @@ make_attrs(
 
   // output-bin-supported
   if (data->num_bin)
-    ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "output-bin-supported", data->num_bin, NULL, data->bin);
+    ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "output-bin-supported", (cups_len_t)data->num_bin, NULL, data->bin);
   else if (data->output_face_up)
     ippAddString(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "output-bin-supported", NULL, "face-up");
   else
     ippAddString(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "output-bin-supported", NULL, "face-down");
 
 
+  // page-ranges-supported
+  if (pdf_supported)
+    ippAddBoolean(attrs, IPP_TAG_PRINTER, "page-ranges-supported", 1);
+
+
   // pages-per-minute
   if (data->ppm > 0)
     ippAddInteger(attrs, IPP_TAG_PRINTER, IPP_TAG_INTEGER, "pages-per-minute", data->ppm);
@@ -825,6 +884,21 @@ make_attrs(
     ippAddInteger(attrs, IPP_TAG_PRINTER, IPP_TAG_INTEGER, "pages-per-minute-color", data->ppm_color);
 
 
+  // pdf-k-octets-supported
+  if (pdf_supported)
+  {
+    int upper = 0, lower = ippGetRange(ippFindAttribute(printer->attrs, "job-k-octets-supported", IPP_TAG_RANGE), 0, &upper);
+					// Range of k-octets-supported...
+
+    ippAddRange(attrs, IPP_TAG_PRINTER, "pdf-k-octets-supported", lower, upper);
+  }
+
+
+  // pdf-versions-supported
+  if (pdf_supported)
+    ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "pdf-versions-supported", (int)(sizeof(pdf_versions_supported) / sizeof(pdf_versions_supported[0])), NULL, pdf_versions_supported);
+
+
   // print-color-mode-supported
   for (num_values = 0, bit = PAPPL_COLOR_MODE_AUTO; bit <= PAPPL_COLOR_MODE_PROCESS_MONOCHROME; bit *= 2)
   {
@@ -869,39 +943,63 @@ make_attrs(
   // printer-output-tray
   if (data->num_bin > 0)
   {
-    for (i = 0, attr = NULL; i < data->num_bin; i ++)
+    for (i = 0, attr = NULL; i < (cups_len_t)data->num_bin; i ++)
     {
       snprintf(output_tray, sizeof(output_tray), "type=unRemovableBin;maxcapacity=-2;remaining=-2;status=0;name=%s;%s", data->bin[i], data->output_face_up ? "stackingorder=lastToFirst;pagedelivery=faceUp;" : "stackingorder=firstToLast;pagedelivery=faceDown;");
       if (attr)
-        ippSetOctetString(attrs, &attr, ippGetCount(attr), output_tray, (int)strlen(output_tray));
+        ippSetOctetString(attrs, &attr, ippGetCount(attr), output_tray, (cups_len_t)strlen(output_tray));
       else
-        attr = ippAddOctetString(attrs, IPP_TAG_PRINTER, "printer-output-tray", output_tray, (int)strlen(output_tray));
+        attr = ippAddOctetString(attrs, IPP_TAG_PRINTER, "printer-output-tray", output_tray, (cups_len_t)strlen(output_tray));
     }
   }
   else if (data->output_face_up)
   {
-    strlcpy(output_tray, "type=unRemovableBin;maxcapacity=-2;remaining=-2;status=0;name=face-up;stackingorder=lastToFirst;pagedelivery=faceUp;", sizeof(output_tray));
-    ippAddOctetString(attrs, IPP_TAG_PRINTER, "printer-output-tray", output_tray, (int)strlen(output_tray));
+    papplCopyString(output_tray, "type=unRemovableBin;maxcapacity=-2;remaining=-2;status=0;name=face-up;stackingorder=lastToFirst;pagedelivery=faceUp;", sizeof(output_tray));
+    ippAddOctetString(attrs, IPP_TAG_PRINTER, "printer-output-tray", output_tray, (cups_len_t)strlen(output_tray));
   }
   else
   {
-    strlcpy(output_tray, "type=unRemovableBin;maxcapacity=-2;remaining=-2;status=0;name=face-down;stackingorder=firstToLast;pagedelivery=faceDown;", sizeof(output_tray));
-    ippAddOctetString(attrs, IPP_TAG_PRINTER, "printer-output-tray", output_tray, (int)strlen(output_tray));
+    papplCopyString(output_tray, "type=unRemovableBin;maxcapacity=-2;remaining=-2;status=0;name=face-down;stackingorder=firstToLast;pagedelivery=faceDown;", sizeof(output_tray));
+    ippAddOctetString(attrs, IPP_TAG_PRINTER, "printer-output-tray", output_tray, (cups_len_t)strlen(output_tray));
   }
 
 
   // printer-resolution-supported
   if (data->num_resolution > 0)
-    ippAddResolutions(attrs, IPP_TAG_PRINTER, "printer-resolution-supported", data->num_resolution, IPP_RES_PER_INCH, data->x_resolution, data->y_resolution);
+    ippAddResolutions(attrs, IPP_TAG_PRINTER, "printer-resolution-supported", (cups_len_t)data->num_resolution, IPP_RES_PER_INCH, data->x_resolution, data->y_resolution);
 
 
   // printer-settable-attributes
-  ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_TAG_KEYWORD, "printer-settable-attributes", (int)(sizeof(printer_settable_attributes) / sizeof(printer_settable_attributes[0])), NULL, printer_settable_attributes);
+  memcpy((void *)svalues, printer_settable_attributes, sizeof(printer_settable_attributes));
+  num_values = (int)(sizeof(printer_settable_attributes) / sizeof(printer_settable_attributes[0]));
+
+  if (data->mode_supported)
+    svalues[num_values ++] = "label-mode-configured";
+  if (data->tear_offset_supported[1])
+    svalues[num_values ++] = "label-tear-off-configured";
+  if (data->darkness_supported)
+    svalues[num_values ++] = "print-darkness-default";
+  if (data->speed_supported[1])
+    svalues[num_values ++] = "print-speed-default";
+  if (data->darkness_supported)
+    svalues[num_values ++] = "printer-darkness-configured";
+  if (system->wifi_join_cb)
+  {
+    svalues[num_values ++] = "printer-wifi-password";
+    svalues[num_values ++] = "printer-wifi-ssid";
+  }
+  for (i = 0; i < (cups_len_t)data->num_vendor && num_values < (int)(sizeof(svalues) / sizeof(svalues[0])); i ++)
+  {
+    snprintf(vvalues[i], sizeof(vvalues[0]), "%s-default", data->vendor[i]);
+    svalues[num_values ++] = vvalues[i];
+  }
+
+  ippAddStrings(attrs, IPP_TAG_PRINTER, IPP_TAG_KEYWORD, "printer-settable-attributes", num_values, NULL, svalues);
 
 
   // pwg-raster-document-resolution-supported
   if (data->num_resolution > 0)
-    ippAddResolutions(attrs, IPP_TAG_PRINTER, "pwg-raster-document-resolution-supported", data->num_resolution, IPP_RES_PER_INCH, data->x_resolution, data->y_resolution);
+    ippAddResolutions(attrs, IPP_TAG_PRINTER, "pwg-raster-document-resolution-supported", (cups_len_t)data->num_resolution, IPP_RES_PER_INCH, data->x_resolution, data->y_resolution);
 
 
   // pwg-raster-document-sheet-back
@@ -954,7 +1052,7 @@ make_attrs(
 		rs[32];			// RS (resolution) values
 
     num_values = 0;
-    svalues[num_values ++] = "V1.4";
+    svalues[num_values ++] = "V1.5";
     svalues[num_values ++] = "W8";
     if (data->raster_types & PAPPL_PWG_RASTER_TYPE_SRGB_8)
       svalues[num_values ++] = "SRGB24";
@@ -1042,13 +1140,13 @@ make_attrs(
         "roll-10"
       };
 
-      for (i = 0, ptr = is, *ptr = '\0', prefix = "IS"; i < data->num_source; i ++)
+      for (i = 0, ptr = is, *ptr = '\0', prefix = "IS"; i < (cups_len_t)data->num_source; i ++)
       {
         for (j = 0; j < (int)(sizeof(iss) / sizeof(iss[0])); j ++)
         {
           if (!strcmp(iss[j], data->source[i]))
           {
-            snprintf(ptr, sizeof(is) - (size_t)(ptr - is), "%s%d", prefix, j);
+            snprintf(ptr, sizeof(is) - (size_t)(ptr - is), "%s%u", prefix, (unsigned)j);
             ptr += strlen(ptr);
             prefix = "-";
           }
@@ -1079,13 +1177,13 @@ make_attrs(
         "other"
       };
 
-      for (i = 0, ptr = mt, *ptr = '\0', prefix = "MT"; i < data->num_type; i ++)
+      for (i = 0, ptr = mt, *ptr = '\0', prefix = "MT"; i < (cups_len_t)data->num_type; i ++)
       {
         for (j = 0; j < (int)(sizeof(mts) / sizeof(mts[0])); j ++)
         {
           if (!strcmp(mts[j], data->type[i]))
           {
-            snprintf(ptr, sizeof(mt) - (size_t)(ptr - mt), "%s%d", prefix, j);
+            snprintf(ptr, sizeof(mt) - (size_t)(ptr - mt), "%s%u", prefix, (unsigned)j);
             ptr += strlen(ptr);
             prefix = "-";
           }
@@ -1146,13 +1244,13 @@ make_attrs(
         "tray-10"
       };
 
-      for (i = 0, ptr = ob, *ptr = '\0', prefix = "OB"; i < data->num_bin; i ++)
+      for (i = 0, ptr = ob, *ptr = '\0', prefix = "OB"; i < (cups_len_t)data->num_bin; i ++)
       {
         for (j = 0; j < (int)(sizeof(obs) / sizeof(obs[0])); j ++)
         {
           if (!strcmp(obs[j], data->bin[i]))
           {
-            snprintf(ptr, sizeof(ob) - (size_t)(ptr - ob), "%s%d", prefix, j);
+            snprintf(ptr, sizeof(ob) - (size_t)(ptr - ob), "%s%u", prefix, (unsigned)j);
             ptr += strlen(ptr);
             prefix = "-";
           }
@@ -1477,7 +1575,7 @@ validate_ready(
   pwg_media_t	*pwg;			// PWG media size
 
 
-  if (num_ready > driver_data->num_source)
+  if (num_ready > PAPPL_MAX_SOURCE)
     return (false);
 
   // Determine the range of media sizes...
