@@ -1,7 +1,7 @@
 //
 // Private printer header file for the Printer Application Framework
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -10,19 +10,13 @@
 
 #ifndef _PAPPL_PRINTER_PRIVATE_H_
 #  define _PAPPL_PRINTER_PRIVATE_H_
-
-//
-// Include necessary headers...
-//
-
 #  include "dnssd-private.h"
 #  include "printer.h"
 #  include "log.h"
-#  include <grp.h>
 #  ifdef __APPLE__
 #    include <sys/param.h>
 #    include <sys/mount.h>
-#  else
+#  elif !_WIN32
 #    include <sys/statfs.h>
 #  endif // __APPLE__
 #  ifdef HAVE_SYS_RANDOM_H
@@ -32,12 +26,6 @@
 #    include <gnutls/gnutls.h>
 #    include <gnutls/crypto.h>
 #  endif // HAVE_GNUTLS_RND
-
-
-//
-// Include necessary headers...
-//
-
 #  include "base-private.h"
 #  include "device.h"
 
@@ -64,15 +52,17 @@ struct _pappl_printer_s			// Printer data
   ipp_pstate_t		state;			// "printer-state" value
   pappl_preason_t	state_reasons;		// "printer-state-reasons" values
   time_t		state_time;		// "printer-state-change-time" value
-  bool			is_stopped,		// Are we stopping this printer?
+  bool			is_accepting,		// Are we accepting jobs?
+			is_stopped,		// Are we stopping this printer?
 			is_deleted;		// Has this printer been deleted?
   char			*device_id,		// "printer-device-id" value
 			*device_uri;		// Device URI
   pappl_device_t	*device;		// Current connection to device (if any)
   bool			device_in_use;		// Is the device in use?
   char			*driver_name;		// Driver name
-  pappl_pr_driver_data_t driver_data;	// Driver data
+  pappl_pr_driver_data_t driver_data;		// Driver data
   ipp_t			*driver_attrs;		// Driver attributes
+  int			num_ready;		// Number of ready media
   ipp_t			*attrs;			// Other (static) printer attributes
   time_t		start_time;		// Startup time
   time_t		config_time;		// "printer-config-change-time" value
@@ -84,14 +74,15 @@ struct _pappl_printer_s			// Printer data
 						// "printer-supply" values
   pappl_job_t		*processing_job;	// Currently printing job, if any
   int			max_active_jobs,	// Maximum number of active jobs to accept
-			max_completed_jobs;	// Maximum number of completed jobs to retain in history
+			max_completed_jobs,	// Maximum number of completed jobs to retain in history
+			max_preserved_jobs;	// Maximum number of completed jobs to preserve in history
   cups_array_t		*active_jobs,		// Array of active jobs
 			*all_jobs,		// Array of all jobs
 			*completed_jobs;	// Array of completed jobs
   int			next_job_id,		// Next "job-id" value
 			impcompleted;		// "printer-impressions-completed" value
   cups_array_t		*links;			// Web navigation links
-#  ifdef HAVE_DNSSD
+#  ifdef HAVE_MDNSRESPONDER
   _pappl_srv_t		dns_sd_ipp_ref,		// DNS-SD IPP service
 			dns_sd_ipps_ref,	// DNS-SD IPPS service
 			dns_sd_http_ref,	// DNS-SD HTTP service
@@ -101,7 +92,7 @@ struct _pappl_printer_s			// Printer data
 			dns_sd_ipps_loc_ref;	// DNS-SD LOC record for IPPS service
 #  elif defined(HAVE_AVAHI)
   _pappl_srv_t		dns_sd_ref;		// DNS-SD services
-#  endif // HAVE_DNSSD
+#  endif // HAVE_MDNSRESPONDER
   unsigned char		dns_sd_loc[16];		// DNS-SD LOC record data
   bool			dns_sd_collision;	// Was there a name collision?
   int			dns_sd_serial;		// DNS-SD serial number (for collisions)
@@ -113,6 +104,8 @@ struct _pappl_printer_s			// Printer data
 			usb_product_id;		// USB product ID
   pappl_uoptions_t	usb_options;		// USB gadget options
   char			*usb_storage;		// USB storage gadget file, if any
+  pappl_pr_usb_cb_t	usb_cb;			// USB processing callback, if any
+  void			*usb_cbdata;		// USB processing callback data, if any
 };
 
 
@@ -126,12 +119,13 @@ extern void		*_papplPrinterRunRaw(pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		*_papplPrinterRunUSB(pappl_printer_t *printer) _PAPPL_PRIVATE;
 
 extern void		_papplPrinterCheckJobs(pappl_printer_t *printer) _PAPPL_PRIVATE;
-extern void		_papplPrinterCleanJobs(pappl_printer_t *printer) _PAPPL_PRIVATE;
-extern void		_papplPrinterCopyAttributes(pappl_client_t *client, pappl_printer_t *printer, cups_array_t *ra, const char *format) _PAPPL_PRIVATE;
-extern void		_papplPrinterCopyState(ipp_t *ipp, pappl_printer_t *printer, cups_array_t *ra) _PAPPL_PRIVATE;
-extern void		_papplPrinterCopyXRI(pappl_client_t *client, ipp_t *ipp, pappl_printer_t *printer) _PAPPL_PRIVATE;
+extern void		_papplPrinterCleanJobsNoLock(pappl_printer_t *printer) _PAPPL_PRIVATE;
+extern void		_papplPrinterCopyAttributes(pappl_printer_t *printer, pappl_client_t *client, cups_array_t *ra, const char *format) _PAPPL_PRIVATE;
+extern void		_papplPrinterCopyState(pappl_printer_t *printer, ipp_tag_t group_tag, ipp_t *ipp, pappl_client_t *client, cups_array_t *ra) _PAPPL_PRIVATE;
+extern void		_papplPrinterCopyXRI(pappl_printer_t *printer, ipp_t *ipp, pappl_client_t *client) _PAPPL_PRIVATE;
 extern void		_papplPrinterDelete(pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterInitDriverData(pappl_pr_driver_data_t *d) _PAPPL_PRIVATE;
+extern bool		_papplPrinterIsAuthorized(pappl_client_t *client) _PAPPL_PRIVATE;
 extern void		_papplPrinterProcessIPP(pappl_client_t *client) _PAPPL_PRIVATE;
 extern bool		_papplPrinterRegisterDNSSDNoLock(pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern bool		_papplPrinterSetAttributes(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
@@ -140,13 +134,14 @@ extern void		_papplPrinterUnregisterDNSSDNoLock(pappl_printer_t *printer) _PAPPL
 extern void		_papplPrinterWebCancelAllJobs(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebCancelJob(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebConfig(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
-extern void		_papplPrinterWebConfigFinalize(pappl_printer_t *printer, int num_form, cups_option_t *form) _PAPPL_PRIVATE;
+extern void		_papplPrinterWebConfigFinalize(pappl_printer_t *printer, cups_len_t num_form, cups_option_t *form) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebDefaults(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebDelete(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebHome(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebIteratorCallback(pappl_printer_t *printer, pappl_client_t *client) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebJobs(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebMedia(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
+extern void		_papplPrinterWebReprintJob(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 extern void		_papplPrinterWebSupplies(pappl_client_t *client, pappl_printer_t *printer) _PAPPL_PRIVATE;
 
 extern const char	*_papplColorModeString(pappl_color_mode_t value) _PAPPL_PRIVATE;
@@ -185,7 +180,9 @@ extern const char	*_papplSidesString(pappl_sides_t value) _PAPPL_PRIVATE;
 extern pappl_sides_t	_papplSidesValue(const char *value) _PAPPL_PRIVATE;
 
 extern const char	*_papplSupplyColorString(pappl_supply_color_t value) _PAPPL_PRIVATE;
+extern pappl_supply_color_t _papplSupplyColorValue(const char *value) _PAPPL_PRIVATE;
 extern const char	*_papplSupplyTypeString(pappl_supply_type_t value) _PAPPL_PRIVATE;
+extern pappl_supply_type_t _papplSupplyTypeValue(const char *value) _PAPPL_PRIVATE;
 
 
 #endif // !_PAPPL_PRINTER_PRIVATE_H_
