@@ -90,8 +90,8 @@ _papplPrinterRunRaw(
   while (!printer->is_deleted && printer->system->is_running)
   {
     // Don't accept connections if we can't accept a new job...
-    while (cupsArrayCount(printer->active_jobs) >= printer->max_active_jobs && !printer->is_deleted && printer->system->is_running)
-      sleep(1);
+    while ((int)cupsArrayGetCount(printer->active_jobs) >= printer->max_active_jobs && !printer->is_deleted && printer->system->is_running)
+      usleep(100000);
 
     if (printer->is_deleted || !printer->system->is_running)
       break;
@@ -119,7 +119,7 @@ _papplPrinterRunRaw(
 
           // Accept the connection...
           sockaddrlen = sizeof(sockaddr);
-          if ((sock = accept(printer->raw_listeners[i].fd, (struct sockaddr *)&sockaddr, &sockaddrlen)) < 0)
+          if ((sock = (int)accept(printer->raw_listeners[i].fd, (struct sockaddr *)&sockaddr, &sockaddrlen)) < 0)
           {
             papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to accept socket print connection: %s", strerror(errno));
             continue;
@@ -141,6 +141,12 @@ _papplPrinterRunRaw(
 	    goto abort_job;
 	  }
 
+	  if ((job->filename = strdup(filename)) == NULL)
+	  {
+            unlink(filename);
+	    goto abort_job;
+	  }
+
 	  papplLogJob(job, PAPPL_LOGLEVEL_DEBUG, "Created job file \"%s\", format \"%s\".", filename, job->format);
 
           activity     = time(NULL);
@@ -167,7 +173,7 @@ _papplPrinterRunRaw(
 
             if (sockp.revents & POLLIN)
             {
-              if ((bytes = read(sock, buffer, sizeof(buffer))) > 0)
+              if ((bytes = recv(sock, buffer, sizeof(buffer), 0)) > 0)
                 write(job->fd, buffer, (size_t)bytes);
               else
                 break;
@@ -186,17 +192,10 @@ _papplPrinterRunRaw(
           if (bytes < 0)
           {
             // Error while reading
-	    unlink(filename);
 	    goto abort_job;
 	  }
 
 	  // Finish the job...
-	  if ((job->filename = strdup(filename)) == NULL)
-	  {
-            unlink(filename);
-	    goto abort_job;
-	  }
-
 	  job->state = IPP_JSTATE_PENDING;
 
 	  _papplPrinterCheckJobs(printer);
