@@ -1,7 +1,7 @@
 //
 // Printer object for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -41,10 +41,10 @@ papplPrinterCancelAllJobs(
 
   // Loop through all jobs and cancel them.
   //
-  // Since we have a writer lock, it is safe to use cupsArrayFirst/Last...
+  // Since we have a writer lock, it is safe to use cupsArrayGetFirst/Last...
   pthread_rwlock_wrlock(&printer->rwlock);
 
-  for (job = (pappl_job_t *)cupsArrayFirst(printer->active_jobs); job; job = (pappl_job_t *)cupsArrayNext(printer->active_jobs))
+  for (job = (pappl_job_t *)cupsArrayGetFirst(printer->active_jobs); job; job = (pappl_job_t *)cupsArrayGetNext(printer->active_jobs))
   {
     // Cancel this job...
     if (job->state == IPP_JSTATE_PROCESSING || (job->state == IPP_JSTATE_HELD && job->fd >= 0))
@@ -114,8 +114,10 @@ papplPrinterCreate(
 			uuid[128],	// printer-uuid
 			print_group[65];// print-group value
   int			k_supported;	// Maximum file size supported
+#if !_WIN32
   struct statfs		spoolinfo;	// FS info for spool directory
   double		spoolsize;	// FS size
+#endif // !_WIN32
   char			path[256];	// Path to resource
   pappl_pr_driver_data_t driver_data;	// Driver data
   ipp_t			*driver_attrs;	// Driver attributes
@@ -137,6 +139,19 @@ papplPrinterCreate(
     IPP_OP_PAUSE_PRINTER,
     IPP_OP_RESUME_PRINTER,
     IPP_OP_SET_PRINTER_ATTRIBUTES,
+    IPP_OP_GET_PRINTER_SUPPORTED_VALUES,
+    IPP_OP_CREATE_PRINTER_SUBSCRIPTIONS,
+    IPP_OP_CREATE_JOB_SUBSCRIPTIONS,
+    IPP_OP_GET_SUBSCRIPTION_ATTRIBUTES,
+    IPP_OP_GET_SUBSCRIPTIONS,
+    IPP_OP_RENEW_SUBSCRIPTION,
+    IPP_OP_CANCEL_SUBSCRIPTION,
+    IPP_OP_GET_NOTIFICATIONS,
+    IPP_OP_ENABLE_PRINTER,
+    IPP_OP_DISABLE_PRINTER,
+    IPP_OP_PAUSE_PRINTER_AFTER_CURRENT_JOB,
+    IPP_OP_CANCEL_CURRENT_JOB,
+    IPP_OP_CANCEL_JOBS,
     IPP_OP_CANCEL_MY_JOBS,
     IPP_OP_CLOSE_JOB,
     IPP_OP_IDENTIFY_PRINTER
@@ -245,7 +260,7 @@ papplPrinterCreate(
     }
   }
   else
-    strlcpy(resource, "/ipp/print", sizeof(resource));
+    papplCopyString(resource, "/ipp/print", sizeof(resource));
 
   // Make sure the printer doesn't already exist...
   if ((printer = papplSystemFindPrinter(system, resource, 0, NULL)) != NULL)
@@ -274,7 +289,7 @@ papplPrinterCreate(
       return (NULL);
     }
 
-    strlcpy(resource, temp, sizeof(resource));
+    papplCopyString(resource, temp, sizeof(resource));
   }
 
   // Allocate memory for the printer...
@@ -291,12 +306,16 @@ papplPrinterCreate(
   // Get the maximum spool size based on the size of the filesystem used for
   // the spool directory.  If the host OS doesn't support the statfs call
   // or the filesystem is larger than 2TiB, always report INT_MAX.
+#if _WIN32
+  k_supported = INT_MAX;
+#else // !_WIN32
   if (statfs(system->directory, &spoolinfo))
     k_supported = INT_MAX;
   else if ((spoolsize = (double)spoolinfo.f_bsize * spoolinfo.f_blocks / 1024) > INT_MAX)
     k_supported = INT_MAX;
   else
     k_supported = (int)spoolsize;
+#endif // _WIN32
 
   // Initialize printer structure and attributes...
   pthread_rwlock_init(&printer->rwlock, NULL);
@@ -316,9 +335,10 @@ papplPrinterCreate(
   printer->state              = IPP_PSTATE_IDLE;
   printer->state_reasons      = PAPPL_PREASON_NONE;
   printer->state_time         = printer->start_time;
-  printer->all_jobs           = cupsArrayNew3((cups_array_func_t)compare_all_jobs, NULL, NULL, 0, NULL, (cups_afree_func_t)_papplJobDelete);
-  printer->active_jobs        = cupsArrayNew((cups_array_func_t)compare_active_jobs, NULL);
-  printer->completed_jobs     = cupsArrayNew((cups_array_func_t)compare_completed_jobs, NULL);
+  printer->is_accepting       = true;
+  printer->all_jobs           = cupsArrayNew((cups_array_cb_t)compare_all_jobs, NULL, NULL, 0, NULL, (cups_afree_cb_t)_papplJobDelete);
+  printer->active_jobs        = cupsArrayNew((cups_array_cb_t)compare_active_jobs, NULL, NULL, 0, NULL, NULL);
+  printer->completed_jobs     = cupsArrayNew((cups_array_cb_t)compare_completed_jobs, NULL, NULL, 0, NULL, NULL);
   printer->next_job_id        = 1;
   printer->max_active_jobs    = (system->options & PAPPL_SOPTIONS_MULTI_QUEUE) ? 0 : 1;
   printer->max_completed_jobs = 100;
@@ -385,11 +405,11 @@ papplPrinterCreate(
 		cmd[128],		// Command (format) list
 		*ptr;			// Pointer into string
     ipp_attribute_t *formats;		// "document-format-supported" attribute
-    int		i,			// Looping var
+    cups_len_t	i,			// Looping var
 		count;			// Number of values
 
     // Assume make and model are separated by a space...
-    strlcpy(mfg, driver_data.make_and_model, sizeof(mfg));
+    papplCopyString(mfg, driver_data.make_and_model, sizeof(mfg));
     if ((mdl = strchr(mfg, ' ')) != NULL)
       *mdl++ = '\0';			// Nul-terminate the make
     else
@@ -426,7 +446,7 @@ papplPrinterCreate(
       if (ptr > cmd)
         snprintf(ptr, sizeof(cmd) - (size_t)(ptr - cmd), ",%s", format);
       else
-        strlcpy(cmd, format, sizeof(cmd));
+        papplCopyString(cmd, format, sizeof(cmd));
 
       ptr += strlen(ptr);
     }
@@ -520,36 +540,30 @@ papplPrinterCreate(
   // natural-language-configured
   ippAddString(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_LANGUAGE), "natural-language-configured", NULL, "en");
 
+  // notify-events-default
+  ippAddString(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "notify-events-default", NULL, "job-completed");
+
+  // notify-events-supported
+  ippAddStrings(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "notify-events-supported", (int)(sizeof(_papplEvents) / sizeof(_papplEvents[0])), NULL, _papplEvents);
+
+  // notify-lease-duration-default
+  ippAddInteger(printer->attrs, IPP_TAG_PRINTER, IPP_TAG_INTEGER, "notify-lease-duration-default", PAPPL_LEASE_DEFAULT);
+
+  // notify-lease-duration-supported
+  ippAddRange(printer->attrs, IPP_TAG_PRINTER, "notify-lease-duration-supported", 0, PAPPL_LEASE_MAX);
+
+  // notify-max-events-supported
+  ippAddInteger(printer->attrs, IPP_TAG_PRINTER, IPP_TAG_INTEGER, "notify-max-events-supported", PAPPL_MAX_EVENTS);
+
+  // notify-pull-method-supported
+  ippAddString(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "notify-pull-method-supported", NULL, "ippget");
+
   // operations-supported
   ippAddIntegers(printer->attrs, IPP_TAG_PRINTER, IPP_TAG_ENUM, "operations-supported", (int)(sizeof(operations) / sizeof(operations[0])), operations);
 
   // orientation-requested-supported
   ippAddIntegers(printer->attrs, IPP_TAG_PRINTER, IPP_TAG_ENUM, "orientation-requested-supported", (int)(sizeof(orientation_requested) / sizeof(orientation_requested[0])), orientation_requested);
 
-  if (_papplSystemFindMIMEFilter(system, "application/pdf", "image/pwg-raster"))
-  {
-    static const char * const pdf_versions_supported[] =
-    {					// "pdf-versions-supported" values
-      "adobe-1.3",
-      "adobe-1.4",
-      "adobe-1.5",
-      "adobe-1.6",
-      "iso-32000-1_2008"		// PDF 1.7
-    };
-
-    // max-page-ranges-supported
-    ippAddInteger(printer->attrs, IPP_TAG_PRINTER, IPP_TAG_INTEGER, "max-page-ranges-supported", 1);
-
-    // page-ranges-supported
-    ippAddBoolean(printer->attrs, IPP_TAG_PRINTER, "page-ranges-supported", 1);
-
-    // pdf-k-octets-supported
-    ippAddRange(printer->attrs, IPP_TAG_PRINTER, "pdf-k-octets-supported", 0, k_supported);
-
-    // pdf-versions-supported
-    ippAddStrings(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "pdf-versions-supported", (int)(sizeof(pdf_versions_supported) / sizeof(pdf_versions_supported[0])), NULL, pdf_versions_supported);
-  }
-
   // pdl-override-supported
   ippAddString(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "pdl-override-supported", NULL, "attempted");
 
@@ -582,7 +596,9 @@ papplPrinterCreate(
   ippAddString(printer->attrs, IPP_TAG_PRINTER, IPP_TAG_URI, "printer-uuid", NULL, uuid);
 
   // uri-security-supported
-  if (papplSystemGetTLSOnly(system))
+  if (system->options & PAPPL_SOPTIONS_NO_TLS)
+    ippAddString(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "uri-security-supported", NULL, "none");
+  else if (papplSystemGetTLSOnly(system))
     ippAddString(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "uri-security-supported", NULL, "tls");
   else
     ippAddStrings(printer->attrs, IPP_TAG_PRINTER, IPP_CONST_TAG(IPP_TAG_KEYWORD), "uri-security-supported", 2, NULL, uri_security);
@@ -613,6 +629,9 @@ papplPrinterCreate(
       {
 	// Detach the main thread from the raw thread to prevent hangs...
 	pthread_detach(tid);
+
+	while (!printer->raw_active)
+	  usleep(1000);			// Wait for raw thread to start
       }
     }
   }
@@ -646,17 +665,20 @@ papplPrinterCreate(
 
     snprintf(path, sizeof(path), "%s/media", printer->uriname);
     papplSystemAddResourceCallback(system, path, "text/html", (pappl_resource_cb_t)_papplPrinterWebMedia, printer);
-    papplPrinterAddLink(printer, "Media", path, PAPPL_LOPTIONS_NAVIGATION | PAPPL_LOPTIONS_STATUS);
+    papplPrinterAddLink(printer, _PAPPL_LOC("Media"), path, PAPPL_LOPTIONS_NAVIGATION | PAPPL_LOPTIONS_STATUS);
 
     snprintf(path, sizeof(path), "%s/printing", printer->uriname);
     papplSystemAddResourceCallback(system, path, "text/html", (pappl_resource_cb_t)_papplPrinterWebDefaults, printer);
-    papplPrinterAddLink(printer, "Printing Defaults", path, PAPPL_LOPTIONS_NAVIGATION | PAPPL_LOPTIONS_STATUS);
+    papplPrinterAddLink(printer, _PAPPL_LOC("Printing Defaults"), path, PAPPL_LOPTIONS_NAVIGATION | PAPPL_LOPTIONS_STATUS);
+
+    snprintf(path, sizeof(path), "%s/reprint", printer->uriname);
+    papplSystemAddResourceCallback(system, path, "text/html", (pappl_resource_cb_t)_papplPrinterWebReprintJob, printer);
 
     if (printer->driver_data.has_supplies)
     {
       snprintf(path, sizeof(path), "%s/supplies", printer->uriname);
       papplSystemAddResourceCallback(system, path, "text/html", (pappl_resource_cb_t)_papplPrinterWebSupplies, printer);
-      papplPrinterAddLink(printer, "Supplies", path, PAPPL_LOPTIONS_STATUS);
+      papplPrinterAddLink(printer, _PAPPL_LOC("Supplies"), path, PAPPL_LOPTIONS_STATUS);
     }
   }
 
@@ -693,7 +715,11 @@ _papplPrinterDelete(
   // Close raw listener sockets...
   for (i = 0; i < printer->num_raw_listeners; i ++)
   {
+#if _WIN32
+    closesocket(printer->raw_listeners[i].fd);
+#else
     close(printer->raw_listeners[i].fd);
+#endif // _WIN32
     printer->raw_listeners[i].fd = -1;
   }
 
@@ -708,8 +734,8 @@ _papplPrinterDelete(
 
   // Note: System writer lock is already held when calling cupsArrayRemove
   // for the system's printer object, so we don't need a separate lock here
-  // and can safely use cupsArrayFirst/Next...
-  for (r = (_pappl_resource_t *)cupsArrayFirst(printer->system->resources); r; r = (_pappl_resource_t *)cupsArrayNext(printer->system->resources))
+  // and can safely use cupsArrayGetFirst/Next...
+  for (r = (_pappl_resource_t *)cupsArrayGetFirst(printer->system->resources); r; r = (_pappl_resource_t *)cupsArrayGetNext(printer->system->resources))
   {
     if (r->cbdata == printer || !strncmp(r->path, prefix, prefixlen))
       cupsArrayRemove(printer->system->resources, r);
@@ -760,6 +786,10 @@ papplPrinterDelete(
   pappl_system_t *system = printer->system;
 					// System
 
+
+  // Deliver delete event...
+  papplSystemAddEvent(system, printer, NULL, PAPPL_EVENT_PRINTER_DELETED | PAPPL_EVENT_SYSTEM_CONFIG_CHANGED, NULL);
+
   // Remove the printer from the system object...
   pthread_rwlock_wrlock(&system->rwlock);
   cupsArrayRemove(system->printers, printer);
@@ -769,6 +799,100 @@ papplPrinterDelete(
 }
 
 
+//
+// 'papplPrinterOpenFile()' - Create or open a file for a printer.
+//
+// This function creates or opens a file for a printer.  The "fname" and
+// "fnamesize" arguments specify the location and size of a buffer to store the
+// printer filename, which incorporates the "directory", printer ID, resource
+// name, and "ext" values.  The job name is "sanitized" to only contain
+// alphanumeric characters.
+//
+// The "mode" argument is "r" to read an existing job file or "w" to write a
+// new job file.  New files are created with restricted permissions for
+// security purposes.
+//
+
+int					// O - File descriptor or -1 on error
+papplPrinterOpenFile(
+    pappl_printer_t *printer,		// I - Printer
+    char            *fname,		// I - Filename buffer
+    size_t          fnamesize,		// I - Size of filename buffer
+    const char      *directory,		// I - Directory to store in (`NULL` for default)
+    const char      *resname,		// I - Resource name
+    const char      *ext,		// I - Extension (`NULL` for none)
+    const char      *mode)		// I - Open mode - "r" for reading or "w" for writing
+{
+  char	name[64],			// "Safe" filename
+	*nameptr;			// Pointer into filename
+
+
+  // Range check input...
+  if (!printer || !fname || fnamesize < 256 || !resname || !mode)
+  {
+    if (fname)
+      *fname = '\0';
+
+    return (-1);
+  }
+
+  // Make sure the spool directory exists...
+  if (!directory)
+    directory = printer->system->directory;
+
+  if (access(directory, X_OK))
+  {
+    if (errno == ENOENT)
+    {
+      // Spool directory does not exist, might have been deleted...
+      if (mkdir(directory, 0777))
+      {
+        papplLogPrinter(printer, PAPPL_LOGLEVEL_FATAL, "Unable to create spool directory '%s': %s", directory, strerror(errno));
+        return (-1);
+      }
+    }
+    else
+    {
+      papplLogPrinter(printer, PAPPL_LOGLEVEL_FATAL, "Unable to access spool directory '%s': %s", directory, strerror(errno));
+      return (-1);
+    }
+  }
+
+  // Make a name from the resource name argument...
+  for (nameptr = name; *resname && nameptr < (name + sizeof(name) - 1); resname ++)
+  {
+    if (isalnum(*resname & 255) || *resname == '-' || *resname == '.')
+    {
+      *nameptr++ = (char)tolower(*resname & 255);
+    }
+    else
+    {
+      *nameptr++ = '_';
+
+      while (resname[1] && !isalnum(resname[1] & 255) && resname[1] != '-' && resname[1] != '.')
+        resname ++;
+    }
+  }
+
+  *nameptr = '\0';
+
+  // Create a filename...
+  if (ext)
+    snprintf(fname, fnamesize, "%s/p%05d-%s.%s", directory, printer->printer_id, name, ext);
+  else
+    snprintf(fname, fnamesize, "%s/p%05d-%s", directory, printer->printer_id, name);
+
+  if (!strcmp(mode, "r"))
+    return (open(fname, O_RDONLY | O_NOFOLLOW | O_CLOEXEC | O_BINARY));
+  else if (!strcmp(mode, "w"))
+    return (open(fname, O_WRONLY | O_CREAT | O_TRUNC | O_NOFOLLOW | O_CLOEXEC | O_BINARY, 0600));
+  else if (!strcmp(mode, "x"))
+    return (unlink(fname));
+  else
+    return (-1);
+}
+
+
 //
 // 'compare_active_jobs()' - Compare two active jobs.
 //
