@@ -1,7 +1,7 @@
 //
 // SNMP functions for the Printer Application Framework.
 //
-// Copyright © 2020 by Michael R Sweet.
+// Copyright © 2020-2022 by Michael R Sweet.
 // Copyright © 2007-2019 by Apple Inc.
 // Copyright © 2006-2007 by Easy Software Products, all rights reserved.
 //
@@ -71,7 +71,7 @@ _papplSNMPCopyOID(int       *dst,	// I - Destination OID
   int	i;				// Looping var
 
 
-  for (i = 0, dstsize --; src[i] >= 0 && i < dstsize; i ++)
+  for (i = 0, dstsize --; i < dstsize && src[i] >= 0; i ++)
     dst[i] = src[i];
 
   dst[i] = -1;
@@ -178,13 +178,13 @@ _papplSNMPOpen(int family)		// I - Address family - @code AF_INET@ or @code AF_I
 
 
   // Create the SNMP socket...
-  if ((fd = socket(family, SOCK_DGRAM, 0)) < 0)
+  if ((fd = (int)socket(family, SOCK_DGRAM, 0)) < 0)
     return (-1);
 
   // Set the "broadcast" flag...
   val = 1;
 
-  if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &val, sizeof(val)))
+  if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (const char *)&val, sizeof(val)))
   {
     close(fd);
 
@@ -280,9 +280,14 @@ _papplSNMPWalk(
   _pappl_snmp_t	packet;			// Current response packet
   int		lastoid[_PAPPL_SNMP_MAX_OID];
 					// Last OID we got
+#ifdef DEBUG
+  char		temp[1024];		// OID returned
+#endif // DEBUG
 
 
   // Range check input...
+  _PAPPL_DEBUG("_papplSNMPWalk(fd=%d, address=%p, version=%d, community=\"%s\", prefix=%s, timeout=%g, cb=%p, data=%p)\n", fd, address, version, community ? community : "(null)", _papplSNMPOIDToString(prefix, temp, sizeof(temp)), timeout, cb, data);
+
   if (fd < 0 || !address || version != _PAPPL_SNMP_VERSION_1 || !community || !prefix || !cb)
     return (-1);
 
@@ -295,16 +300,30 @@ _papplSNMPWalk(
     request_id ++;
 
     if (!_papplSNMPWrite(fd, address, version, community, _PAPPL_ASN1_GET_NEXT_REQUEST, request_id, packet.object_name))
+    {
+      _PAPPL_DEBUG("_papplSNMPWalk: Unable to send Get-Next-Request.\n");
       return (-1);
+    }
 
     if (!_papplSNMPRead(fd, &packet, timeout))
+    {
+      _PAPPL_DEBUG("_papplSNMPWalk: Unable to read response.\n");
       return (-1);
+    }
+
+    _PAPPL_DEBUG("_papplSNMPWalk: OID %s.\n", _papplSNMPOIDToString(packet.object_name, temp, sizeof(temp)));
 
     if (!_papplSNMPIsOIDPrefixed(&packet, prefix) || _papplSNMPIsOID(&packet, lastoid))
+    {
+      _PAPPL_DEBUG("_papplSNMPWalk: Different prefix or same OID as last, returning %d.\n", count);
       return (count);
+    }
 
     if (packet.error || packet.error_status)
+    {
+      _PAPPL_DEBUG("_papplSNMPWalk: error=\"%s\", error_status=%d, returning %d.\n", packet.error, packet.error_status, count > 0 ? count : -1);
       return (count > 0 ? count : -1);
+    }
 
     _papplSNMPCopyOID(lastoid, packet.object_name, _PAPPL_SNMP_MAX_OID);
 
@@ -351,9 +370,9 @@ _papplSNMPWrite(
   packet.request_id   = request_id;
   packet.object_type  = _PAPPL_ASN1_NULL_VALUE;
 
-  strlcpy(packet.community, community, sizeof(packet.community));
+  papplCopyString(packet.community, community, sizeof(packet.community));
 
-  for (i = 0; oid[i] >= 0 && i < (_PAPPL_SNMP_MAX_OID - 1); i ++)
+  for (i = 0; i < (_PAPPL_SNMP_MAX_OID - 1) && oid[i] >= 0; i ++)
     packet.object_name[i] = oid[i];
   packet.object_name[i] = -1;
 
@@ -375,7 +394,11 @@ _papplSNMPWrite(
   temp               = *address;
   temp.ipv4.sin_port = htons(_PAPPL_SNMP_PORT);
 
+#if _WIN32
+  return (sendto(fd, buffer, (int)bytes, 0, (void *)&temp, (socklen_t)httpAddrLength(&temp)) == bytes);
+#else
   return (sendto(fd, buffer, (size_t)bytes, 0, (void *)&temp, (socklen_t)httpAddrLength(&temp)) == bytes);
+#endif // _WIN32
 }
 
 
