@@ -1,7 +1,7 @@
 //
 // System accessor functions for the Printer Application Framework
 //
-// Copyright © 2020-2021 by Michael R Sweet.
+// Copyright © 2020-2022 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -22,6 +22,9 @@
 #ifdef HAVE_LIBPNG
 #  include <png.h>
 #endif // HAVE_LIBPNG
+#ifndef _WIN32
+#  include <sys/resource.h>
+#endif // !_WIN32
 
 
 //
@@ -63,12 +66,17 @@ papplSystemAddListeners(
     return (false);
   }
 
+#if !_WIN32
   if (name && *name == '/')
   {
     // Add a domain socket listener...
     ret = add_listeners(system, name, 0, AF_LOCAL);
+    if (ret && !system->domain_path)
+      system->domain_path = strdup(name);
   }
-  else if (name && isdigit(*name & 255))
+  else
+#endif // !_WIN32
+  if (name && isdigit(*name & 255))
   {
     // Add IPv4 listener...
     if (system->port)
@@ -77,8 +85,13 @@ papplSystemAddListeners(
     }
     else
     {
-      int	port = 7999 + (getuid() % 1000);
-					// Current port
+      int	port;			// Current port
+
+#if _WIN32
+      port = 7999;
+#else
+      port = 7999 + (getuid() % 1000);
+#endif // _WIN32
 
       do
       {
@@ -100,8 +113,13 @@ papplSystemAddListeners(
     }
     else
     {
-      int	port = 7999 + (getuid() % 1000);
-					// Current port
+      int	port;			// Current port
+
+#if _WIN32
+      port = 7999;
+#else
+      port = 7999 + (getuid() % 1000);
+#endif // _WIN32
 
       do
       {
@@ -124,8 +142,13 @@ papplSystemAddListeners(
     }
     else
     {
-      int	port = 7999 + (getuid() % 1000);
-					// Current port
+      int	port;		// Current port
+
+#if _WIN32
+      port = 7999;
+#else
+      port = 7999 + (getuid() % 1000);
+#endif // _WIN32
 
       do
       {
@@ -175,7 +198,7 @@ papplSystemAddMIMEFilter(
     return;
 
   if (!system->filters)
-    system->filters = cupsArrayNew3((cups_array_func_t)compare_filters, NULL, NULL, 0, (cups_acopy_func_t)copy_filter, (cups_afree_func_t)free);
+    system->filters = cupsArrayNew((cups_array_cb_t)compare_filters, NULL, NULL, 0, (cups_acopy_cb_t)copy_filter, (cups_afree_cb_t)free);
 
   key.src    = srctype;
   key.dst    = dsttype;
@@ -201,7 +224,7 @@ _papplSystemExportVersions(
     ipp_tag_t      group_tag,		// I - Group (`IPP_TAG_PRINTER` or `IPP_TAG_SYSTEM`)
     cups_array_t   *ra)			// I - Requested attributes or `NULL` for all
 {
-  int		i;			// Looping var
+  cups_len_t	i;			// Looping var
   ipp_attribute_t *attr;		// Attribute
   char		name[128];		// Attribute name
   const char	*name_prefix = (group_tag == IPP_TAG_PRINTER) ? "printer" : "system";
@@ -373,7 +396,7 @@ papplSystemGetAdminGroup(
 
     if (system->admin_group)
     {
-      strlcpy(buffer, system->admin_group, bufsize);
+      papplCopyString(buffer, system->admin_group, bufsize);
       ret = buffer;
     }
     else
@@ -470,7 +493,7 @@ papplSystemGetDefaultPrintGroup(
 
     if (system->default_print_group)
     {
-      strlcpy(buffer, system->default_print_group, bufsize);
+      papplCopyString(buffer, system->default_print_group, bufsize);
       ret = buffer;
     }
     else
@@ -507,7 +530,7 @@ papplSystemGetDNSSDName(
 
     if (system->dns_sd_name)
     {
-      strlcpy(buffer, system->dns_sd_name, bufsize);
+      papplCopyString(buffer, system->dns_sd_name, bufsize);
       ret = buffer;
     }
     else
@@ -559,7 +582,7 @@ papplSystemGetGeoLocation(
 
     if (system->geo_location)
     {
-      strlcpy(buffer, system->geo_location, bufsize);
+      papplCopyString(buffer, system->geo_location, bufsize);
       ret = buffer;
     }
     else
@@ -577,7 +600,10 @@ papplSystemGetGeoLocation(
 //
 // 'papplSystemGetHostname()' - Get the system hostname.
 //
-// This function copies the current system hostname to the specified buffer.
+// This function is deprecated.  Use the @link papplSystemGetHostName@ function
+// instead.
+//
+// @deprecated@ @exclude all@
 //
 
 char *					// O - Hostname
@@ -585,6 +611,22 @@ papplSystemGetHostname(
     pappl_system_t *system,		// I - System
     char           *buffer,		// I - String buffer
     size_t         bufsize)		// I - Size of string buffer
+{
+  return (papplSystemGetHostName(system, buffer, bufsize));
+}
+
+
+//
+// 'papplSystemGetHostName()' - Get the system hostname.
+//
+// This function copies the current system hostname to the specified buffer.
+//
+
+char *					// O - Hostname
+papplSystemGetHostName(
+    pappl_system_t *system,		// I - System
+    char           *buffer,		// I - String buffer
+    size_t         bufsize)		// I - Size of string buffer
 {
   char	*ret = NULL;			// Return value
 
@@ -595,7 +637,7 @@ papplSystemGetHostname(
 
     if (system->hostname)
     {
-      strlcpy(buffer, system->hostname, bufsize);
+      papplCopyString(buffer, system->hostname, bufsize);
       ret = buffer;
     }
     else
@@ -610,6 +652,22 @@ papplSystemGetHostname(
 }
 
 
+//
+// 'papplSystemGetHostPort()' - Get the port number for network connections to
+//                              the system.
+//
+// This function returns the port number that is used for network connections
+// to the system.
+//
+
+int					// O - Port number
+papplSystemGetHostPort(
+    pappl_system_t *system)		// I - System
+{
+  return (system ? system->port : 0);
+}
+
+
 //
 // 'papplSystemGetLocation()' - Get the system location string, if any.
 //
@@ -632,7 +690,7 @@ papplSystemGetLocation(
 
     if (system->location)
     {
-      strlcpy(buffer, system->location, bufsize);
+      papplCopyString(buffer, system->location, bufsize);
       ret = buffer;
     }
     else
@@ -653,13 +711,29 @@ papplSystemGetLocation(
 // This function returns the current system log level as an enumeration.
 //
 
-pappl_loglevel_t
+pappl_loglevel_t			// O - Log level
 papplSystemGetLogLevel(
-    pappl_system_t *system)     // I - System
+    pappl_system_t *system)		// I - System
 {
   return (system ? system->loglevel : PAPPL_LOGLEVEL_UNSPEC);
 }
 
+
+//
+// 'papplSystemGetMaxClients()' - Get the maximum number of clients.
+//
+// This function gets the maximum number of simultaneous clients that are
+// allowed by the system.
+//
+
+int					// O - Maximum number of clients
+papplSystemGetMaxClients(
+    pappl_system_t *system)		// I - System
+{
+  return (system ? system->max_clients : 0);
+}
+
+
 //
 // 'papplSystemGetMaxLogSize()' - Get the maximum log file size.
 //
@@ -679,6 +753,23 @@ papplSystemGetMaxLogSize(
 }
 
 
+//
+// 'papplSystemGetMaxSubscriptions()' - Get the maximum number of event subscriptions.
+//
+// This function gets the maximum number of event subscriptions that are
+// allowed.  A maximum of `0` means there is no limit.
+//
+// The default maximum number of event subscriptions is 100.
+//
+
+size_t					// O - Maximum number of subscriptions or `0`
+papplSystemGetMaxSubscriptions(
+    pappl_system_t *system)		// I - System
+{
+  return (system ? system->max_subscriptions : 0);
+}
+
+
 //
 // 'papplSystemGetName()' - Get the system name.
 //
@@ -700,7 +791,7 @@ papplSystemGetName(
 
     if (system->name)
     {
-      strlcpy(buffer, system->name, bufsize);
+      papplCopyString(buffer, system->name, bufsize);
       ret = buffer;
     }
     else
@@ -766,7 +857,7 @@ papplSystemGetOrganization(
 
     if (system->organization)
     {
-      strlcpy(buffer, system->organization, bufsize);
+      papplCopyString(buffer, system->organization, bufsize);
       ret = buffer;
     }
     else
@@ -803,7 +894,7 @@ papplSystemGetOrganizationalUnit(
 
     if (system->org_unit)
     {
-      strlcpy(buffer, system->org_unit, bufsize);
+      papplCopyString(buffer, system->org_unit, bufsize);
       ret = buffer;
     }
     else
@@ -838,7 +929,7 @@ papplSystemGetPassword(
   {
     pthread_rwlock_rdlock(&system->rwlock);
 
-    strlcpy(buffer, system->password_hash, bufsize);
+    papplCopyString(buffer, system->password_hash, bufsize);
 
     pthread_rwlock_unlock(&system->rwlock);
   }
@@ -853,8 +944,10 @@ papplSystemGetPassword(
 // 'papplSystemGetPort()' - Get the port number for network connections to the
 //                          system.
 //
-// This function returns the port number that is used for network connections
-// to the system.
+// This function is deprecated.  Use the @link papplSystemGetHostName@ function
+// instead.
+//
+// @deprecated@ @exclude all@
 //
 
 int					// O - Port number
@@ -904,7 +997,7 @@ papplSystemGetSessionKey(
       // Lock for updating the session key with random data...
       pthread_rwlock_wrlock(&system->session_rwlock);
 
-      snprintf(system->session_key, sizeof(system->session_key), "%08x%08x%08x%08x%08x%08x%08x%08x", _papplGetRand(), _papplGetRand(), _papplGetRand(), _papplGetRand(), _papplGetRand(), _papplGetRand(), _papplGetRand(), _papplGetRand());
+      snprintf(system->session_key, sizeof(system->session_key), "%08x%08x%08x%08x%08x%08x%08x%08x", papplGetRand(), papplGetRand(), papplGetRand(), papplGetRand(), papplGetRand(), papplGetRand(), papplGetRand(), papplGetRand());
       system->session_time = curtime;
     }
     else
@@ -913,7 +1006,7 @@ papplSystemGetSessionKey(
       pthread_rwlock_rdlock(&system->session_rwlock);
     }
 
-    strlcpy(buffer, system->session_key, bufsize);
+    papplCopyString(buffer, system->session_key, bufsize);
 
     pthread_rwlock_unlock(&system->session_rwlock);
   }
@@ -975,15 +1068,15 @@ papplSystemGetVersions(
   {
     pthread_rwlock_rdlock(&system->rwlock);
 
-    if (max_versions > system->num_versions)
-      memcpy(versions, system->versions, (size_t)system->num_versions * sizeof(pappl_version_t));
+    if (max_versions > (int)system->num_versions)
+      memcpy(versions, system->versions, system->num_versions * sizeof(pappl_version_t));
     else
       memcpy(versions, system->versions, (size_t)max_versions * sizeof(pappl_version_t));
 
     pthread_rwlock_unlock(&system->rwlock);
   }
 
-  return (system ? system->num_versions : 0);
+  return (system ? (int)system->num_versions : 0);
 }
 
 
@@ -1017,14 +1110,14 @@ papplSystemHashPassword(
     if (salt && strchr(salt, '~'))
     {
       // Copy existing nonce from the salt string...
-      strlcpy(nonce, salt, sizeof(nonce));
+      papplCopyString(nonce, salt, sizeof(nonce));
       if ((ptr = strchr(nonce, ':')) != NULL)
         *ptr = '\0';
     }
     else
     {
       // Generate a new random nonce...
-      snprintf(nonce, sizeof(nonce), "%08x%08x", _papplGetRand(), _papplGetRand());
+      snprintf(nonce, sizeof(nonce), "%08x%08x", papplGetRand(), papplGetRand());
     }
 
     snprintf(temp, sizeof(temp), "%s:%s", nonce, password);
@@ -1082,7 +1175,7 @@ papplSystemIteratePrinters(
     pappl_printer_cb_t cb,		// I - Callback function
     void               *data)		// I - Callback data
 {
-  int			i,		// Looping var
+  size_t		i,		// Looping var
 			count;		// Number of printers
 
 
@@ -1091,12 +1184,12 @@ papplSystemIteratePrinters(
 
   // Loop through the printers.
   //
-  // Note: Cannot use cupsArrayFirst/Last since other threads might be
+  // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
   // enumerating the printers array.
 
   pthread_rwlock_rdlock(&system->rwlock);
-  for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
-    (cb)((pappl_printer_t *)cupsArrayIndex(system->printers, i), data);
+  for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
+    (cb)((pappl_printer_t *)cupsArrayGetElement(system->printers, i), data);
   pthread_rwlock_unlock(&system->rwlock);
 }
 
@@ -1110,7 +1203,7 @@ papplSystemMatchDriver(
     pappl_system_t *system,		// I - System
     const char     *device_id)		// I - IEEE-1284 device ID string
 {
-  int		i;			// Looping var
+  size_t	i;			// Looping var
   pappl_pr_driver_t *driver;		// Current driver
   const char	*drvstart,		// Start of key/value pair
 		*drvend,		// End of key/value pair
@@ -1204,6 +1297,7 @@ papplSystemSetAdminGroup(
     free(system->admin_group);
     system->admin_group = value ? strdup(value) : NULL;
 
+#if !_WIN32 // TODO: Implement Windows admin group support
     if (system->admin_group && strcmp(system->admin_group, "none"))
     {
       char		buffer[8192];	// Buffer for strings
@@ -1216,6 +1310,7 @@ papplSystemSetAdminGroup(
 	system->admin_gid = grp->gr_gid;
     }
     else
+#endif // !_WIN32
       system->admin_gid = (gid_t)-1;
 
     _papplSystemConfigChanged(system);
@@ -1225,6 +1320,40 @@ papplSystemSetAdminGroup(
 }
 
 
+//
+// 'papplSystemSetAuthCallback()' - Set an authentication callback for the specified scheme.
+//
+// This function sets the authentication callback that is used for Client
+// requests.  The authentication callback is used for every Client request
+// containing the WWW-Authenticate header (`HTTP_FIELD_WWW_AUTHENTICATE`).
+// The callback returns one of the following status codes:
+//
+// - `HTTP_STATUS_CONTINUE` if the authentication succeeded,
+// - `HTTP_STATUS_UNAUTHORIZED` if the authentication failed, or
+// - `HTTP_STATUS_FORBIDDEN` if the authentication succeeded but the user is
+//   not part of the specified group.
+//
+
+void
+papplSystemSetAuthCallback(
+    pappl_system_t  *system,		// I - System
+    const char      *auth_scheme,	// I - Authentication scheme
+    pappl_auth_cb_t auth_cb,		// I - Callback function
+    void            *auth_cbdata)	// I - Callback data
+{
+  if (system)
+  {
+    pthread_rwlock_wrlock(&system->rwlock);
+
+    free(system->auth_scheme);
+    system->auth_scheme = auth_scheme ? strdup(auth_scheme) : NULL;
+    system->auth_cb     = auth_cb;
+    system->auth_cbdata = auth_cbdata;
+
+    pthread_rwlock_unlock(&system->rwlock);
+  }
+}
+
 //
 // 'papplSystemSetContact()' - Set the "system-contact" value.
 //
@@ -1335,6 +1464,31 @@ papplSystemSetDNSSDName(
 }
 
 
+//
+// 'papplSystemSetEventCallback()' - Set a callback for monitoring system events.
+//
+// This function sets a callback function to receive event notifications from
+// the system.
+//
+
+void
+papplSystemSetEventCallback(
+    pappl_system_t   *system,		// I - System
+    pappl_event_cb_t event_cb,		// I - Event callback function
+    void             *event_data)	// I - Event callback data
+{
+  if (system && event_cb)
+  {
+    pthread_rwlock_wrlock(&system->rwlock);
+
+    system->event_cb   = event_cb;
+    system->event_data = event_data;
+
+    pthread_rwlock_unlock(&system->rwlock);
+  }
+}
+
+
 //
 // 'papplSystemSetFooterHTML()' - Set the footer HTML for the web interface.
 //
@@ -1399,20 +1553,39 @@ papplSystemSetGeoLocation(
 //
 // 'papplSystemSetHostname()' - Set the system hostname.
 //
-// This function sets the system hostname.  If `NULL`, the default hostname
-// is used.
+// This function is deprecated.  Use the @link papplSystemSetHostName@ function
+// instead.
+//
+// @deprecated@ @exclude all@
 //
 
 void
 papplSystemSetHostname(
     pappl_system_t *system,		// I - System
     const char     *value)		// I - Hostname or `NULL` for default
+{
+  papplSystemSetHostName(system, value);
+}
+
+
+//
+// 'papplSystemSetHostName()' - Set the system hostname.
+//
+// This function sets the system hostname.  If `NULL`, the default hostname
+// is used.
+//
+
+void
+papplSystemSetHostName(
+    pappl_system_t *system,		// I - System
+    const char     *value)		// I - Hostname or `NULL` for default
 {
   if (system)
   {
-    pthread_rwlock_wrlock(&system->rwlock);
+    char	temp[1024],		// Temporary hostname string
+		*ptr;			// Pointer in temporary hostname
 
-    free(system->hostname);
+    pthread_rwlock_wrlock(&system->rwlock);
 
     if (value)
     {
@@ -1434,15 +1607,12 @@ papplSystemSetHostname(
         avahi_client_set_host_name(master, value);
 #endif // HAVE_AVAHI
 
+#if !_WIN32
       sethostname(value, (int)strlen(value));
-
-      system->hostname = strdup(value);
+#endif // !_WIN32
     }
     else
     {
-      char	temp[1024],		// Temporary hostname string
-		*ptr;			// Pointer in temporary hostname
-
 #ifdef HAVE_AVAHI
       _pappl_dns_sd_t	master = _papplDNSSDInit(system);
 					  // DNS-SD master reference
@@ -1450,7 +1620,7 @@ papplSystemSetHostname(
 					  // mDNS hostname
 
       if (avahi_name)
-	strlcpy(temp, avahi_name, sizeof(temp));
+	papplCopyString(temp, avahi_name, sizeof(temp));
       else
 #endif /* HAVE_AVAHI */
       httpGetHostname(NULL, temp, sizeof(temp));
@@ -1458,20 +1628,30 @@ papplSystemSetHostname(
       if ((ptr = strstr(temp, ".lan")) != NULL && !ptr[4])
       {
         // Replace hostname.lan with hostname.local
-        strlcpy(ptr, ".local", sizeof(temp) - (size_t)(ptr - temp));
+        papplCopyString(ptr, ".local", sizeof(temp) - (size_t)(ptr - temp));
       }
       else if (!strrchr(temp, '.'))
       {
         // No domain information, so append .local to hostname...
         ptr = temp + strlen(temp);
-        strlcpy(ptr, ".local", sizeof(temp) - (size_t)(ptr - temp));
+        papplCopyString(ptr, ".local", sizeof(temp) - (size_t)(ptr - temp));
       }
 
-      system->hostname = strdup(temp);
+      value = temp;
     }
 
-    // Force an update of all DNS-SD registrations...
-    system->dns_sd_host_changes = -1;
+    if (system->hostname && strcasecmp(system->hostname, value) && system->is_running)
+    {
+      // Force an update of all DNS-SD registrations...
+      system->dns_sd_host_changes = -1;
+    }
+
+    // Save the new hostname value
+    free(system->hostname);
+    system->hostname = strdup(value);
+
+    // Set the system TLS credentials...
+    cupsSetServerCredentials(NULL, system->hostname, 1);
 
     pthread_rwlock_unlock(&system->rwlock);
   }
@@ -1528,6 +1708,90 @@ papplSystemSetLogLevel(
   }
 }
 
+
+//
+// 'papplSystemSetMaxClients()' - Set the maximum number of clients.
+//
+// This function sets the maximum number of simultaneous clients that are
+// allowed by the system from 0 (auto) to 32768 (half of the available TCP
+// port numbers).
+//
+// The default maximum number of clients is based on available system resources.
+//
+
+void
+papplSystemSetMaxClients(
+    pappl_system_t *system,		// I - System
+    int            max_clients)		// I - Maximum number of clients or `0` for auto
+{
+  if (!system)
+    return;
+
+  if (max_clients <= 0)
+  {
+    // Determine the maximum number of clients to support...
+#ifdef _WIN32
+    max_clients = 100;			// Use a default of 100...
+
+#else
+    struct rlimit	file_limits,	// Current file descriptor limits
+			mem_limits;	// Current memory limits
+
+    max_clients = 100;			// Use a default of 100...
+
+    if (!getrlimit(RLIMIT_NOFILE, &file_limits) && !getrlimit(RLIMIT_DATA, &mem_limits))
+    {
+      // Calculate a maximum number of clients...
+      int max_files, max_mem;		// Maximum files and memory
+
+      if (file_limits.rlim_cur != file_limits.rlim_max && file_limits.rlim_cur < 65536)
+      {
+        // Try increasing the limit to the maximum allowed...
+        if (file_limits.rlim_max > 65536)
+	  file_limits.rlim_cur = 65536;
+        else
+	  file_limits.rlim_cur = file_limits.rlim_max;
+
+        if (setrlimit(RLIMIT_NOFILE, &file_limits))
+          getrlimit(RLIMIT_NOFILE, &file_limits);
+      }
+
+      // Max clients based on file descriptors is 1/2 the limit...
+      if (file_limits.rlim_cur == RLIM_INFINITY)
+        max_files = 32768;
+      else
+        max_files = (int)(file_limits.rlim_cur / 2);
+
+      // Max clients based on memory is 1/64k the limit...
+      if (mem_limits.rlim_cur == RLIM_INFINITY)
+        max_mem = 32768;
+      else
+        max_mem = (int)(mem_limits.rlim_cur / 65536);
+
+      // Use min(max_files,max_mem)...
+      if (max_files > max_mem)
+        max_clients = max_mem;
+      else
+        max_clients = max_files;
+    }
+#endif // _WIN32
+  }
+
+  // Restrict max_clients to <= 32768...
+  if (max_clients > 32768)
+    max_clients = 32768;
+
+  // Set the new value...
+  pthread_rwlock_wrlock(&system->rwlock);
+
+  system->max_clients = max_clients;
+
+  _papplSystemConfigChanged(system);
+
+  pthread_rwlock_unlock(&system->rwlock);
+}
+
+
 //
 // 'papplSystemSetMaxLogSize()' - Set the maximum log file size in bytes.
 //
@@ -1557,6 +1821,33 @@ papplSystemSetMaxLogSize(
 }
 
 
+//
+// 'papplSystemSetMaxSubscriptions()' - Set the maximum number of event subscriptions.
+//
+// This function Sets the maximum number of event subscriptions that are
+// allowed.  A maximum of `0` means there is no limit.
+//
+// The default maximum number of event subscriptions is `100`.
+//
+
+void
+papplSystemSetMaxSubscriptions(
+    pappl_system_t *system,		// I - System
+    size_t         max_subscriptions)	// I - Maximum number of subscriptions or `0` for no limit
+{
+  if (system)
+  {
+    pthread_rwlock_wrlock(&system->rwlock);
+
+    system->max_subscriptions = max_subscriptions;
+
+    _papplSystemConfigChanged(system);
+
+    pthread_rwlock_unlock(&system->rwlock);
+  }
+}
+
+
 //
 // 'papplSystemSetMIMECallback()' - Set the MIME typing callback for the system.
 //
@@ -1581,7 +1872,6 @@ papplSystemSetMIMECallback(
   {
     pthread_rwlock_wrlock(&system->rwlock);
 
-    system->config_time = time(NULL);
     system->mime_cb     = cb;
     system->mime_cbdata = data;
 
@@ -1718,7 +2008,7 @@ papplSystemSetPassword(
   {
     pthread_rwlock_wrlock(&system->rwlock);
 
-    strlcpy(system->password_hash, hash, sizeof(system->password_hash));
+    papplCopyString(system->password_hash, hash, sizeof(system->password_hash));
 
     _papplSystemConfigChanged(system);
 
@@ -1763,8 +2053,7 @@ papplSystemSetPrinterDrivers(
   {
     pthread_rwlock_wrlock(&system->rwlock);
 
-    system->config_time   = time(NULL);
-    system->num_drivers   = num_drivers;
+    system->num_drivers   = (size_t)num_drivers;
     system->drivers       = drivers;
     system->autoadd_cb    = autoadd_cb;
     system->create_cb     = create_cb;
@@ -1867,17 +2156,53 @@ papplSystemSetVersions(
     pthread_rwlock_wrlock(&system->rwlock);
 
     if (num_versions > (int)(sizeof(system->versions) / sizeof(system->versions[0])))
-      system->num_versions = (int)(sizeof(system->versions) / sizeof(system->versions[0]));
+      system->num_versions = sizeof(system->versions) / sizeof(system->versions[0]);
     else
-      system->num_versions = num_versions;
+      system->num_versions = (size_t)num_versions;
 
-    memcpy(system->versions, versions, (size_t)system->num_versions * sizeof(pappl_version_t));
+    memcpy(system->versions, versions, system->num_versions * sizeof(pappl_version_t));
 
     pthread_rwlock_unlock(&system->rwlock);
   }
 }
 
 
+//
+// 'papplSystemSetWiFiCallbacks()' - Set Wi-Fi callbacks.
+//
+// This function sets the 802.11 Wi-Fi interface callbacks for the system.  The
+// "join_cb" is used to join a Wi-Fi network, the "list_cb" is used to list
+// available networks, and the "status_cb" is used to query the current Wi-Fi
+// network connection status and Secure Set Identifier (SSID).  The "join_cb"
+// and "status_cb" functions are used to support getting and setting the IPP
+// "printer-wifi-state", "printer-wifi-ssid", and "printer-wifi-password"
+// attributes, while the "list_cb" function enables changing the Wi-Fi network
+// from the network web interface, if enabled.
+//
+// Note: The Wi-Fi callbacks can only be set prior to calling
+// @link papplSystemRun@.
+//
+
+void
+papplSystemSetWiFiCallbacks(
+    pappl_system_t         *system,	// I - System
+    pappl_wifi_join_cb_t   join_cb,	// I - Join callback
+    pappl_wifi_list_cb_t   list_cb,	// I - List callback
+    pappl_wifi_status_cb_t status_cb,	// I - Status callback
+    void                   *data)	// I - Callback data pointer
+{
+  if (system && !system->is_running && join_cb && status_cb)
+  {
+    pthread_rwlock_wrlock(&system->rwlock);
+    system->wifi_join_cb   = join_cb;
+    system->wifi_list_cb   = list_cb;
+    system->wifi_status_cb = status_cb;
+    system->wifi_cbdata    = data;
+    pthread_rwlock_unlock(&system->rwlock);
+  }
+}
+
+
 //
 // 'add_listeners()' - Create and add listener sockets to a system.
 //
