@@ -1,7 +1,7 @@
 //
 // IPP processing for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -23,7 +23,7 @@ typedef struct _pappl_attr_s		// Input attribute structure
 {
   const char	*name;			// Attribute name
   ipp_tag_t	value_tag;		// Value tag
-  int		max_count;		// Max number of values
+  cups_len_t	max_count;		// Max number of values
 } _pappl_attr_t;
 
 
@@ -31,13 +31,16 @@ typedef struct _pappl_attr_s		// Input attribute structure
 // Local functions...
 //
 
-static void		ipp_create_printer(pappl_client_t *client);
-static void		ipp_delete_printer(pappl_client_t *client);
-static void		ipp_get_printers(pappl_client_t *client);
-static void		ipp_get_system_attributes(pappl_client_t *client);
-static void		ipp_set_system_attributes(pappl_client_t *client);
-static void		ipp_shutdown_all_printers(pappl_client_t *client);
-
+static void	ipp_create_printer(pappl_client_t *client);
+static void	ipp_delete_printer(pappl_client_t *client);
+static void	ipp_disable_all_printers(pappl_client_t *client);
+static void	ipp_enable_all_printers(pappl_client_t *client);
+static void	ipp_get_printers(pappl_client_t *client);
+static void	ipp_get_system_attributes(pappl_client_t *client);
+static void	ipp_pause_all_printers(pappl_client_t *client);
+static void	ipp_resume_all_printers(pappl_client_t *client);
+static void	ipp_set_system_attributes(pappl_client_t *client);
+static void	ipp_shutdown_all_printers(pappl_client_t *client);
 
 //
 // '_papplSystemProcessIPP()' - Process an IPP System request.
@@ -76,10 +79,51 @@ _papplSystemProcessIPP(
 	ipp_set_system_attributes(client);
 	break;
 
+    case IPP_OP_DISABLE_ALL_PRINTERS :
+        ipp_disable_all_printers(client);
+        break;
+
+    case IPP_OP_ENABLE_ALL_PRINTERS :
+        ipp_enable_all_printers(client);
+        break;
+
+    case IPP_OP_PAUSE_ALL_PRINTERS :
+    case IPP_OP_PAUSE_ALL_PRINTERS_AFTER_CURRENT_JOB :
+        ipp_pause_all_printers(client);
+        break;
+
+    case IPP_OP_RESUME_ALL_PRINTERS :
+        ipp_resume_all_printers(client);
+        break;
+
     case IPP_OP_SHUTDOWN_ALL_PRINTERS :
 	ipp_shutdown_all_printers(client);
 	break;
 
+    case IPP_OP_CREATE_SYSTEM_SUBSCRIPTIONS :
+        _papplSubscriptionIPPCreate(client);
+        break;
+
+    case IPP_OP_GET_SUBSCRIPTION_ATTRIBUTES :
+        _papplSubscriptionIPPGetAttributes(client);
+        break;
+
+    case IPP_OP_GET_SUBSCRIPTIONS :
+        _papplSubscriptionIPPList(client);
+        break;
+
+    case IPP_OP_RENEW_SUBSCRIPTION :
+        _papplSubscriptionIPPRenew(client);
+        break;
+
+    case IPP_OP_CANCEL_SUBSCRIPTION :
+        _papplSubscriptionIPPCancel(client);
+        break;
+
+    case IPP_OP_GET_NOTIFICATIONS :
+        _papplSubscriptionIPPGetNotifications(client);
+        break;
+
     default :
         if (client->system->op_cb && (client->system->op_cb)(client, client->system->op_cbdata))
           break;
@@ -230,7 +274,7 @@ ipp_create_printer(
   // Return the printer
   papplClientRespondIPP(client, IPP_STATUS_OK, NULL);
 
-  ra = cupsArrayNew((cups_array_func_t)strcmp, NULL);
+  ra = cupsArrayNew((cups_array_cb_t)strcmp, NULL, NULL, 0, NULL, NULL);
   cupsArrayAdd(ra, "printer-id");
   cupsArrayAdd(ra, "printer-is-accepting-jobs");
   cupsArrayAdd(ra, "printer-state");
@@ -238,7 +282,7 @@ ipp_create_printer(
   cupsArrayAdd(ra, "printer-uuid");
   cupsArrayAdd(ra, "printer-xri-supported");
 
-  _papplPrinterCopyAttributes(client, printer, ra, NULL);
+  _papplPrinterCopyAttributes(printer, client, ra, NULL);
   cupsArrayDelete(ra);
 }
 
@@ -276,6 +320,68 @@ ipp_delete_printer(
 }
 
 
+//
+// 'ipp_disable_all_printers()' - Disable all printers.
+//
+
+static void
+ipp_disable_all_printers(
+    pappl_client_t *client)		// I - Client
+{
+  pappl_system_t	*system = client->system;
+					// System
+  pappl_printer_t	*printer;	// Current printer
+  http_status_t		auth_status;	// Authorization status
+
+
+  // Verify the connection is authorized...
+  if ((auth_status = papplClientIsAuthorized(client)) != HTTP_STATUS_CONTINUE)
+  {
+    papplClientRespond(client, auth_status, NULL, NULL, 0, 0);
+    return;
+  }
+
+  // Loop through the printers...
+  pthread_rwlock_rdlock(&system->rwlock);
+  for (printer = (pappl_printer_t *)cupsArrayGetFirst(system->printers); printer; printer = (pappl_printer_t *)cupsArrayGetNext(system->printers))
+  {
+    papplPrinterDisable(printer);
+  }
+  pthread_rwlock_unlock(&system->rwlock);
+}
+
+
+//
+// 'ipp_enable_all_printers()' - Enable all printers.
+//
+
+static void
+ipp_enable_all_printers(
+    pappl_client_t *client)		// I - Client
+{
+  pappl_system_t	*system = client->system;
+					// System
+  pappl_printer_t	*printer;	// Current printer
+  http_status_t		auth_status;	// Authorization status
+
+
+  // Verify the connection is authorized...
+  if ((auth_status = papplClientIsAuthorized(client)) != HTTP_STATUS_CONTINUE)
+  {
+    papplClientRespond(client, auth_status, NULL, NULL, 0, 0);
+    return;
+  }
+
+  // Loop through the printers...
+  pthread_rwlock_rdlock(&system->rwlock);
+  for (printer = (pappl_printer_t *)cupsArrayGetFirst(system->printers); printer; printer = (pappl_printer_t *)cupsArrayGetNext(system->printers))
+  {
+    papplPrinterEnable(printer);
+  }
+  pthread_rwlock_unlock(&system->rwlock);
+}
+
+
 //
 // 'ipp_get_printers()' - Get printers.
 //
@@ -287,7 +393,7 @@ ipp_get_printers(
   pappl_system_t	*system = client->system;
 					// System
   cups_array_t		*ra;		// Requested attributes array
-  int			i,		// Looping var
+  size_t		i,		// Looping var
 			count,		// Number of printers
 			limit;		// Maximum number to return
   pappl_printer_t	*printer;	// Current printer
@@ -295,7 +401,7 @@ ipp_get_printers(
 
 
   // Get request attributes...
-  limit  = ippGetInteger(ippFindAttribute(client->request, "limit", IPP_TAG_INTEGER), 0);
+  limit  = (size_t)ippGetInteger(ippFindAttribute(client->request, "limit", IPP_TAG_INTEGER), 0);
   ra     = ippCreateRequestedArray(client->request);
   format = ippGetString(ippFindAttribute(client->request, "document-format", IPP_TAG_MIMETYPE), 0, NULL);
 
@@ -304,14 +410,14 @@ ipp_get_printers(
   pthread_rwlock_rdlock(&system->rwlock);
 
   // Enumerate the printers for the client...
-  count = cupsArrayCount(system->printers);
+  count = cupsArrayGetCount(system->printers);
 
   if (limit > 0 && limit < count)
     count = limit;
 
   for (i = 0; i < count; i ++)
   {
-    printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+    printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
     if (limit && i >= limit)
       break;
@@ -320,7 +426,7 @@ ipp_get_printers(
       ippAddSeparator(client->response);
 
     pthread_rwlock_rdlock(&printer->rwlock);
-    _papplPrinterCopyAttributes(client, printer, ra, format);
+    _papplPrinterCopyAttributes(printer, client, ra, format);
     pthread_rwlock_unlock(&printer->rwlock);
   }
 
@@ -341,7 +447,7 @@ ipp_get_system_attributes(
   pappl_system_t	*system = client->system;
 					// System
   cups_array_t		*ra;		// Requested attributes array
-  int			i,		// Looping var
+  size_t		i,		// Looping var
 			count;		// Count of values
   pappl_printer_t	*printer;	// Current printer
   ipp_attribute_t	*attr;		// Current attribute
@@ -361,9 +467,9 @@ ipp_get_system_attributes(
 
   if (!ra || cupsArrayFind(ra, "system-config-change-date-time") || cupsArrayFind(ra, "system-config-change-time"))
   {
-    for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+    for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
     {
-      printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+      printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
       if (config_time < printer->config_time)
         config_time = printer->config_time;
@@ -378,11 +484,11 @@ ipp_get_system_attributes(
 
   if (!ra || cupsArrayFind(ra, "system-configured-printers"))
   {
-    attr = ippAddCollections(client->response, IPP_TAG_SYSTEM, "system-configured-printers", cupsArrayCount(system->printers), NULL);
+    attr = ippAddCollections(client->response, IPP_TAG_SYSTEM, "system-configured-printers", IPP_NUM_CAST cupsArrayGetCount(system->printers), NULL);
 
-    for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+    for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
     {
-      printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+      printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
       col = ippNew();
 
@@ -390,15 +496,14 @@ ipp_get_system_attributes(
 
       ippAddInteger(col, IPP_TAG_SYSTEM, IPP_TAG_INTEGER, "printer-id", printer->printer_id);
       ippAddString(col, IPP_TAG_SYSTEM, IPP_TAG_TEXT, "printer-info", NULL, printer->name);
-      ippAddBoolean(col, IPP_TAG_SYSTEM, "printer-is-accepting-jobs", 1);
       ippAddString(col, IPP_TAG_SYSTEM, IPP_TAG_TEXT, "printer-name", NULL, printer->name);
       ippAddString(col, IPP_TAG_SYSTEM, IPP_TAG_KEYWORD, "printer-service-type", NULL, "print");
-      _papplPrinterCopyState(col, printer, NULL);
-      _papplPrinterCopyXRI(client, col, printer);
+      _papplPrinterCopyState(printer, IPP_TAG_PRINTER, col, client, NULL);
+      _papplPrinterCopyXRI(printer, col, client);
 
       pthread_rwlock_unlock(&printer->rwlock);
 
-      ippSetCollection(client->response, &attr, i, col);
+      ippSetCollection(client->response, &attr, IPP_NUM_CAST i, col);
       ippDelete(col);
     }
   }
@@ -442,9 +547,9 @@ ipp_get_system_attributes(
   {
     int	state = IPP_PSTATE_IDLE;	// System state
 
-    for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+    for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
     {
-      printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+      printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
       if (printer->state == IPP_PSTATE_PROCESSING)
       {
@@ -458,9 +563,9 @@ ipp_get_system_attributes(
 
   if (!ra || cupsArrayFind(ra, "system-state-change-date-time") || cupsArrayFind(ra, "system-state-change-time"))
   {
-    for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+    for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
     {
-      printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+      printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
       if (state_time < printer->state_time)
         state_time = printer->state_time;
@@ -477,9 +582,9 @@ ipp_get_system_attributes(
   {
     pappl_preason_t	state_reasons = PAPPL_PREASON_NONE;
 
-    for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+    for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
     {
-      printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+      printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
       state_reasons |= printer->state_reasons;
     }
@@ -515,10 +620,10 @@ ipp_get_system_attributes(
   {
     char	uri[1024];		// URI value
 
-    httpAssembleURI(HTTP_URI_CODING_ALL, uri, sizeof(uri), "ipps", NULL, client->host_field, client->host_port, "/ipp/system");
+    httpAssembleURI(HTTP_URI_CODING_ALL, uri, sizeof(uri), (system->options & PAPPL_SOPTIONS_NO_TLS) ? "ipp" : "ipps", NULL, client->host_field, client->host_port, "/ipp/system");
     col = ippNew();
     ippAddString(col, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "xri-authentication", NULL, client->system->auth_service ? "basic" : "none");
-    ippAddString(col, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "xri-security", NULL, "tls");
+    ippAddString(col, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "xri-security", NULL, (system->options & PAPPL_SOPTIONS_NO_TLS) ? "none" : "tls");
     ippAddString(col, IPP_TAG_SYSTEM, IPP_TAG_URI, "xri-uri", NULL, uri);
 
     ippAddCollection(client->response, IPP_TAG_SYSTEM, "system-xri-supported", col);
@@ -531,6 +636,68 @@ ipp_get_system_attributes(
 }
 
 
+//
+// 'ipp_pause_all_printers()' -  all printers.
+//
+
+static void
+ipp_pause_all_printers(
+    pappl_client_t *client)		// I - Client
+{
+  pappl_system_t	*system = client->system;
+					// System
+  pappl_printer_t	*printer;	// Current printer
+  http_status_t		auth_status;	// Authorization status
+
+
+  // Verify the connection is authorized...
+  if ((auth_status = papplClientIsAuthorized(client)) != HTTP_STATUS_CONTINUE)
+  {
+    papplClientRespond(client, auth_status, NULL, NULL, 0, 0);
+    return;
+  }
+
+  // Loop through the printers...
+  pthread_rwlock_rdlock(&system->rwlock);
+  for (printer = (pappl_printer_t *)cupsArrayGetFirst(system->printers); printer; printer = (pappl_printer_t *)cupsArrayGetNext(system->printers))
+  {
+    papplPrinterPause(client->printer);
+  }
+  pthread_rwlock_unlock(&system->rwlock);
+}
+
+
+//
+// 'ipp_resume_all_printers()' -  all printers.
+//
+
+static void
+ipp_resume_all_printers(
+    pappl_client_t *client)		// I - Client
+{
+  pappl_system_t	*system = client->system;
+					// System
+  pappl_printer_t	*printer;	// Current printer
+  http_status_t		auth_status;	// Authorization status
+
+
+  // Verify the connection is authorized...
+  if ((auth_status = papplClientIsAuthorized(client)) != HTTP_STATUS_CONTINUE)
+  {
+    papplClientRespond(client, auth_status, NULL, NULL, 0, 0);
+    return;
+  }
+
+  // Loop through the printers...
+  pthread_rwlock_rdlock(&system->rwlock);
+  for (printer = (pappl_printer_t *)cupsArrayGetFirst(system->printers); printer; printer = (pappl_printer_t *)cupsArrayGetNext(system->printers))
+  {
+    papplPrinterResume(client->printer);
+  }
+  pthread_rwlock_unlock(&system->rwlock);
+}
+
+
 //
 // 'ipp_set_system_attributes()' - Set system attributes.
 //
@@ -543,9 +710,9 @@ ipp_set_system_attributes(
 					// System
   ipp_attribute_t	*rattr;		// Current request attribute
   ipp_tag_t		value_tag;	// Value tag
-  int			count;		// Number of values
+  cups_len_t		count;		// Number of values
   const char		*name;		// Attribute name
-  int			i;		// Looping var
+  size_t		i;		// Looping var
   http_status_t		auth_status;	// Authorization status
   static _pappl_attr_t	sattrs[] =	// Settable system attributes
   {
@@ -584,13 +751,13 @@ ipp_set_system_attributes(
     value_tag = ippGetValueTag(rattr);
     count     = ippGetCount(rattr);
 
-    for (i = 0; i < (int)(sizeof(sattrs) / sizeof(sattrs[0])); i ++)
+    for (i = 0; i < (sizeof(sattrs) / sizeof(sattrs[0])); i ++)
     {
       if (!strcmp(name, sattrs[i].name) && value_tag == sattrs[i].value_tag && count <= sattrs[i].max_count)
         break;
     }
 
-    if (i >= (int)(sizeof(sattrs) / sizeof(sattrs[0])))
+    if (i >= (sizeof(sattrs) / sizeof(sattrs[0])))
       papplClientRespondIPPUnsupported(client, rattr);
 
     if (!strcmp(name, "system-default-printer-id"))
@@ -651,6 +818,8 @@ ipp_set_system_attributes(
 
   pthread_rwlock_unlock(&system->rwlock);
 
+  papplSystemAddEvent(system, NULL, NULL, PAPPL_EVENT_SYSTEM_CONFIG_CHANGED, NULL);
+
   papplClientRespondIPP(client, IPP_STATUS_OK, NULL);
 }
 
