@@ -1,7 +1,7 @@
 //
 // System load/save functions for the Printer Application Framework
 //
-// Copyright © 2020-2021 by Michael R Sweet.
+// Copyright © 2020-2022 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
@@ -23,7 +23,7 @@ static void	parse_media_col(char *value, pappl_media_col_t *media);
 static char	*read_line(cups_file_t *fp, char *line, size_t linesize, char **value, int *linenum);
 static void	write_contact(cups_file_t *fp, pappl_contact_t *contact);
 static void	write_media_col(cups_file_t *fp, const char *name, pappl_media_col_t *media);
-static void	write_options(cups_file_t *fp, const char *name, int num_options, cups_option_t *options);
+static void	write_options(cups_file_t *fp, const char *name, cups_len_t num_options, cups_option_t *options);
 
 
 //
@@ -120,7 +120,7 @@ papplSystemLoadState(
     else if (!strcasecmp(line, "<Printer") && value)
     {
       // Read a printer...
-      int		num_options;	// Number of options
+      cups_len_t	num_options;	// Number of options
       cups_option_t	*options = NULL;// Options
       const char	*printer_id,	// Printer ID
 			*printer_name,	// Printer name
@@ -324,7 +324,8 @@ papplSystemLoadState(
 
 	    if ((attr_fd = papplJobOpenFile(job, job_attr_filename, sizeof(job_attr_filename), system->directory, "ipp", "r")) < 0)
 	    {
-	      papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to open file for job attributes: '%s'.", job_attr_filename);
+	      if (errno != ENOENT)
+		papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to open file for job attributes: '%s'.", job_attr_filename);
 	      continue;
 	    }
 
@@ -387,7 +388,7 @@ papplSystemSaveState(
     pappl_system_t *system,		// I - System
     const char     *filename)		// I - File to save
 {
-  int			i, j,		// Looping vars
+  size_t		i, j,		// Looping vars
 			count;		// Number of printers
   cups_file_t		*fp;		// Output file
   pappl_printer_t	*printer;	// Current printer
@@ -427,16 +428,16 @@ papplSystemSaveState(
 
   // Loop through the printers.
   //
-  // Note: Cannot use cupsArrayFirst/Last since other threads might be
+  // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
   // enumerating the printers array.
 
-  for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+  for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
   {
-    int			jcount;		// Number of jobs
-    int			num_options = 0;// Number of options
+    size_t		jcount;		// Number of jobs
+    cups_len_t		num_options = 0;// Number of options
     cups_option_t	*options = NULL;// Options
 
-    printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+    printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
     if (printer->is_deleted)
       continue;
@@ -480,13 +481,13 @@ papplSystemSaveState(
 
     write_media_col(fp, "media-col-default", &printer->driver_data.media_default);
 
-    for (j = 0; j < printer->driver_data.num_source; j ++)
+    for (j = 0; j < (size_t)printer->driver_data.num_source; j ++)
     {
       if (printer->driver_data.media_ready[j].size_name[0])
       {
         char	name[128];		// Attribute name
 
-        snprintf(name, sizeof(name), "media-col-ready%d", j);
+        snprintf(name, sizeof(name), "media-col-ready%u", (unsigned)j);
         write_media_col(fp, name, printer->driver_data.media_ready + j);
       }
     }
@@ -510,7 +511,7 @@ papplSystemSaveState(
       cupsFilePutConf(fp, "sides-default", _papplSidesString(printer->driver_data.sides_default));
     if (printer->driver_data.x_default)
       cupsFilePrintf(fp, "printer-resolution-default %dx%ddpi\n", printer->driver_data.x_default, printer->driver_data.y_default);
-    for (j = 0; j < printer->driver_data.num_vendor; j ++)
+    for (j = 0; j < (size_t)printer->driver_data.num_vendor; j ++)
     {
       char	defname[128],		// xxx-default name
 	      	defvalue[1024];		// xxx-default value
@@ -521,13 +522,15 @@ papplSystemSaveState(
       cupsFilePutConf(fp, defname, defvalue);
     }
 
-    // Note: Cannot use cupsArrayFirst/Last since other threads might be
+    // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
     // enumerating the all_jobs array.
 
-    for (j = 0, jcount = cupsArrayCount(printer->all_jobs); j < jcount; j ++)
+    for (j = 0, jcount = cupsArrayGetCount(printer->all_jobs); j < jcount; j ++)
     {
-      job = (pappl_job_t *)cupsArrayIndex(printer->all_jobs, j);
-
+      job = (pappl_job_t *)cupsArrayGetElement(printer->all_jobs, j);
+ 
+      pthread_rwlock_rdlock(&job->rwlock);
+ 
       // Add basic job attributes...
       num_options = 0;
       num_options = cupsAddIntegerOption("id", job->job_id, num_options, &options);
@@ -537,7 +540,9 @@ papplSystemSaveState(
 
       if (job->filename)
         num_options = cupsAddOption("filename", job->filename, num_options, &options);
-      if (job->state)
+      if (job->is_canceled)
+        num_options = cupsAddIntegerOption("state", (int)IPP_JSTATE_CANCELED, num_options, &options);
+      else if (job->state)
         num_options = cupsAddIntegerOption("state", (int)job->state, num_options, &options);
       if (job->state_reasons)
         num_options = cupsAddIntegerOption("state_reasons", (int)job->state_reasons, num_options, &options);
@@ -547,6 +552,8 @@ papplSystemSaveState(
         num_options = cupsAddIntegerOption("processing", (int)job->processing, num_options, &options);
       if (job->completed)
         num_options = cupsAddIntegerOption("completed", (int)job->completed, num_options, &options);
+      else if (job->is_canceled)
+        num_options = cupsAddIntegerOption("completed", (int)time(NULL), num_options, &options);
       if (job->impressions)
         num_options = cupsAddIntegerOption("impressions", job->impressions, num_options, &options);
       if (job->impcompleted)
@@ -563,6 +570,7 @@ papplSystemSaveState(
           if ((attr_fd = papplJobOpenFile(job, job_attr_filename, sizeof(job_attr_filename), system->directory, "ipp", "w")) < 0)
           {
             papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to create file for job attributes: '%s'.", job_attr_filename);
+            pthread_rwlock_unlock(&job->rwlock);
             continue;
           }
 
@@ -578,6 +586,8 @@ papplSystemSaveState(
 
       write_options(fp, "Job", num_options, options);
       cupsFreeOptions(num_options, options);
+
+      pthread_rwlock_unlock(&job->rwlock);
     }
 
     cupsFilePuts(fp, "</Printer>\n");
@@ -601,7 +611,7 @@ static void
 parse_contact(char            *value,	// I - Value
               pappl_contact_t *contact)	// O - Contact
 {
-  int		i,			// Looping var
+  cups_len_t	i,			// Looping var
 		num_options;		// Number of options
   cups_option_t	*options = NULL,	// Options
 		*option;		// Current option
@@ -613,11 +623,11 @@ parse_contact(char            *value,	// I - Value
   for (i = num_options, option = options; i > 0; i --, option ++)
   {
     if (!strcasecmp(option->name, "name"))
-      strlcpy(contact->name, option->value, sizeof(contact->name));
+      papplCopyString(contact->name, option->value, sizeof(contact->name));
     else if (!strcasecmp(option->name, "email"))
-      strlcpy(contact->email, option->value, sizeof(contact->email));
+      papplCopyString(contact->email, option->value, sizeof(contact->email));
     else if (!strcasecmp(option->name, "telephone"))
-      strlcpy(contact->telephone, option->value, sizeof(contact->telephone));
+      papplCopyString(contact->telephone, option->value, sizeof(contact->telephone));
   }
 
   cupsFreeOptions(num_options, options);
@@ -633,7 +643,7 @@ parse_media_col(
     char              *value,		// I - Value
     pappl_media_col_t *media)		// O - Media collection
 {
-  int		i,			// Looping var
+  cups_len_t	i,			// Looping var
 		num_options;		// Number of options
   cups_option_t	*options = NULL,	// Options
 		*option;		// Current option
@@ -651,13 +661,13 @@ parse_media_col(
     else if (!strcasecmp(option->name, "right"))
       media->right_margin = (int)strtol(option->value, NULL, 10);
     else if (!strcasecmp(option->name, "name"))
-      strlcpy(media->size_name, option->value, sizeof(media->size_name));
+      papplCopyString(media->size_name, option->value, sizeof(media->size_name));
     else if (!strcasecmp(option->name, "width"))
       media->size_width = (int)strtol(option->value, NULL, 10);
     else if (!strcasecmp(option->name, "length"))
       media->size_length = (int)strtol(option->value, NULL, 10);
     else if (!strcasecmp(option->name, "source"))
-      strlcpy(media->source, option->value, sizeof(media->source));
+      papplCopyString(media->source, option->value, sizeof(media->source));
     else if (!strcasecmp(option->name, "top"))
       media->top_margin = (int)strtol(option->value, NULL, 10);
     else if (!strcasecmp(option->name, "offset"))
@@ -665,7 +675,7 @@ parse_media_col(
     else if (!strcasecmp(option->name, "tracking"))
       media->tracking = _papplMediaTrackingValue(option->value);
     else if (!strcasecmp(option->name, "type"))
-      strlcpy(media->type, option->value, sizeof(media->type));
+      papplCopyString(media->type, option->value, sizeof(media->type));
   }
 
   cupsFreeOptions(num_options, options);
@@ -722,7 +732,7 @@ static void
 write_contact(cups_file_t     *fp,	// I - File
               pappl_contact_t *contact)	// I - Contact
 {
-  int		num_options = 0;	// Number of options
+  cups_len_t	num_options = 0;	// Number of options
   cups_option_t	*options = NULL;	// Options
 
 
@@ -748,7 +758,7 @@ write_media_col(
     const char        *name,		// I - Attribute name
     pappl_media_col_t *media)		// I - Media value
 {
-  int		num_options = 0;	// Number of options
+  cups_len_t	num_options = 0;	// Number of options
   cups_option_t	*options = NULL;	// Options
 
 
@@ -787,7 +797,7 @@ write_media_col(
 static void
 write_options(cups_file_t   *fp,	// I - File
               const char    *name,	// I - Attribute name
-              int           num_options,// I - Number of options
+              cups_len_t    num_options,// I - Number of options
               cups_option_t *options)	// I - Options
 {
   const char	*start,			// Start of current subset
