@@ -1,7 +1,7 @@
 //
 // Printer object for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -41,7 +41,7 @@ _papplSystemAddPrinter(
     printer->printer_id = system->next_printer_id ++;
 
   if (!system->printers)
-    system->printers = cupsArrayNew3((cups_array_func_t)compare_printers, NULL, NULL, 0, NULL, (cups_afree_func_t)_papplPrinterDelete);
+    system->printers = cupsArrayNew((cups_array_cb_t)compare_printers, NULL, NULL, 0, NULL, (cups_afree_cb_t)_papplPrinterDelete);
 
   cupsArrayAdd(system->printers, printer);
 
@@ -51,6 +51,7 @@ _papplSystemAddPrinter(
   pthread_rwlock_unlock(&system->rwlock);
 
   _papplSystemConfigChanged(system);
+  papplSystemAddEvent(system, printer, NULL, PAPPL_EVENT_PRINTER_CREATED | PAPPL_EVENT_SYSTEM_CONFIG_CHANGED, NULL);
 }
 
 
@@ -69,7 +70,7 @@ papplSystemFindPrinter(
     int            printer_id,		// I - Printer ID or `0`
     const char     *device_uri)		// I - Device URI or `NULL`
 {
-  int			i,		// Current printer index
+  size_t		i,		// Current printer index
 			count;		// Printer count
   pappl_printer_t	*printer = NULL;// Matching printer
 
@@ -92,12 +93,12 @@ papplSystemFindPrinter(
 
   // Loop through the printers to find the one we want...
   //
-  // Note: Cannot use cupsArrayFirst/Last since other threads might be
+  // Note: Cannot use cupsArrayGetFirst/Last since other threads might be
   // enumerating the printers array.
 
-  for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+  for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
   {
-    printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+    printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
     papplLog(system, PAPPL_LOGLEVEL_DEBUG, "papplSystemFindPrinter: printer '%s' - resource=\"%s\", printer_id=%d, device_uri=\"%s\"", printer->name, printer->resource, printer->printer_id, printer->device_uri);
 
