@@ -1,7 +1,7 @@
 //
 // System object for the Printer Application Framework
 //
-// Copyright © 2019-2020 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 // Copyright © 2010-2019 by Apple Inc.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
@@ -160,7 +160,8 @@ papplSystemCreate(
     bool             tls_only)		// I - Only support TLS connections?
 {
   pappl_system_t	*system;	// System object
-  const char		*tmpdir;	// Temporary directory
+  const char		*tmpdir = papplGetTempDir();
+					// Temporary directory
 
 
   if (!name)
@@ -174,42 +175,37 @@ papplSystemCreate(
   pthread_rwlock_init(&system->rwlock, NULL);
   pthread_rwlock_init(&system->session_rwlock, NULL);
   pthread_mutex_init(&system->config_mutex, NULL);
+  pthread_mutex_init(&system->subscription_mutex, NULL);
+  pthread_cond_init(&system->subscription_cond, NULL);
 
-  system->options         = options;
-  system->start_time      = time(NULL);
-  system->name            = strdup(name);
-  system->dns_sd_name     = strdup(name);
-  system->port            = port;
-  system->directory       = spooldir ? strdup(spooldir) : NULL;
-  system->logfd           = -1;
-  system->logfile         = logfile ? strdup(logfile) : NULL;
-  system->loglevel        = loglevel;
-  system->logmaxsize      = 1024 * 1024;
-  system->next_client     = 1;
-  system->next_printer_id = 1;
-  system->subtypes        = subtypes ? strdup(subtypes) : NULL;
-  system->tls_only        = tls_only;
-  system->admin_gid       = (gid_t)-1;
-  system->auth_service    = auth_service ? strdup(auth_service) : NULL;
+  system->options           = options;
+  system->start_time        = time(NULL);
+  system->name              = strdup(name);
+  system->dns_sd_name       = strdup(name);
+  system->port              = port;
+  system->directory         = spooldir ? strdup(spooldir) : NULL;
+  system->logfd             = -1;
+  system->logfile           = logfile ? strdup(logfile) : NULL;
+  system->loglevel          = loglevel;
+  system->logmaxsize        = 1024 * 1024;
+  system->next_client       = 1;
+  system->next_printer_id   = 1;
+  system->subtypes          = subtypes ? strdup(subtypes) : NULL;
+  system->tls_only          = tls_only;
+  system->admin_gid         = (gid_t)-1;
+  system->auth_service      = auth_service ? strdup(auth_service) : NULL;
+  system->max_subscriptions = 100;
+
+  papplSystemSetMaxClients(system, 0);
 
   if (!system->name || !system->dns_sd_name || (spooldir && !system->directory) || (logfile && !system->logfile) || (subtypes && !system->subtypes) || (auth_service && !system->auth_service))
     goto fatal;
 
   // Make sure the system name and UUID are initialized...
-  papplSystemSetHostname(system, NULL);
+  papplSystemSetHostName(system, NULL);
   papplSystemSetUUID(system, NULL);
 
-  // Set the system TLS credentials...
-  cupsSetServerCredentials(NULL, system->hostname, 1);
-
   // See if the spool directory can be created...
-  if ((tmpdir = getenv("TMPDIR")) == NULL)
-#ifdef __APPLE__
-    tmpdir = "/private/tmp";
-#else
-    tmpdir = "/tmp";
-#endif // __APPLE__
-
   if (!system->directory)
   {
     char	newspooldir[256];	// Spool directory
@@ -257,6 +253,9 @@ papplSystemCreate(
   papplSystemAddMIMEFilter(system, "image/png", "image/pwg-raster", _papplJobFilterPNG, NULL);
 #endif // HAVE_LIBPNG
 
+  // Load base localizations...
+  _papplLocLoadAll(system);
+
   return (system);
 
   // If we get here, something went wrong...
@@ -278,7 +277,7 @@ void
 papplSystemDelete(
     pappl_system_t *system)		// I - System object
 {
-  int	i;				// Looping var
+  size_t	i;			// Looping var
 
 
   if (!system || system->is_running)
@@ -292,10 +291,12 @@ papplSystemDelete(
   free(system->name);
   free(system->dns_sd_name);
   free(system->hostname);
+  free(system->domain_path);
   free(system->server_header);
   free(system->directory);
   free(system->logfile);
   free(system->subtypes);
+  free(system->auth_scheme);
   free(system->auth_service);
   free(system->admin_group);
   free(system->default_print_group);
@@ -304,11 +305,21 @@ papplSystemDelete(
     close(system->logfd);
 
   for (i = 0; i < system->num_listeners; i ++)
+#if _WIN32
+    closesocket(system->listeners[i].fd);
+#else
     close(system->listeners[i].fd);
+#endif // _WIN32
 
   cupsArrayDelete(system->filters);
   cupsArrayDelete(system->links);
   cupsArrayDelete(system->resources);
+  cupsArrayDelete(system->localizations);
+
+  _papplSystemCleanSubscriptions(system, true);
+  cupsArrayDelete(system->subscriptions);
+  pthread_cond_destroy(&system->subscription_cond);
+  pthread_mutex_destroy(&system->subscription_mutex);
 
   pthread_rwlock_destroy(&system->rwlock);
   pthread_rwlock_destroy(&system->session_rwlock);
@@ -329,7 +340,7 @@ char *					// I - UUID string
 _papplSystemMakeUUID(
     pappl_system_t *system,		// I - System
     const char     *printer_name,	// I - Printer name or `NULL` for none
-    int            job_id,		// I - Job ID or `0` for none
+    int            job_id,		// I - Job ID, negative subscription ID, or `0` for none
     char           *buffer,		// I - String buffer
     size_t         bufsize)		// I - Size of buffer
 {
@@ -341,8 +352,10 @@ _papplSystemMakeUUID(
   //
   // Start with the SHA2-256 sum of the hostname, port, object name and
   // number, and some random data on the end for jobs (to avoid duplicates).
-  if (printer_name && job_id)
-    snprintf(data, sizeof(data), "_PAPPL_JOB_:%s:%d:%s:%d:%08x", system->hostname, system->port, printer_name, job_id, _papplGetRand());
+  if (job_id < 0)			// Negative job ID == subscription ID
+    snprintf(data, sizeof(data), "_PAPPL_SUBSCRIPTION_:%s:%d:%s:%d:%08x", system->hostname, system->port, printer_name ? printer_name : "", -job_id, papplGetRand());
+  else if (printer_name && job_id)
+    snprintf(data, sizeof(data), "_PAPPL_JOB_:%s:%d:%s:%d:%08x", system->hostname, system->port, printer_name, job_id, papplGetRand());
   else if (printer_name)
     snprintf(data, sizeof(data), "_PAPPL_PRINTER_:%s:%d:%s", system->hostname, system->port, printer_name);
   else
@@ -368,16 +381,18 @@ _papplSystemMakeUUID(
 void
 papplSystemRun(pappl_system_t *system)	// I - System
 {
-  int			i,		// Looping var
+  size_t		i,		// Looping var
 			count;		// Number of listeners that fired
+  int			pcount;		// Poll count
   pappl_client_t	*client;	// New client
   char			header[HTTP_MAX_VALUE];
 					// Server: header value
   int			dns_sd_host_changes;
 					// Current number of host name changes
   pappl_printer_t	*printer;	// Current printer
-  pappl_printer_t	*usb_printer = NULL;
-					// USB printer
+  pthread_attr_t	tattr;		// Thread creation attributes
+  time_t		curtime,	// Current time
+			subtime = 0;	// Subscription checking time
 
 
   // Range check...
@@ -407,7 +422,7 @@ papplSystemRun(pappl_system_t *system)	// I - System
   {
     papplSystemAddResourceCallback(system, "/logfile.txt", "text/plain", (pappl_resource_cb_t)_papplSystemWebLogFile, system);
     papplSystemAddResourceCallback(system, "/logs", "text/html", (pappl_resource_cb_t)_papplSystemWebLogs, system);
-    papplSystemAddLink(system, "View Logs", "/logs", PAPPL_LOPTIONS_LOGGING | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+    papplSystemAddLink(system, _PAPPL_LOC("View Logs"), "/logs", PAPPL_LOPTIONS_LOGGING | PAPPL_LOPTIONS_HTTPS_REQUIRED);
   }
 
   if (system->options & PAPPL_SOPTIONS_WEB_INTERFACE)
@@ -416,39 +431,43 @@ papplSystemRun(pappl_system_t *system)	// I - System
     {
       papplSystemAddResourceCallback(system, "/", "text/html", (pappl_resource_cb_t)_papplSystemWebHome, system);
       papplSystemAddResourceCallback(system, "/addprinter", "text/html", (pappl_resource_cb_t)_papplSystemWebAddPrinter, system);
-      papplSystemAddLink(system, "Add Printer", "/addprinter", PAPPL_LOPTIONS_PRINTER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+      papplSystemAddLink(system, _PAPPL_LOC("Add Printer"), "/addprinter", PAPPL_LOPTIONS_PRINTER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
     }
     if (system->options & PAPPL_SOPTIONS_MULTI_QUEUE)
       papplSystemAddResourceCallback(system, "/config", "text/html", (pappl_resource_cb_t)_papplSystemWebConfig, system);
     if (system->options & PAPPL_SOPTIONS_WEB_NETWORK)
     {
       papplSystemAddResourceCallback(system, "/network", "text/html", (pappl_resource_cb_t)_papplSystemWebNetwork, system);
-      papplSystemAddLink(system, "Network", "/network", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+      papplSystemAddLink(system, _PAPPL_LOC("Network"), "/network", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+      if (system->wifi_join_cb && system->wifi_list_cb && system->wifi_status_cb)
+        papplSystemAddResourceCallback(system, "/network-wifi", "text/html", (pappl_resource_cb_t)_papplSystemWebWiFi, system);
     }
     if (system->options & PAPPL_SOPTIONS_WEB_SECURITY)
     {
       papplSystemAddResourceCallback(system, "/security", "text/html", (pappl_resource_cb_t)_papplSystemWebSecurity, system);
-      papplSystemAddLink(system, "Security", "/security", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+      papplSystemAddLink(system, _PAPPL_LOC("Security"), "/security", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
     }
-#ifdef HAVE_GNUTLS
+#if defined(HAVE_GNUTLS) || defined(HAVE_OPENSSL)
     if (system->options & PAPPL_SOPTIONS_WEB_TLS)
     {
       papplSystemAddResourceCallback(system, "/tls-install-crt", "text/html", (pappl_resource_cb_t)_papplSystemWebTLSInstall, system);
-      papplSystemAddLink(system, "Install TLS Certificate", "/tls-install-crt", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+      papplSystemAddLink(system, _PAPPL_LOC("Install TLS Certificate"), "/tls-install-crt", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
       papplSystemAddResourceCallback(system, "/tls-new-crt", "text/html", (pappl_resource_cb_t)_papplSystemWebTLSNew, system);
-      papplSystemAddLink(system, "Create New TLS Certificate", "/tls-new-crt", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+      papplSystemAddLink(system, _PAPPL_LOC("Create New TLS Certificate"), "/tls-new-crt", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
       papplSystemAddResourceCallback(system, "/tls-new-csr", "text/html", (pappl_resource_cb_t)_papplSystemWebTLSNew, system);
-      papplSystemAddLink(system, "Create TLS Certificate Request", "/tls-new-csr", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
+      papplSystemAddLink(system, _PAPPL_LOC("Create TLS Certificate Request"), "/tls-new-csr", PAPPL_LOPTIONS_OTHER | PAPPL_LOPTIONS_HTTPS_REQUIRED);
     }
-#endif // HAVE_GNUTLS
+#endif // HAVE_GNUTLS || HAVE_OPENSSL
   }
 
   // Catch important signals...
   papplLog(system, PAPPL_LOGLEVEL_INFO, "Starting system.");
 
+#if !_WIN32
   signal(SIGTERM, sigterm_handler);
   signal(SIGINT, sigterm_handler);
   signal(SIGHUP, sighup_handler);
+#endif // !_WIN32
 
   // Set the server header...
   free(system->server_header);
@@ -459,7 +478,7 @@ papplSystemRun(pappl_system_t *system)	// I - System
 
     // Replace spaces and other not-allowed characters in the firmware name
     // with an underscore...
-    strlcpy(safe_name, system->versions[0].name, sizeof(safe_name));
+    papplCopyString(safe_name, system->versions[0].name, sizeof(safe_name));
     for (safe_ptr = safe_name; *safe_ptr; safe_ptr ++)
     {
       if (*safe_ptr <= ' ' || *safe_ptr == '/' || *safe_ptr == 0x7f || (*safe_ptr & 0x80))
@@ -473,7 +492,7 @@ papplSystemRun(pappl_system_t *system)	// I - System
   {
     // If no version information is registered, just say "unknown" for the
     // main name...
-    strlcpy(header, "Unknown PAPPL/" PAPPL_VERSION " CUPS IPP/2.0", sizeof(header));
+    papplCopyString(header, "Unknown PAPPL/" PAPPL_VERSION " CUPS IPP/2.0", sizeof(header));
   }
 
   if ((system->server_header = strdup(header)) == NULL)
@@ -486,14 +505,18 @@ papplSystemRun(pappl_system_t *system)	// I - System
   // Make the static attributes...
   make_attributes(system);
 
+  // Start all child threads in a detached state...
+  pthread_attr_init(&tattr);
+  pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
+
   // Advertise the system via DNS-SD as needed...
   if (system->dns_sd_name)
     _papplSystemRegisterDNSSDNoLock(system);
 
   // Start up printers...
-  for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+  for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
   {
-    printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+    printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
     // Advertise via DNS-SD as needed...
     if (printer->dns_sd_name)
@@ -502,39 +525,25 @@ papplSystemRun(pappl_system_t *system)	// I - System
     // Start the raw socket listeners as needed...
     if ((system->options & PAPPL_SOPTIONS_RAW_SOCKET) && printer->num_raw_listeners > 0)
     {
-      pthread_t	tid;		// Thread ID
+      pthread_t	tid;			// Thread ID
 
-      if (pthread_create(&tid, NULL, (void *(*)(void *))_papplPrinterRunRaw, printer))
+      if (pthread_create(&tid, &tattr, (void *(*)(void *))_papplPrinterRunRaw, printer))
       {
 	// Unable to create listener thread...
 	papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to create raw listener thread: %s", strerror(errno));
       }
-      else
-      {
-	// Detach the main thread from the raw thread to prevent hangs...
-	pthread_detach(tid);
-      }
     }
   }
 
-  // Start the USB listener as needed...
-  if (system->options & PAPPL_SOPTIONS_USB_PRINTER)
+  // Start the USB gadget as needed...
+  if ((system->options & PAPPL_SOPTIONS_USB_PRINTER) && (printer = papplSystemFindPrinter(system, NULL, system->default_printer_id, NULL)) != NULL)
   {
-    // USB support is limited to a single (default) printer...
-    if ((usb_printer = papplSystemFindPrinter(system, NULL, system->default_printer_id, NULL)) != NULL)
-    {
-      pthread_t	tid;			// Thread ID
+    pthread_t	tid;			// Thread ID
 
-      if (pthread_create(&tid, NULL, (void *(*)(void *))_papplPrinterRunUSB, usb_printer))
-      {
-	// Unable to create USB thread...
-	papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to create USB printer thread: %s", strerror(errno));
-      }
-      else
-      {
-	// Detach the main thread from the raw thread to prevent hangs...
-	pthread_detach(tid);
-      }
+    if (pthread_create(&tid, &tattr, (void *(*)(void *))_papplPrinterRunUSB, printer))
+    {
+      // Unable to create USB thread...
+      papplLogPrinter(printer, PAPPL_LOGLEVEL_ERROR, "Unable to create USB gadget thread: %s", strerror(errno));
     }
   }
 
@@ -547,35 +556,45 @@ papplSystemRun(pappl_system_t *system)	// I - System
       _papplLogOpen(system);
     }
 
-    if ((count = poll(system->listeners, (nfds_t)system->num_listeners, 1000)) < 0 && errno != EINTR && errno != EAGAIN)
+    if ((pcount = poll(system->listeners, (nfds_t)system->num_listeners, 1000)) < 0 && errno != EINTR && errno != EAGAIN)
     {
       papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to accept new connections: %s", strerror(errno));
       break;
     }
 
-    if (count > 0)
+    if (pcount > 0)
     {
       // Accept client connections as needed...
-      for (i = 0; i < system->num_listeners; i ++)
+      for (i = 0; i < (size_t)system->num_listeners; i ++)
       {
 	if (system->listeners[i].revents & POLLIN)
 	{
-	  if ((client = _papplClientCreate(system, system->listeners[i].fd)) != NULL)
+	  if ((client = _papplClientCreate(system, (int)system->listeners[i].fd)) != NULL)
 	  {
-	    if (pthread_create(&client->thread_id, NULL, (void *(*)(void *))_papplClientRun, client))
+	    pthread_rwlock_wrlock(&system->rwlock);
+	    system->num_clients ++;
+	    pthread_rwlock_unlock(&system->rwlock);
+
+	    if (pthread_create(&client->thread_id, &tattr, (void *(*)(void *))_papplClientRun, client))
 	    {
 	      // Unable to create client thread...
 	      papplLog(system, PAPPL_LOGLEVEL_ERROR, "Unable to create client thread: %s", strerror(errno));
 	      _papplClientDelete(client);
 	    }
-	    else
-	    {
-	      // Detach the main thread from the client thread to prevent hangs...
-	      pthread_detach(client->thread_id);
-	    }
 	  }
 	}
       }
+
+      if (system->num_clients >= system->max_clients)
+      {
+	for (i = 0; i < (size_t)system->num_listeners; i ++)
+	  system->listeners[i].events = 0;
+      }
+    }
+    else if (system->num_clients < system->max_clients)
+    {
+      for (i = 0; i < (size_t)system->num_listeners; i ++)
+	system->listeners[i].events = POLLIN;
     }
 
     dns_sd_host_changes = _papplDNSSDGetHostChanges();
@@ -587,16 +606,16 @@ papplSystemRun(pappl_system_t *system)	// I - System
 					// Force re-registration?
 
       if (force_dns_sd)
-        papplSystemSetHostname(system, NULL);
+        papplSystemSetHostName(system, NULL);
 
       pthread_rwlock_rdlock(&system->rwlock);
 
       if (system->dns_sd_collision || force_dns_sd)
         _papplSystemRegisterDNSSDNoLock(system);
 
-      for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+      for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
       {
-	printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+	printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
         if (printer->dns_sd_collision || force_dns_sd)
           _papplPrinterRegisterDNSSDNoLock(printer);
@@ -626,7 +645,7 @@ papplSystemRun(pappl_system_t *system)	// I - System
     if (system->shutdown_time || sigterm_time)
     {
       // Shutdown requested, see if we can do so safely...
-      int		jcount = 0;	// Number of active jobs
+      size_t		jcount = 0;	// Number of active jobs
 
       // Force shutdown after 60 seconds
       if (system->shutdown_time && (time(NULL) - system->shutdown_time) > 60)
@@ -637,12 +656,12 @@ papplSystemRun(pappl_system_t *system)	// I - System
 
       // Otherwise shutdown immediately if there are no more active jobs...
       pthread_rwlock_rdlock(&system->rwlock);
-      for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+      for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
       {
-	printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+	printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
         pthread_rwlock_rdlock(&printer->rwlock);
-        jcount += cupsArrayCount(printer->active_jobs);
+        jcount += cupsArrayGetCount(printer->active_jobs);
         pthread_rwlock_unlock(&printer->rwlock);
       }
       pthread_rwlock_unlock(&system->rwlock);
@@ -651,9 +670,17 @@ papplSystemRun(pappl_system_t *system)	// I - System
         break;
     }
 
-    // Clean out old jobs...
-    if (system->clean_time && time(NULL) >= system->clean_time)
+    // Clean out old jobs and subscriptions...
+    curtime = time(NULL);
+
+    if (system->clean_time && curtime >= system->clean_time)
       papplSystemCleanJobs(system);
+
+    if (curtime >= subtime)
+    {
+      _papplSystemCleanSubscriptions(system, false);
+      subtime = curtime + 10;
+    }
   }
 
   papplLog(system, PAPPL_LOGLEVEL_INFO, "Shutting down system.");
@@ -661,12 +688,14 @@ papplSystemRun(pappl_system_t *system)	// I - System
   ippDelete(system->attrs);
   system->attrs = NULL;
 
+  pthread_attr_destroy(&tattr);
+
   if (system->dns_sd_name)
     _papplSystemUnregisterDNSSDNoLock(system);
 
-  for (i = 0, count = cupsArrayCount(system->printers); i < count; i ++)
+  for (i = 0, count = cupsArrayGetCount(system->printers); i < count; i ++)
   {
-    printer = (pappl_printer_t *)cupsArrayIndex(system->printers, i);
+    printer = (pappl_printer_t *)cupsArrayGetElement(system->printers, i);
 
     // Remove advertising via DNS-SD as needed...
     if (printer->dns_sd_name)
@@ -681,9 +710,10 @@ papplSystemRun(pappl_system_t *system)	// I - System
 
   system->is_running = false;
 
-  if (usb_printer)
+  if ((system->options & PAPPL_SOPTIONS_USB_PRINTER) && (printer = papplSystemFindPrinter(system, NULL, system->default_printer_id, NULL)) != NULL)
   {
-    while (usb_printer->usb_active)
+    // Wait for the USB gadget thread(s) to complete...
+    while (printer->usb_active)
       usleep(100000);
   }
 }
@@ -712,8 +742,42 @@ papplSystemShutdown(
 static void
 make_attributes(pappl_system_t *system)	// I - System
 {
-  int			i;		// Looping var
+  cups_len_t		i;		// Looping var
   ipp_attribute_t	*attr;		// Attribute
+  static const char * const charset[] =	// charset-supported values
+  {
+    "us-ascii",
+    "utf-8"
+  };
+  static const char * const ipp_versions[] =
+  {					// ipp-versions-supported values
+    "1.1",
+    "2.0"
+  };
+  static const int	operations[] =	// operations-supported values
+  {
+    IPP_OP_GET_PRINTER_ATTRIBUTES,
+    IPP_OP_GET_SUBSCRIPTION_ATTRIBUTES,
+    IPP_OP_GET_SUBSCRIPTIONS,
+    IPP_OP_RENEW_SUBSCRIPTION,
+    IPP_OP_CANCEL_SUBSCRIPTION,
+    IPP_OP_GET_NOTIFICATIONS,
+    IPP_OP_CREATE_PRINTER,
+    IPP_OP_DELETE_PRINTER,
+    IPP_OP_GET_PRINTERS,
+    IPP_OP_CREATE_SYSTEM_SUBSCRIPTIONS,
+    IPP_OP_DISABLE_ALL_PRINTERS,
+    IPP_OP_ENABLE_ALL_PRINTERS,
+    IPP_OP_GET_SYSTEM_ATTRIBUTES,
+    IPP_OP_GET_SYSTEM_SUPPORTED_VALUES,
+    IPP_OP_PAUSE_ALL_PRINTERS,
+    IPP_OP_PAUSE_ALL_PRINTERS_AFTER_CURRENT_JOB,
+    IPP_OP_RESUME_ALL_PRINTERS,
+    IPP_OP_SET_SYSTEM_ATTRIBUTES,
+    IPP_OP_SHUTDOWN_ALL_PRINTERS,
+    IPP_OP_CUPS_GET_DEFAULT,
+    IPP_OP_CUPS_GET_PRINTERS
+  };
   static const char * const printer_creation_attributes_supported[] =
   {					// "printer-creation-attributes-supported" Values
     "copies-default",
@@ -756,24 +820,60 @@ make_attributes(pappl_system_t *system)	// I - System
 
   system->attrs = ippNew();
 
+  // charset-configured
+  ippAddString(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_CHARSET), "charset-configured", NULL, "utf-8");
+
+  // charset-supported
+  ippAddStrings(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_CHARSET), "charset-supported", sizeof(charset) / sizeof(charset[0]), NULL, charset);
+
+  // generated-natural-language-supported
+  ippAddString(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_LANGUAGE), "generated-natural-language-supported", NULL, "en");
+
+  // ipp-versions-supported
+  ippAddStrings(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "ipp-versions-supported", (int)(sizeof(ipp_versions) / sizeof(ipp_versions[0])), NULL, ipp_versions);
+
+  // natural-language-configured
+  ippAddString(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_LANGUAGE), "natural-language-configured", NULL, "en");
+
+  // notify-events-default
+  ippAddString(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "notify-events-default", NULL, "job-completed");
+
+  // notify-events-supported
+  ippAddStrings(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "notify-events-supported", (int)(sizeof(_papplEvents) / sizeof(_papplEvents[0])), NULL, _papplEvents);
+
+  // notify-lease-duration-default
+  ippAddInteger(system->attrs, IPP_TAG_SYSTEM, IPP_TAG_INTEGER, "notify-lease-duration-default", PAPPL_LEASE_DEFAULT);
+
+  // notify-lease-duration-supported
+  ippAddRange(system->attrs, IPP_TAG_SYSTEM, "notify-lease-duration-supported", 0, PAPPL_LEASE_MAX);
+
+  // notify-max-events-supported
+  ippAddInteger(system->attrs, IPP_TAG_SYSTEM, IPP_TAG_INTEGER, "notify-max-events-supported", PAPPL_MAX_EVENTS);
+
+  // notify-pull-method-supported
+  ippAddString(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "notify-pull-method-supported", NULL, "ippget");
+
+  // operations-supported
+  ippAddIntegers(system->attrs, IPP_TAG_SYSTEM, IPP_TAG_ENUM, "operations-supported", (int)(sizeof(operations) / sizeof(operations[0])), operations);
+
   // printer-creation-attributes-supported
   ippAddStrings(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_KEYWORD), "printer-creation-attributes-supported", (int)(sizeof(printer_creation_attributes_supported) / sizeof(printer_creation_attributes_supported[0])), NULL, printer_creation_attributes_supported);
 
   // smi2699-device-command-supported
   if (system->num_drivers > 0)
   {
-    int num_drivers = system->num_drivers;
+    size_t num_drivers = system->num_drivers;
 					// Number of drivers to report
     if (system->autoadd_cb)
       num_drivers ++;
 
-    attr = ippAddStrings(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_NAME), "smi2699-device-command-supported", num_drivers, NULL, NULL);
+    attr = ippAddStrings(system->attrs, IPP_TAG_SYSTEM, IPP_CONST_TAG(IPP_TAG_NAME), "smi2699-device-command-supported", (cups_len_t)num_drivers, NULL, NULL);
 
-    for (i = 0; i < system->num_drivers; i ++)
+    for (i = 0; i < (cups_len_t)system->num_drivers; i ++)
       ippSetString(system->attrs, &attr, i, system->drivers[i].name);
 
     if (system->autoadd_cb)
-      ippSetString(system->attrs, &attr, system->num_drivers, "auto");
+      ippSetString(system->attrs, &attr, (cups_len_t)system->num_drivers, "auto");
   }
 
   // smi2699-device-uri-schemes-supported
