@@ -13,7 +13,7 @@
 
 #define PWG_DRIVER 1
 #include "testpappl.h"
-#include <pappl/base-private.h>		// For strlcpy
+#include <pappl/base-private.h>		// For papplCopyString
 #include <cups/dir.h>
 #include "label-png.h"
 
@@ -41,8 +41,6 @@ static const char * const pwg_2inch_media[] =
 
   "custom_max_2x3600in",
   "custom_min_0.25x0.25in"
-  "roll_max_2x3600in",
-  "roll_min_0.25x0.25in"
 };
 
 static const char * const pwg_4inch_media[] =
@@ -52,11 +50,10 @@ static const char * const pwg_4inch_media[] =
   "oe_multipurpose-label_2x2.3125in",
   "na_index-3x5_3x5in",
   "na_index-4x6_4x6in",
+//  "na_letter_8.5x11in",
 
   "custom_max_4x3600in",
   "custom_min_0.25x0.25in"
-  "roll_max_4x3600in",
-  "roll_min_0.25x0.25in"
 };
 
 static const char * const pwg_common_media[] =
@@ -103,7 +100,7 @@ pwg_autoadd(const char *device_info,	// I - Device information string (not used)
             const char *device_id,	// I - IEEE-1284 device ID
             void       *data)		// I - Callback data (not used)
 {
-  int		num_did;		// Number of device ID pairs
+  cups_len_t	num_did;		// Number of device ID pairs
   cups_option_t	*did;			// Device ID pairs
   const char	*cmd,			// Command set value
 		*ret = NULL;		// Return value
@@ -113,7 +110,7 @@ pwg_autoadd(const char *device_info,	// I - Device information string (not used)
   (void)device_uri;
   (void)data;
 
-  num_did = papplDeviceParseID(device_id, &did);
+  num_did = (cups_len_t)papplDeviceParseID(device_id, &did);
 
   if ((cmd = cupsGetOption("COMMAND SET", num_did, did)) == NULL)
     cmd = cupsGetOption("CMD", num_did, did);
@@ -224,9 +221,12 @@ pwg_callback(
     return (false);
   }
 
+  if (strstr(driver_name, "-pdf"))
+    driver_data->format = "application/pdf";
+
   if (!strncmp(driver_name, "pwg_2inch-", 10))
   {
-    strlcpy(driver_data->make_and_model, "PWG 2-inch Label Printer", sizeof(driver_data->make_and_model));
+    papplCopyString(driver_data->make_and_model, "PWG 2-inch Label Printer", sizeof(driver_data->make_and_model));
 
     driver_data->kind       = PAPPL_KIND_LABEL | PAPPL_KIND_ROLL;
     driver_data->ppm        = 20;	// 20 labels per minute
@@ -234,12 +234,12 @@ pwg_callback(
     driver_data->bottom_top = 625;	// 1/8" top and bottom
 
     driver_data->num_media = (int)(sizeof(pwg_2inch_media) / sizeof(pwg_2inch_media[0]));
-    memcpy(driver_data->media, pwg_2inch_media, sizeof(pwg_2inch_media));
+    memcpy((void *)driver_data->media, pwg_2inch_media, sizeof(pwg_2inch_media));
 
     driver_data->num_source = 1;
     driver_data->source[0]  = "main-roll";
 
-    strlcpy(driver_data->media_ready[0].size_name, "oe_address-label_1.25x3.5in", sizeof(driver_data->media_ready[0].size_name));
+    papplCopyString(driver_data->media_ready[0].size_name, "oe_address-label_1.25x3.5in", sizeof(driver_data->media_ready[0].size_name));
 
     driver_data->darkness_configured = 53;
     driver_data->darkness_supported  = 16;
@@ -247,7 +247,7 @@ pwg_callback(
   }
   else if (!strncmp(driver_name, "pwg_4inch-", 10))
   {
-    strlcpy(driver_data->make_and_model, "PWG 4-inch Label Printer", sizeof(driver_data->make_and_model));
+    papplCopyString(driver_data->make_and_model, "PWG 4-inch Label Printer", sizeof(driver_data->make_and_model));
 
     driver_data->kind       = PAPPL_KIND_LABEL | PAPPL_KIND_ROLL;
     driver_data->ppm        = 20;	// 20 labels per minute
@@ -255,13 +255,13 @@ pwg_callback(
     driver_data->bottom_top = 1;	// Not quite borderless top and bottom
 
     driver_data->num_media = (int)(sizeof(pwg_4inch_media) / sizeof(pwg_4inch_media[0]));
-    memcpy(driver_data->media, pwg_4inch_media, sizeof(pwg_4inch_media));
+    memcpy((void *)driver_data->media, pwg_4inch_media, sizeof(pwg_4inch_media));
 
     driver_data->num_source = 1;
     driver_data->source[0]  = "main-roll";
 
-    strlcpy(driver_data->media_ready[0].size_name, "na_index-4x6_4x6in", sizeof(driver_data->media_ready[0].size_name));
-    strlcpy(driver_data->media_ready[1].size_name, "oe_address-label_1.25x3.5in", sizeof(driver_data->media_ready[1].size_name));
+    papplCopyString(driver_data->media_ready[0].size_name, "na_index-4x6_4x6in", sizeof(driver_data->media_ready[0].size_name));
+    papplCopyString(driver_data->media_ready[1].size_name, "oe_address-label_1.25x3.5in", sizeof(driver_data->media_ready[1].size_name));
 
     driver_data->darkness_configured = 53;
     driver_data->darkness_supported  = 16;
@@ -269,7 +269,7 @@ pwg_callback(
   }
   else if (!strncmp(driver_name, "pwg_common-", 11))
   {
-    strlcpy(driver_data->make_and_model, "PWG Office Printer", sizeof(driver_data->make_and_model));
+    papplCopyString(driver_data->make_and_model, "PWG Office Printer", sizeof(driver_data->make_and_model));
 
     driver_data->has_supplies = true;
     driver_data->kind         = PAPPL_KIND_DOCUMENT | PAPPL_KIND_PHOTO | PAPPL_KIND_POSTCARD;
@@ -282,7 +282,7 @@ pwg_callback(
     driver_data->finishings = PAPPL_FINISHINGS_PUNCH | PAPPL_FINISHINGS_STAPLE;
 
     driver_data->num_media = (int)(sizeof(pwg_common_media) / sizeof(pwg_common_media[0]));
-    memcpy(driver_data->media, pwg_common_media, sizeof(pwg_common_media));
+    memcpy((void *)driver_data->media, pwg_common_media, sizeof(pwg_common_media));
 
     driver_data->num_source = 4;
     driver_data->source[0]  = "main";
@@ -304,8 +304,8 @@ pwg_callback(
       driver_data->bin[0]  = "center";
     }
 
-    strlcpy(driver_data->media_ready[0].size_name, "na_letter_8.5x11in", sizeof(driver_data->media_ready[0].size_name));
-    strlcpy(driver_data->media_ready[1].size_name, "iso_a4_210x297mm", sizeof(driver_data->media_ready[1].size_name));
+    papplCopyString(driver_data->media_ready[0].size_name, "na_letter_8.5x11in", sizeof(driver_data->media_ready[0].size_name));
+    papplCopyString(driver_data->media_ready[1].size_name, "iso_a4_210x297mm", sizeof(driver_data->media_ready[1].size_name));
 
     driver_data->sides_supported = PAPPL_SIDES_ONE_SIDED | PAPPL_SIDES_TWO_SIDED_LONG_EDGE | PAPPL_SIDES_TWO_SIDED_SHORT_EDGE;
     driver_data->sides_default   = PAPPL_SIDES_TWO_SIDED_LONG_EDGE;
@@ -337,9 +337,9 @@ pwg_callback(
     driver_data->media_default.size_width    = 21590;
     driver_data->media_default.size_length   = 27940;
     driver_data->media_default.top_margin    = driver_data->bottom_top;
-    strlcpy(driver_data->media_default.size_name, "na_letter_8.5x11in", sizeof(driver_data->media_default.size_name));
-    strlcpy(driver_data->media_default.source, "main", sizeof(driver_data->media_default.source));
-    strlcpy(driver_data->media_default.type, "stationery", sizeof(driver_data->media_default.type));
+    papplCopyString(driver_data->media_default.size_name, "na_letter_8.5x11in", sizeof(driver_data->media_default.size_name));
+    papplCopyString(driver_data->media_default.source, "main", sizeof(driver_data->media_default.source));
+    papplCopyString(driver_data->media_default.type, "stationery", sizeof(driver_data->media_default.type));
   }
   else
   {
@@ -433,8 +433,8 @@ pwg_callback(
       driver_data->media_ready[i].size_length   = pwg->length;
       driver_data->media_ready[i].top_margin    = driver_data->bottom_top;
       driver_data->media_ready[i].tracking      = PAPPL_MEDIA_TRACKING_MARK;
-      strlcpy(driver_data->media_ready[i].source, driver_data->source[i], sizeof(driver_data->media_ready[i].source));
-      strlcpy(driver_data->media_ready[i].type, driver_data->type[0], sizeof(driver_data->media_ready[i].type));
+      papplCopyString(driver_data->media_ready[i].source, driver_data->source[i], sizeof(driver_data->media_ready[i].source));
+      papplCopyString(driver_data->media_ready[i].type, driver_data->type[0], sizeof(driver_data->media_ready[i].type));
     }
   }
 
@@ -617,7 +617,7 @@ pwg_rstartjob(
 
   papplJobSetData(job, pwg);
 
-  pwg->ras = cupsRasterOpenIO((cups_raster_iocb_t)papplDeviceWrite, device, CUPS_RASTER_WRITE_PWG);
+  pwg->ras = cupsRasterOpenIO((cups_raster_cb_t)papplDeviceWrite, device, CUPS_RASTER_WRITE_PWG);
 
   return (1);
 }
@@ -642,7 +642,7 @@ pwg_rstartpage(
 
   memset(pwg->colorants, 0, sizeof(pwg->colorants));
 
-  return (cupsRasterWriteHeader2(pwg->ras, &options->header) != 0);
+  return (cupsRasterWriteHeader(pwg->ras, &options->header) != 0);
 }
 
 
@@ -834,7 +834,7 @@ pwg_testpage(
   else
     testfile = "portrait-gray.png";
 
-  strlcpy(buffer, testfile, bufsize);
+  papplCopyString(buffer, testfile, bufsize);
   if (access(buffer, R_OK))
     snprintf(buffer, bufsize, "testsuite/%s", testfile);
 
