@@ -1,16 +1,24 @@
 //
 // Xcode configuration header file for the Printer Application Framework
 //
-// Copyright © 2019-2021 by Michael R Sweet.
+// Copyright © 2019-2022 by Michael R Sweet.
 //
 // Licensed under Apache License v2.0.  See the file "LICENSE" for more
 // information.
 //
 
 // Version numbers
-#define PAPPL_VERSION		"1.0.3"
+#define PAPPL_VERSION		"1.2.1"
 #define PAPPL_VERSION_MAJOR	1
-#define PAPPL_VERSION_MINOR	0
+#define PAPPL_VERSION_MINOR	2
+
+
+// Location of PAPPL state and spool data (when run as root)
+#define PAPPL_STATEDIR		"/Library"
+
+
+// Location of PAPPL domain socket (when run as root)
+#define PAPPL_SOCKDIR		"/private/var/run"
 
 
 // Location of CUPS config files
@@ -19,11 +27,13 @@
 
 // DNS-SD (mDNSResponder or Avahi)
 #define HAVE_DNSSD 1
+#define HAVE_MDNSRESPONDER 1
 /* #undef HAVE_AVAHI */
 
 
-// GNU TLS
+// GNU TLS, LibreSSL/OpenSSL
 /* #undef HAVE_GNUTLS */
+#define HAVE_OPENSSL 1
 
 
 // libjpeg
